/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.annotation.flowControl;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * Makes the compiler use the inner-class as scope for the sub-elements of the matched html-element.
 * <p/>
 * The complete element is processed using the scope, and not just its childs.
 * An instance of the inner-class is always created, even if the element is omitted using EIf.
 * <p/>
 * todo: web-page with examples.
 * todo: add link to page describing scope-instance-creation-method.
 * <p/>
 * Must be applied to an inner-class, not an outer-class.
 *
 * @see EScopeChilds
 * @see EScopingWhile
 * @see EIterate
 */
@Target({ElementType.TYPE})
@Retention(RetentionPolicy.RUNTIME)
public @interface EScopeTag
{
	/**
	 * Makes the annotation match html elements which contains the value in the 'id' attribute or as a name in the 'class' attribute.
	 * Leading underscores in the names specified in the html are ignored when comparing the names.
	 * <p/>
	 * When the parameter is specified, the 'id', 'className' or 'tag' parameters must not be specified.
	 */
	public abstract String[] value() default {};

	/**
	 * Makes the annotation match html elements which contains the value in the 'id' attribute.
	 * Leading underscores in the 'id' attribute value are ignored when comparing the names.
	 * <p/>
	 * When the parameter is specified, the 'value' parameter must not be specified. It is allowed to specify
	 * the 'className' and 'tag' parameters.
	 */
	public abstract String[] id() default {};

	/**
	 * Makes the annotation match html elements which contains the value as a name in the 'class' attribute.
	 * Leading underscores in the names specified in the html are ignored when comparing the names.
	 * <p/>
	 * When the parameter is specified, the 'value' parameter must not be specified. It is allowed to specify
	 * the 'id' and 'tag' parameters.
	 */
	public abstract String[] className() default {};

	/**
	 * Makes the annotation match html elements whose tag is the specified value.
	 * <p/>
	 * When the parameter is specified, the 'value' parameter must not be specified. It is allowed to specify
	 * the 'id' and 'className' parameters.
	 */
	public abstract String[] tag() default {};

	/**
	 * Specifies whether the compiler should emit a warning (the default) if the annotation doesn't match any html element at all.
	 */
	public abstract boolean warnIfUnused() default true;
}