/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.application.scope;

import java.util.Arrays;
import java.util.EnumSet;
import org.jabhts.annotation.global.TOverrides;
import org.jabhts.translator.compiler.application.baseclass.InvalidApplicationException;
import org.jabhts.translator.compiler.utils.AnnotationHelper;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jabhts.translator.compiler.utils.CompilerUtils;
import org.jabhts.translator.compiler.utils.InternalCompilerException;
import org.jabhts.translator.compiler.utils.LogLevel;

public class TargetingInfo extends AnnotationHelper
{
	public final Scope scope;
	public final boolean isInherited, warnIfUnused;
	public final String[] defparam, id, className, tag;
	public final boolean hasTOverrides;
	//
	public final String basename;
	public final Type type;

	public enum Type
	{
		DefaultParam, Triplets, Appliance
	}

	public TargetingInfo(Scope scope, AnnotationHelper helper, boolean isInherited) throws CompilerException
	{
		super(helper.log, helper.annotation, helper.cls, helper.field, helper.method);
		this.scope = scope;
		this.isInherited = isInherited;
		this.defparam = getAnnotationStringArrayParam("value");
		this.id = getAnnotationStringArrayParam("id");
		this.className = getAnnotationStringArrayParam("className");
		this.tag = getAnnotationStringArrayParam("tag");
		this.warnIfUnused = getAnnotationBooleanParam("warnIfUnused");
		if (method != null)
			hasTOverrides = method.isAnnotationPresent(TOverrides.class);
		else if (field != null)
			hasTOverrides = field.isAnnotationPresent(TOverrides.class);
		else
			hasTOverrides = (cls != null) && cls.isAnnotationPresent(TOverrides.class);
		String bn;
		if (defparam.length > 0) {
			type = Type.DefaultParam;
			if (id.length + className.length + tag.length > 0) {
				throw new InvalidApplicationException("id/className/tag may not be specified" +
						" when the default parameter is specified: " + annotationToString());
			}
			bn = defparam[0];
		} else if (id.length + className.length + tag.length > 0) {
			bn = (id.length > 0) ? id[0] : ((className.length > 0) ? className[0] : tag[0]);
			type = Type.Triplets;
		} else if (method != null) {
			bn = CompilerUtils.stripLowercasePrefix(method.getName());
			if (bn.isEmpty())
				bn = method.getName();
			type = Type.Appliance;
		} else if (field != null) {
			bn = field.getName();
			type = Type.Appliance;
		} else if (cls.getEnclosingClass() != null) {
			// innerclass
			bn = cls.getSimpleName();
			type = Type.Appliance;
		} else {
			throw new InvalidApplicationException("id/className/tag or default-parm must be specified" +
					" when annotation is applied directly on the class: " + annotationToString());
		}
		if (bn.isEmpty())
			throw new InternalCompilerException("bn is null for " + annotationToString());
		// make initial letter uppercase
		if (!Character.isJavaIdentifierStart(bn.charAt(0)))
			bn = "X" + bn;
		basename = CompilerUtils.makeFirstCharacterUppercase(bn);
	}

	public String targetToString()
	{
		return type.name()
				+ (defparam.length > 0 ? "{defparam = " + Arrays.toString(defparam) + "}" : "")
				+ (id.length > 0 ? "{id = " + Arrays.toString(id) + "}" : "")
				+ (className.length > 0 ? "{class = " + Arrays.toString(className) + "}" : "")
				+ (tag.length > 0 ? "{tag = " + Arrays.toString(tag) + "}" : "")
				;
	}

	protected EnumSet<SelectorType> matchedBy = EnumSet.noneOf(SelectorType.class);

	/**
	 * Huge problem: it doesn't know which of the params that was matched in this manner,
	 * so if one default param matches by id and another by class, then an "Matched by..."
	 * warning is created.
	 */
	public void notifyMatched(SelectorType st)
	{
		// 	Id, Class, Tag
		if (matchedBy.contains(st))
			return;
		matchedBy.add(st);
	}

	protected boolean hasMadePostCompilationWarnings;

	/**
	 * can be called several times
	 */
	public void emitPostCompilationWarnings()
	{
		if (hasMadePostCompilationWarnings)
			return;
		if (warnIfUnused && matchedBy.isEmpty() && !isInherited) {
			log.log(LogLevel.Notice, "Not matched by html: " + annotationToString());
			hasMadePostCompilationWarnings = true;
		}
		if (type == Type.DefaultParam && matchedBy.size() > 1) {
			StringBuilder sb = new StringBuilder();
			sb.append("Matched by  ");
			int num = 0;
			for (SelectorType st : matchedBy) {
				sb.append(++num == 1 ? "" : " and ");
				sb.append(st.description);
			}
			sb.append(": ").append(annotationToString());
			log.log(LogLevel.Notice, sb.toString());
			hasMadePostCompilationWarnings = true;
		}
	}
}
