/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.codeGeneration.impl;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;
import org.jabhts.runtime.lib.Opcode;
import org.jabhts.runtime.lib.TemplateBase;
import org.jabhts.runtime.lib.TemplateData;
import org.jabhts.translator.compiler.application.scope.ScopeImpl;
import org.jabhts.translator.compiler.codeGeneration.util.ClassType;
import org.jabhts.translator.compiler.codeGeneration.util.JavaClass;
import org.jabhts.translator.compiler.codeGeneration.util.JavaPrinter;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jabhts.translator.compiler.utils.CompilerLog;
import org.jabhts.translator.compiler.utils.CompilerUtils;
import org.jabhts.translator.compiler.utils.LogLevel;
import org.jabhts.translator.globals.CompilerOptions;
import org.jabhts.translator.globals.ProgressInfo;

public class ProgramEmit
{
	protected final CompilerOptions options;
	protected final CompilerLog log;
	protected final ProgramData data;
	protected final String packageNameDot, fullClassName, simpleClassName;
	protected final List<Callback> uniqueIterCallbacks;

	protected JavaClass jclass;
	protected JavaPrinter jp;

	public TemplateData templateData;

	protected static final String md5Marker = "MD5.MD5.MD5.MD5.MD5.MD5.MD5.MD5.";

	public ProgramEmit(CompilerOptions options, CompilerLog log, ProgramData data, String packageName, String fullClassName) throws CompilerException
	{
		this.options = options;
		this.log = log;
		this.data = data;
		this.fullClassName = fullClassName;
		this.packageNameDot = packageName + ".";
		this.simpleClassName = fullClassName.substring(packageNameDot.length());
		if (Opcode.values().length > Opcode.OPCODE_MASK)
			throw new CompilerException("Invalid Opcode.OPCODE_MASK: too few bits.");
		Set<String> seen = new HashSet<>();
		uniqueIterCallbacks = new ArrayList<>(); // must keep sorting
		for (Callback cb : data.callbacks) {
			/*
			System.err.println("CB." + cb.callbackType + ": " + cb.scope.scopeName
				+ (cb.method != null ? ", m = " + cb.method.getName() : "")
				+ (cb.field != null ? ", f = " + cb.field.getName() : ""));
			*/
			if (cb.iterInfo == null || cb.callbackType != Callback.CallbackType.EnterScope)
				continue;
			//System.err.println("SEE: " + cb.scope.scopeName + " for " + cb.iterInfo.description);
			if (seen.contains(cb.iterInfo.itemVar))
				continue;
			seen.add(cb.iterInfo.itemVar);
			uniqueIterCallbacks.add(cb);
		}
	}

	public void main(ProgressInfo pi) throws CompilerException
	{
		try {
			jclass = new EmitJavaClass(options, fullClassName);
			jclass.encoding = options.sourceEncoding;
			jclass.addImport(TemplateBase.class);
			jclass.classExtends = "TemplateBase";
			jclass.isFinal = true;
			jp = new JavaPrinter(jclass.pw, 1);
			makeConstructor();
			makeRootScopeField();
			makeMagicCommentStart();
			makeEmitSetup();
			makeSubScopeFields();
			makeCallbackFields();
			makeCallbacks();
			//if (options.verbose)
			//	System.err.println(jclass.getCode());
			File dirF = jclass.javaF.getParentFile();
			if (!dirF.exists() && !dirF.mkdirs())
				throw new IOException("Can't create dir: " + dirF.getPath());
			if (options.saveSourceToDisk) {
				if (jclass.save()) {
					pi.numJavaSourcesUpdated++;
					log.log(LogLevel.Notice, "Updated java source: " + jclass.shortClassName);
				}
			} else {
				jclass.getCode(); // to calc MD5s
			}
			if (options.verbose) {
				dumpOpcodes();
				dumpStringTable();
			}
		}
		catch (CompilerException ex) {
			throw ex;
		}
		catch (Exception ex) {
			throw new CompilerException(ex);
		}
	}

	public File getGeneratedJavaF()
	{
		return jclass.javaF;
	}

	public class EmitJavaClass extends JavaClass
	{
		public EmitJavaClass(CompilerOptions options, String fullClassName)
		{
			super(options.generatedSourceBaseDirF, fullClassName, ClassType.Class);
			addGeneratedAnnotation("org.jabhts");
		}

		public String getCode() throws Exception
		{
			String s = super.getCode();
			String md5 = CompilerUtils.md5LcHex(s);
			templateData = new TemplateData(data.texts, data.opcodes, md5);
			int idx = s.indexOf(md5Marker);
			return s.substring(0, idx) + md5 + s.substring(idx + md5Marker.length());
		}
	}

	protected void dumpStringTable() throws CompilerException
	{
		System.err.println("--------- string table");
		StringBuilder sb = new StringBuilder(200);
		for (int i = 0; i < data.texts.length; i++) {
			String s = data.texts[i];
			System.err.println("[" + i + "]: {{" + s + "}}");
			sb.append(s);
		}
		int chars = sb.length();
		int bytes;
		bytes = sb.toString().getBytes(StandardCharsets.UTF_8).length;
		System.err.println("--------- end, " + chars + " chars, " + bytes + " UTF-8 bytes.");
	}

	protected void dumpOpcodes() throws CompilerException
	{
		System.err.println("--------- opcodes");
		for (int i = 0; i < data.opcodes.length; i++) {
			int encoded = data.opcodes[i];
			String txt = String.format("%8s", "[" + i + "]: ");
			try {
				int param = Opcode.decodeParam(encoded);
				String s;
				switch (Opcode.decodeOpcode(encoded)) {
					case PRINT_STRINGTABLE:
						s = data.texts[param];
						if (s.length() > 40)
							s = s.substring(0, 20) + "[...]" + s.substring(s.length() - 20);
						s = s.replaceAll("\\n", "\\\\n");
						txt += String.format("%-25s %s", Opcode.toString(encoded), s);
						break;
					case CALLBACK:
						Callback cb = data.callbacks[param];
						txt += String.format("%-25s %s", Opcode.toString(encoded), cb.key);
						//txt += ":             " + cb.key;//cb.callbackType.name();
						break;
					default:
						txt += Opcode.toString(encoded);
				}
			}
			catch (Exception ex) {
				txt += ": INVALID encoded = " + encoded + ": " + CompilerUtils.getSomeMessage(ex);
			}
			System.err.println(txt);
		}
		int stringsLen = 0;
		//noinspection UnusedCatchParameter
		for (String s : data.texts)
			stringsLen += s.getBytes(StandardCharsets.UTF_8).length;
		System.err.println("--------- end, " + data.opcodes.length * 4 + " bytes, strings = " + stringsLen + " bytes.");
	}

	protected void makeConstructor() throws CompilerException
	{
		jp.println("public " + jclass.shortClassName + "("
				+ relClassName(data.rootScope) + " " + data.rootScope.scopeFieldName + ")");
		jp.println("{");
		jp.println("super(\"" + md5Marker + "\");");
		jp.println("this." + data.rootScope.scopeFieldName + " = " + data.rootScope.scopeFieldName + ";");
		jp.println("}");
	}

	protected void makeMagicCommentStart() throws CompilerException
	{
		jp.println();
		jp.println("/*");
		jp.println("// remove this line to comment out rest of class" +
				" (to temporary silence compile errors) */");
	}

	protected void makeRootScopeField() throws CompilerException
	{
		jp.println();
		jp.println("final " + relClassName(data.rootScope) + " " + data.rootScope.scopeFieldName + ";");
	}

	protected void makeSubScopeFields() throws CompilerException
	{
		for (ScopeImpl scope : data.scopes) {
			if (scope == data.rootScope)
				continue;
			jp.println();
			jp.println(relClassName(scope) + " " + scope.scopeFieldName + ";");
		}
	}

	protected void makeCallbackFields() throws CompilerException
	{
		for (Callback cb : uniqueIterCallbacks) {
			jclass.addImport(Iterator.class);
			jp.println();
			jp.println("// callback fields for " + cb.iterInfo.description);
			jp.println("Iterator<" + relClassName(cb.iterInfo.itemType) + "> " + cb.iterInfo.iterVar + ";");
			jp.println(relClassName(cb.iterInfo.itemType) + " " + cb.iterInfo.itemVar + ";");
		}
	}

	protected void makeEmitSetup() throws CompilerException
	{
	}

	protected void makeCallbacks() throws CompilerException
	{
		jclass.addImport(SuppressWarnings.class);
		jp.println();
		jp.println("@SuppressWarnings(\"ConstantConditions\")");
		jp.println("protected void executeCallback(int idx) throws Exception");
		jp.println("{");
		jp.println("switch (idx) {");
		for (int cbIdx = 0; cbIdx < data.callbacks.length; cbIdx++) {
			Callback cb = data.callbacks[cbIdx];
			jp.println("case " + cbIdx + ":");
			jp.indent++;
			switch (cb.callbackType) {
				case CallMethod:
					if (cb.iterInfo != null) {
						emitSetupIterator(cb, false);
						break;
					}
					Class<?> t = cb.method.getReturnType();
					if (t.equals(String.class))
						jp.print("stringValue = ");
					else if (t.equals(Integer.class) || t.getName().equals("int"))
						jp.print("intValue = ");
					else if (t.equals(Boolean.class) || t.getName().equals("boolean"))
						jp.print("boolValue = ");
					jp.println(getMethodRef(cb.scope, cb.method) + "();");
					break;

				case GetStringField:
					jp.println("stringValue = " + getFieldRef(cb.scope, cb.field) + ";");
					break;

				case GetIntegerField:
					jp.println("intValue = " + getFieldRef(cb.scope, cb.field) + ";");
					break;

				case GetBooleanField:
					jp.println("boolValue = " + getFieldRef(cb.scope, cb.field) + ";");
					break;

				case GetIteratorField:
					emitSetupIterator(cb, true);
					break;

				case EnterScope:
					jp.print(cb.scope.scopeFieldName + " = ");
					String prefix = "";
					String whoseFault = null;
					if (cb.scope.scopeInstanceCreatorMethod != null) {
						jp.print(getMethodRef(cb.scope.parent, cb.scope.scopeInstanceCreatorMethod) + "(");
						whoseFault = relClassName(cb.scope.parent) + "." + cb.scope.scopeInstanceCreatorMethod.getName() + "()";
					} else if (Modifier.isStatic(cb.scope.scopeClass.getModifiers())) {
						jp.print("new " + relClassName(cb.scope) + "(");
					} else {
						String s = relClassName(cb.scope);
						int z = s.lastIndexOf('.');
						if (z > 0)
							s = s.substring(z + 1);
						jp.print(cb.scope.parent.scopeFieldName + ".new " + s + "(");
					}
					if (cb.iterInfo != null) {
						jp.print(prefix);
						jp.print(cb.iterInfo.itemVar);
					}
					jp.println(");");
					if (whoseFault != null) {
						jp.println("if (" + cb.scope.scopeFieldName + " == null)");
						jp.println("throw new NullPointerException(\"" + whoseFault + " returned null\");");
					}
					break;

				case LeaveScope:
					jp.println(cb.scope.scopeFieldName + " = null;");
					break;

				case IteratorStep:
					jp.println("boolValue = " + cb.iterInfo.iterVar + ".hasNext();");
					jp.println("if (boolValue) {");
					jp.println(cb.iterInfo.itemVar + " = " + cb.iterInfo.iterVar + ".next();");
					jp.println("} else {");
					jp.println(cb.iterInfo.itemVar + " = null;");
					jp.println(cb.iterInfo.iterVar + " = null;");
					jp.println("}");
					break;

				default:
					throw new CompilerException("Invalid type = " + cb.callbackType);
			}
			jp.println("break;");
			jp.println();
			jp.indent--;
		}
		jp.println("default:");
		jp.println("\tthrow new Exception(\"Unknown callback idx = \" + idx);");
		jp.println("}");
		jp.println("}");
	}

	protected void emitSetupIterator(Callback cb, boolean isField) throws CompilerException
	{
		String who = isField
				? relClassName(cb.scope) + "." + cb.field.getName()
				: relClassName(cb.scope) + "." + cb.method.getName();
		String get = isField ? getFieldRef(cb.scope, cb.field) : getMethodRef(cb.scope, cb.method) + "()";
		String isNullTxt = isField ? " is null" : " returned null";
		if (cb.iterInfo.isIterable) {
			jp.println("{");
			jp.println("Iterable<" + relClassName(cb.iterInfo.itemType) + "> able = " + get + ";");
			jp.println("if (able == null)");
			jp.println("throw new NullPointerException(\"" + who + isNullTxt + "\");");
			jp.println(cb.iterInfo.iterVar + " = able.iterator();");
			jp.println("if (" + cb.iterInfo.iterVar + " == null)");
			jp.println("throw new NullPointerException(\"" + who + ".iterator() returned null\");");
			jp.println("}");
		} else {
			jp.println(cb.iterInfo.iterVar + " = " + get + ";");
			jp.println("if (" + cb.iterInfo.iterVar + "== null)");
			jp.println("throw new NullPointerException(\"" + who + isNullTxt + "\");");
		}
	}

	protected String getMethodRef(ScopeImpl scope, Method method) throws CompilerException
	{
		checkAccess(method.getDeclaringClass(), method.getModifiers(), method.getName());
		if (Modifier.isStatic(method.getModifiers()))
			return relClassName(scope) + "." + method.getName();
		else
			return scope.scopeFieldName + "." + method.getName();
	}

	protected String getFieldRef(ScopeImpl scope, Field field) throws CompilerException
	{
		checkAccess(field.getDeclaringClass(), field.getModifiers(), field.getName());
		if (Modifier.isStatic(field.getModifiers()))
			return relClassName(scope) + "." + field.getName();
		else
			return scope.scopeFieldName + "." + field.getName();
	}

	protected void checkAccess(Class<?> cls, int modifiers, String name)
	{
		if (isSamePackage(cls)) {
			if (Modifier.isPrivate(modifiers))
				log.log(LogLevel.Error, relClassName(cls) + "." + name + " is not private: cannot be accessed");
		} else if (!Modifier.isPublic(modifiers)) {
			log.log(LogLevel.Error, relClassName(cls) + "." + name + " is not public: cannot be accessed");
		}
	}

	public static final Pattern javaLangRE = Pattern.compile("(?<![\\.\\w$])(java\\.lang\\.)");

	protected String relClassName(String fcn)
	{
		if (isSamePackage(fcn))
			return CompilerUtils.zapPrefix(fcn, packageNameDot).replace('$', '.');
		fcn = javaLangRE.matcher(fcn).replaceAll("");
		return fcn.replace('$', '.');
	}

	protected boolean isSamePackage(Class<?> cls)
	{
		return isSamePackage(cls.getName());
	}

	protected boolean isSamePackage(String s)
	{
		if (!s.startsWith(packageNameDot))
			return false;
		int idx = s.indexOf('$');
		if (idx > 0)
			s = s.substring(0, idx);
		return s.indexOf('.', packageNameDot.length()) < 0;
	}

	@SuppressWarnings({"ThrowableInstanceNeverThrown"})
	protected String relClassName(Class<?> cls)
	{
		if (cls == null)
			new Exception("cls is null").printStackTrace();
		return relClassName(cls.getName());
	}

	@SuppressWarnings({"ThrowableInstanceNeverThrown"})
	protected String relClassName(ScopeImpl scope)
	{
		if (scope == null)
			new Exception("scope is null").printStackTrace();
		return relClassName(scope.scopeClass);
	}
}
