/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.tree;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.jabhts.runtime.html.HtmlAttr;
import org.jabhts.runtime.html.HtmlTags;
import org.jabhts.translator.compiler.application.attribute.AttributeEmitter;
import org.jabhts.translator.compiler.application.attribute.PostAttributeEmitter;
import org.jabhts.translator.compiler.application.attribute.sub.SubAttributeEmitter;
import org.jabhts.translator.compiler.application.element.simple.EContentIfApplication;
import org.jabhts.translator.compiler.application.element.simple.EContentPrePostApplication;
import org.jabhts.translator.compiler.application.element.simple.EIfApplication;
import org.jabhts.translator.compiler.application.element.simple.ETagIfApplication;
import org.jabhts.translator.compiler.application.scope.Scope;
import org.jabhts.translator.compiler.codeGeneration.api.Program;
import org.jabhts.translator.compiler.htmlLoader.HtmlDocument;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jabhts.translator.compiler.utils.CompilerLog;
import org.jabhts.translator.compiler.utils.CompilerUtils;
import org.jabhts.translator.compiler.utils.LogLevel;
import org.jabhts.translator.globals.CompilerOptions;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

public class TreeEmit
{
	public final CompilerOptions options;
	public final Program program;
	public final CompilerLog log;

	boolean debugFlag = false;

	public TreeEmit(CompilerOptions options, Program program, CompilerLog log)
	{
		this.options = options;
		this.program = program;
		this.log = log;
	}

	public void emit(HtmlDocument hfi, boolean includeDoctype, boolean onlyChilds,
	                 Node rootE, Scope scope, String annotDescription) throws CompilerException
	{
		if (includeDoctype && hfi.docType != null) {
			program.addEmitString(hfi.docType);
			program.addSoftNewline();
		}
		if (onlyChilds) {
			ScopeLink rootScope = new ScopeLink(null, scope, annotDescription);
			for (Node childE = rootE.getFirstChild(); childE != null; childE = childE.getNextSibling())
				emitNode(childE, rootScope, 0, "", true);
		} else {
			emitNode(rootE, new ScopeLink(null, scope, annotDescription), 0, "", true);
		}
	}

	protected void emitNode(Node n, ScopeLink scopeLink, int level, String indent, boolean allowIdAttribute) throws CompilerException
	{
		switch (n.getNodeType()) {
			case Node.ELEMENT_NODE:
				emitElement((Element)n, scopeLink, level, indent, allowIdAttribute);
				break;
			case Node.TEXT_NODE:
				emitTextNode(n);
				break;
			case Node.COMMENT_NODE:
				emitCommentNode(n);
				break;
			default:
				error(n, "Unsupported node type for html");
		}
	}

	protected void emitElement(Element e, ScopeLink scopeLink, int level, String indent, boolean allowIdAttribute) throws CompilerException
	{
		String nodeName = e.getNodeName();
		if (debugFlag)
			log.log(LogLevel.Verbose, indent + "Element: " + nodeName);
		ElementEmitters emitters = new ElementEmitters(log, debugFlag, program, scopeLink, e, level, allowIdAttribute);
		if (!emitters.shouldProcessElement)
			return;
		if (emitters.eOuterScope != null)
			emitters.eOuterScope.preElement();
		for (EIfApplication.Emitter emit : emitters.eConditionals)
			emit.preElement();
		if (emitters.eLoop != null)
			emitters.eLoop.preElement();
		if (emitters.eInnerScope != null)
			emitters.eInnerScope.preElement();
		if (HtmlTags.elementsPrependNewline.contains(nodeName))
			program.addSoftNewline();
		if (emitters.eReplacer != null) {
			emitters.eReplacer.generateElement();
		} else {
			for (ETagIfApplication.Emitter emit : emitters.eTag)
				emit.preElementOpenTag();
			program.addEmitString("<");
			program.addEmitString(nodeName);
			emitAttributes(e, emitters.allowIdAttribute, emitters);
			program.addEmitString(">");
			Collections.reverse(emitters.eTag);
			for (ETagIfApplication.Emitter emit : emitters.eTag)
				emit.postElementOpenTag();
			//
			if (!HtmlTags.emptyElements.contains(nodeName)) {
				for (EContentIfApplication.Emitter emit : emitters.eContentIf)
					emit.preElementContent();
				for (EContentPrePostApplication.Emitter emit : emitters.ePrePostContent)
					emit.preContent();
				if (emitters.eContent != null) {
					emitters.eContent.generateContent();
				} else {
					String childIndent = indent + "  ";
					for (Node child = e.getFirstChild(); child != null; child = child.getNextSibling())
						emitNode(child, emitters.scopeLink, level + 1, childIndent, emitters.allowIdAttribute);
				}
				Collections.reverse(emitters.ePrePostContent);
				for (EContentPrePostApplication.Emitter emit : emitters.ePrePostContent)
					emit.postContent();
				//
				Collections.reverse(emitters.eContentIf);
				for (EContentIfApplication.Emitter emit : emitters.eContentIf)
					emit.postElementContent();
				//
				for (ETagIfApplication.Emitter emit : emitters.eTag)
					emit.preElementCloseTag();
				//
				program.addEmitString("</");
				program.addEmitString(nodeName);
				program.addEmitString(">");
				Collections.reverse(emitters.eTag);
				for (ETagIfApplication.Emitter emit : emitters.eTag)
					emit.postElementCloseTag();
			}
		}
		if (HtmlTags.elementsAppendNewline.contains(nodeName))
			program.addSoftNewline();
		if (emitters.eInnerScope != null)
			emitters.eInnerScope.postElement();
		if (emitters.eLoop != null)
			emitters.eLoop.postElement();
		for (EIfApplication.Emitter emit : emitters.eConditionals)
			emit.postElement();
		if (emitters.eOuterScope != null)
			emitters.eOuterScope.postElement();
	}

	protected void emitAttributes(Element e, boolean allowIdAttribute,
	                              ElementEmitters emitters) throws CompilerException
	{
		// make union set of attributes specified on the element and those we have emitters for.
		Set<String> nameSet = new HashSet<>();
		NamedNodeMap map = e.getAttributes();
		int num = map.getLength();
		for (int ai = 0; ai < num; ai++)
			nameSet.add(map.item(ai).getNodeName());
		nameSet.addAll(emitters.eAttributes.keySet());
		nameSet.addAll(emitters.eSubAttributes.keySet());
		//
		boolean snakeOil = e.getTagName().equals("meta") && nameSet.contains("name");
		if (snakeOil) {
			// Some webdesigners believe that the order of attrs is important.
			emitAttribute(e, allowIdAttribute, emitters, "name", snakeOil);
			nameSet.remove("name");
		}
		if (nameSet.isEmpty())
			return;
		List<String> list = new ArrayList<>(nameSet);
		list.sort(null);
		for (String attrName : list) {
			emitAttribute(e, allowIdAttribute, emitters, attrName, snakeOil);
		}
	}

	private void emitAttribute(Element e, boolean allowIdAttribute,
							   ElementEmitters emitters, String attrName,
							   boolean snakeOil) throws CompilerException
	{
		AttributeEmitter<?> valueEmitter = emitters.eAttributes.get(attrName);
		PostAttributeEmitter<?> postValueEmitter = emitters.ePostAttributes.get(attrName);
		List<SubAttributeEmitter<?>> subEmitters = emitters.eSubAttributes.get(attrName);
		String value = e.getAttribute(attrName);
		if (attrName.equals("id") && valueEmitter == null) {
			// if we have a processor, it is probably for generating unique id's, so keep it.
			if (!allowIdAttribute || value.startsWith("_"))
				return;
		}
		boolean empty = HtmlTags.emptyAttributes.contains(attrName);
		if (valueEmitter != null) {
			if (postValueEmitter != null) {
				attrError(e, attrName, "Having both plain attribute annotation " +
						"and post-content attribute annotation is not allowed");
			}
			if (subEmitters != null) {
				attrError(e, attrName, "Having both plain attribute annotation " +
						"and sub-attribute annotation is not allowed");
			}
			if (valueEmitter.processed) // attribute collection which is already generated.
				return;
			valueEmitter.processed = true;
			valueEmitter.generateAttribute();
			return;
		}
		if (postValueEmitter != null) {
			if (empty)
				attrError(e, attrName, "Can't make post-content for empty (boolean) attribute!");
			if (subEmitters != null) {
				attrError(e, attrName, "Having both post-content attribute annotation " +
						"and sub-attribute annotation is not allowed");
			}
		}
		if (subEmitters != null) {
			if (empty) {
				attrError(e, attrName, "Can't make sub-attrs for empty (boolean) attribute");
				return;
			}
			switch (attrName) {
				case "class":
					new ClassSubAttributesHandler(program, log, subEmitters).main(value);
					break;
				case "style":
					new StyleSubAttributesHandler(program, log, subEmitters).main(value);
					break;
				default:
					attrError(e, attrName, "Internal error: Unsupported sub attr name: " + attrName);
					break;
			}
			return;
		}
		if (empty) {
			program.addEmitChar(' ');
			program.addEmitString(attrName);
			return;
		}
		if (attrName.equals("class")) {
			value = CompilerUtils.zapUnderscorePrefixedClassNames(value);
			if (postValueEmitter == null && value.isEmpty())
				return;
		}
		program.addEmitChar(' ');
		program.addEmitString(attrName);
		if (postValueEmitter != null) {
			// always quote
			program.addEmitString("='");
			postValueEmitter.generatePostContent(value);
			program.addEmitString("'");
			return;
		}
		if (value == null || value.isEmpty()) {
			program.addEmitString("=''");
		} else if (snakeOil) {
			program.addEmitString("=");
			program.addEmitString(HtmlAttr.forceQuoted(value));
		} else {
			program.addEmitString("=");
			program.addEmitString(HtmlAttr.quoted(value));
		}
	}

	protected void error(Node e, String msg)
	{
		attrError(e, null, msg);
	}

	protected void attrError(Node e, String attrName, String msg)
	{
		StringBuilder sb = new StringBuilder();
		makeLocation(sb, e);
		if (attrName != null)
			sb.append("/@").append(attrName);
		sb.append(": ").append(msg);
		log.log(LogLevel.Error, sb.toString());
	}

	/**
	 * Appends path to current node to the StringBuilder, used for error reporting.
	 */
	protected void makeLocation(StringBuilder sb, Node e)
	{
		Node parent = e.getParentNode();
		if (parent != null && parent.getNodeType() != Node.DOCUMENT_NODE) {
			makeLocation(sb, parent);
			sb.append("/");
		}
		sb.append(e.getNodeName());
		if (e.getNodeType() != Node.ELEMENT_NODE)
			sb.append("[type=").append(e.getNodeType()).append("]");
	}

	protected void emitTextNode(Node n) throws CompilerException
	{
		Node parent = n.getParentNode();
		if (parent != null && parent.getNodeType() == Node.ELEMENT_NODE) {
			String name = parent.getNodeName();
			if (name.equals("style")) {
				emitStyleTagContent(n);
				return;
			}
			if (name.equals("script")) {
				emitScriptTagContent(n);
				return;
			}
		}
		boolean afterLF = false;
		Node prev = n.getPreviousSibling();
		if (prev != null && prev.getNodeType() == Node.ELEMENT_NODE && HtmlTags.elementsAppendNewline.contains(prev.getNodeName()))
			afterLF = true; // might have a goto-label at this text node which shouldn't start by emitting LF as we're already at a new line
		String s = n.getNodeValue();
		for (int i = 0; i < s.length(); i++) {
			char ch = s.charAt(i);
			if (ch == '\r') {
				if (i + 1 < s.length() && s.charAt(i + 1) == '\n')
					continue;
				ch = '\n';
			}
			if (ch == '\n') {
				if (!afterLF)
					program.addSoftNewline();
				afterLF = true;
				continue;
			}
			if (afterLF && (ch == ' ' || ch == '\t'))
				continue;
			afterLF = false;
			switch (ch) {
				case '<':
					program.addEmitString("&lt;");
					break;
				case '&':
					program.addEmitString("&amp;");
					break;
				case '>':
					program.addEmitString("&gt;");
					break;
				case 160:
					program.addEmitString("&nbsp;");
					break;
				default:
					program.addEmitChar(ch);
			}
		}
	}

	/**
	 * Special handling of &lt;style&gt;...&lt;/style&gt; content: prepend newline and remove
	 * leading spaces from each line. End with newline.
	 */
	protected void emitStyleTagContent(Node n) throws CompilerException
	{
		String s = n.getNodeValue().trim();
		s = options.treeModificator.modifyStyleElementContent(s);
		if (s == null || s.isEmpty())
			return;
		program.addEmitChar('\n');
		boolean afterLF = true;
		for (int i = 0; i < s.length(); i++) {
			char ch = s.charAt(i);
			if (ch == '\n') {
				afterLF = true;
			} else if (ch <= 32 && afterLF) {
				continue;
			} else {
				afterLF = false;
			}
			program.addEmitChar(ch);
		}
		program.addEmitChar('\n');
	}

	/**
	 * Special handling of &lt;script&gt;...&lt;/script&gt; content if it is not empty:
	 * prepend newline and append newline.
	 */
	protected void emitScriptTagContent(Node n) throws CompilerException
	{
		String s = n.getNodeValue();
		s = options.treeModificator.modifyScriptElementContent(s);
		if (s == null || s.isEmpty())
			return;
		if (s.indexOf('\n') < 0) {
			program.addEmitString(s);
			return;
		}
		program.addSoftNewline();
		program.addEmitString(s);
		program.addSoftNewline();
	}

	protected void emitCommentNode(Node n) throws CompilerException
	{
		String s = n.getNodeValue();
		s = options.treeModificator.modifyCommentContent(s);
		program.addEmitString("<!--");
		program.addEmitString(s);
		program.addEmitString("-->");
	}

}
