/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.utils;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;
import dk.vitality.lib.foundation.misc.ByteBuffer;

public class CompilerUtils
{
	public static final String[] emptyStringArray = new String[0];

	public static String[] splitCssClassNames(String attr)
	{
		if (attr == null || attr.isEmpty())
			return emptyStringArray;
		StringTokenizer st = new StringTokenizer(attr, " \t\n\r");
		String[] array = new String[st.countTokens()];
		for (int i = 0; st.hasMoreTokens(); i++)
			array[i] = st.nextToken();
		return array;
	}

	public static String[] splitCssClassNamesWithoutPrefixedUnderscore(String attr)
	{
		String[] a = splitCssClassNames(attr);
		for (int i = 0; i < a.length; i++) {
			String s = a[i];
			if (s.charAt(0) == '_')
				a[i] = s.substring(1);
		}
		return a;
	}

	public static String zapUnderscorePrefixedClassNames(String attr)
	{
		if (attr == null || attr.indexOf('_') < 0)
			return attr;
		String[] a = splitCssClassNames(attr);
		StringBuilder sb = new StringBuilder(attr.length());
		for (String s : a) {
			if (s.charAt(0) == '_')
				continue;
			if (sb.length() > 0)
				sb.append(' ');
			sb.append(s);
		}
		return sb.toString();
	}

	public static <T> List<T> concatLists(List<T> accum, List<T> additional)
	{
		if (additional == null || additional.isEmpty())
			return accum;
		if (accum == null)
			accum = new ArrayList<>();
		accum.addAll(additional);
		return accum;
	}

	public static String stripLowercasePrefix(String s) throws CompilerException
	{
		int len = s.length();
		for (int i = 0; i < len; i++) {
			if (!Character.isLowerCase(s.charAt(i)))
				return s.substring(i);
		}
		throw new InvalidElementTargetException("Can't create default name by stripping [a-z]+ from '" + s + "'");
	}

	public static String makeFirstCharacterUppercase(String s)
	{
		if (s == null || s.isEmpty() || Character.isUpperCase(s.charAt(0)))
			return s;
		else
			return Character.toUpperCase(s.charAt(0)) + s.substring(1);
	}

	public static String annotationToString(Annotation a, Class<?> cls, Field f, Method m)
	{
		StringBuilder sb = new StringBuilder(20);
		sb.append('@').append(a.annotationType().getSimpleName()).append(' ');
		if (f != null || m != null) {
			if (f != null)
				sb.append(f.getName());
			if (m != null) {
				sb.append(m.getReturnType().getSimpleName());
				sb.append(' ').append(m.getName()).append("()");
			}
			sb.append(" in ");
			addClass(sb, cls);
		} else {
			sb.append("class ");
			addClass(sb, cls);
		}
		return sb.toString();//StringUtils.replaceSubstring(sb.toString(), " java.lang.", " ");
	}

	public static void addClass(StringBuilder sb, Class<?> theClass)
	{
		Class<?> parent = theClass.getEnclosingClass();
		if (parent != null) {
			addClass(sb, parent);
			sb.append('.');
		}
		sb.append(theClass.getSimpleName());
	}


	public static String getInnerclassBasename(Class<?> innerClass)
	{
		return zapPrefix(getInnerclassName(innerClass), "Emit");
	}

	public static String getInnerclassName(Class<?> innerClass)
	{
		String s = innerClass.getSimpleName();
		int idx = s.lastIndexOf('$');
		if (idx >= 0)
			s = s.substring(idx + 1);
		return s;
	}

	public static boolean isValidJavaIdentifier(String s)
	{
		if (s.isEmpty())
			return false;
		if (!Character.isJavaIdentifierStart(s.charAt(0)))
			return false;
		for (int i = 1; i < s.length(); i++) {
			if (!Character.isJavaIdentifierPart(s.charAt(i)))
				return false;
		}
		return true;
	}

	public static String dotJoin(String... a)
	{
		if (a.length == 0)
			return null;
		if (a.length == 1)
			return a[0];
		int len = a.length - 1;
		for (String s : a)
			len += (s != null) ? s.length() : len;
		StringBuilder sb = new StringBuilder();
		for (String s : a) {
			if (s == null)
				continue;
			if (sb.length() > 0)
				sb.append('.');
			int idx1 = 0;
			while (idx1 < s.length() && s.charAt(idx1) == '.')
				idx1++;
			int idx2 = s.length();
			while (idx2 > 0 && s.charAt(idx2 - 1) == '.')
				idx2--;
			sb.append(s.substring(idx1, idx2));
		}
		return sb.toString();
	}

	public static String zapPrefix(String s, String prefix)
	{
		if (s == null)
			return null;
		if (prefix != null && !prefix.isEmpty() && s.startsWith(prefix))
			s = s.substring(prefix.length());
		return s;
	}

	public static String getSomeMessage(Throwable t)
	{
		if (t == null)
			return "(exception object is null)";
		String msg = t.getMessage();
		if (msg == null || msg.isEmpty())
			msg = t.getClass().getName();
		msg = zapPrefix(msg, "java.lang.Exception:");
		msg = zapPrefix(msg, "java.lang.");
		return msg;
	}

	public static String md5LcHex(String s) throws Exception
	{
		return ByteBuffer.toHexString(MessageDigest.getInstance("MD5").digest(s.getBytes(StandardCharsets.UTF_8)));
	}
}
