/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.html;

import java.net.URLEncoder;
import dk.vitality.lib.foundation.exception.ExceptionUtils;

public class URLUtils
{
	public static String decodeURL(String s, boolean decodeUtf8)
	{
		if (s == null)
			return null;
		StringBuilder sb = new StringBuilder();
		boolean hasDouble8 = false, has16 = false, prevWas8 = false;
		boolean hasSimple8 = false;
		int len = s.length();
		for (int i = 0; i < len; ++i) {
			char ch = s.charAt(i);
			if (ch == '+') {
				sb.append(' ');
				prevWas8 = false;
			} else if (ch == '%' && i + 5 < len && s.charAt(i + 1) == 'u') {
				int i1 = dehex(s.charAt(i + 2)) << 12;
				int i2 = dehex(s.charAt(i + 3)) << 8;
				int i3 = dehex(s.charAt(i + 4)) << 4;
				int i4 = dehex(s.charAt(i + 5));
				if (i1 >= 0 && i2 >= 0 && i3 >= 0 && i4 >= 0) {
					sb.append((char)(i1 | i2 | i3 | i4));
					i += 5;
					has16 = true;
				} else {
					sb.append(ch);
				}
				prevWas8 = false;
			} else if (ch == '%' && i + 2 < len) {
				int i1 = dehex(s.charAt(i + 1)) << 4;
				int i2 = dehex(s.charAt(i + 2));
				if (i1 >= 0 && i2 >= 0) {
					ch = (char)(i1 | i2);
					if (ch == 10 && sb.length() > 0 && sb.charAt(sb.length() - 1) == 13)
						sb.setLength(sb.length() - 1); // cr lf -> lf
					sb.append(ch);
					hasSimple8 |= ch >= 128;
					i += 2;
					if (ch >= 128) {
						hasDouble8 |= prevWas8;
						prevWas8 = true;
					} else {
						prevWas8 = false;
					}
				} else {
					sb.append(ch);
					prevWas8 = false;
				}
			} else {
				sb.append(ch);
				prevWas8 = false;
			}
		}
		final boolean debugFlag = false;
		if (decodeUtf8 && hasDouble8 && !has16) {
			try {
				len = sb.length();
				byte[] b = new byte[len];
				for (int i = 0; i < len; i++)
					b[i] = (byte)sb.charAt(i);
				String result = new String(b, "UTF-8");
				if (debugFlag)
					System.err.println("Did decode til UTF-8: '" + s + "' -> '" + result + "'");
				return result;
			}
			catch (Exception ex) {
				if (debugFlag)
					System.err.println("Can't decode as UTF-8: '" + sb + "': " + ExceptionUtils.getExMsg(ex));
			}
		} else if (debugFlag && hasSimple8) {
			System.err.println("NON UTF-8: '" + s + "' -> '" + sb + "'");
		}
		return sb.toString();
	}

	// Compared to java.net.URLEncoder.encode:
	// It doesn't encode with the platform encoding, but always UTF-8.
	public static String encodeURL(String s)
	{
		if (s == null)
			return "";
		if (!needUrlEncode(s))
			return s;
		int len = s.length();
		StringBuilder sb = new StringBuilder(len + 25);
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch == '_' || ch == '-' || ch == '.'
					|| (ch >= '0' && ch <= '9')
					|| (ch >= 'A' && ch <= 'Z')
					|| (ch >= 'a' && ch <= 'z')) {
				sb.append(ch);
			} else if (ch <= 127) {
				sb.append('%');
				sb.append(tohex(ch >> 4));
				sb.append(tohex(ch));
			} else if (ch <= 0x7FF) {
				sb.append('%');
				sb.append(tohex(((ch >> 10) & 1) | 12));
				sb.append(tohex(ch >> 6));
				sb.append('%');
				sb.append(tohex(((ch >> 4) & 3) | 8));
				sb.append(tohex(ch));
			} else if (ch <= 0xFFFF) {
				sb.append('%');
				sb.append(tohex(14));
				sb.append(tohex(ch >> 12));
				sb.append('%');
				sb.append(tohex(((ch >> 10) & 3) | 8));
				sb.append(tohex(ch >> 6));
				sb.append('%');
				sb.append(tohex(((ch >> 4) & 3) | 8));
				sb.append(tohex(ch));
			} else {
				try {
					sb.append(URLEncoder.encode(s.substring(i, i + 1), "UTF-8"));
				}
				catch (Exception ex) {
					System.err.println("URLEncoder charset: " + ExceptionUtils.getExMsg(ex));
				}
			}
		}
		//System.err.println("URL encode '" + s + "' -> '" + sb + "'");
		return sb.toString();
	}

	private static boolean needUrlEncode(String s)
	{
		int len = s.length();
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch == '_' || ch == '-' || ch == '.'
					|| (ch >= '0' && ch <= '9')
					|| (ch >= 'A' && ch <= 'Z')
					|| (ch >= 'a' && ch <= 'z')) {
				continue;
			}
			return true;
		}
		return false;
	}

	private static char tohex(int i)
	{
		i &= 15;
		return i < 10 ? (char)('0' + i) : (char)('A' + i - 10);
	}

	private static int dehex(char c)
	{
		if (c >= '0' && c <= '9')
			return c - '0';
		if (c >= 'a' && c <= 'f')
			return c - 'a' + 10;
		if (c >= 'A' && c <= 'F')
			return c - 'A' + 10;
		return -1;
	}
}
