/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.io;

import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.util.zip.GZIPInputStream;
import org.jetbrains.annotations.NotNull;
import dk.vitality.lib.foundation.misc.ByteBuffer;

public class FileContentsUtils
{
	public static void readFully(@NotNull InputStream is,
	                             @NotNull byte[] buffer) throws IOException
	{
		readFully(is, buffer, 0, buffer.length);
	}

	/**
	 * Reads from the InputStream until bytesToRead bytes have been read.
	 *
	 * @throws IOException Throws EOFException if less than bytesToRead bytes can be read before getting EOF from the stream.
	 */
	public static void readFully(@NotNull InputStream is,
	                             @NotNull byte[] buffer,
	                             int offset,
	                             int bytesToRead) throws IOException
	{
		int bytesRead = 0;
		int dst = offset;
		while (bytesRead < bytesToRead) {
			int num = is.read(buffer, dst, bytesToRead - bytesRead);
			//System.err.println("read " + num + " -> " + (numBytesRead + num) + " of " + maxToRead);
			if (num < 0)
				throw new EOFException("End of stream reached prematurely.");
			bytesRead += num;
			dst += num;
		}
	}

	@NotNull
	public static byte[] readFully(@NotNull InputStream is) throws IOException
	{
		ByteBuffer bb = new ByteBuffer(64);
		bb.readFully(is);
		return bb.toBytes();
	}

	/**
	 * Gets bytes from /dev/random
	 */
	public static byte[] getDevRandomBytes(int num) throws IOException
	{
		try (FileInputStream fis = new FileInputStream("/dev/random")) {
			byte[] buffer = new byte[num];
			readFully(fis, buffer);
			return buffer;
		}
	}

	/**
	 * Saves the content into the file by writing to a temporary file and then renaming the file.
	 */
	public static void saveFile(@NotNull File f,
	                            @NotNull byte[] data) throws IOException
	{
		File tmpF = new File(f.getParentFile(), f.getName() + "-" + System.currentTimeMillis() + ".tmp");
		try {
			Files.write(tmpF.toPath(), data);
		}
		catch (IOException ex) {
			if (!tmpF.delete())
				System.err.println("Couldn't delete tmp file: " + tmpF.getPath());
			throw ex;
		}
		// Must delete file and not just overwrite it using renameTo() as
		// IDEA doesn't detect filechanges otherwise...
		if (f.exists() && !f.delete())
			throw new IOException("Can't delete " + f.getPath());
		if (!tmpF.renameTo(f)) {
			if (!tmpF.delete())
				System.err.println("Couldn't delete tmp file: " + tmpF.getPath());
			throw new IOException("Can't rename " + tmpF.getPath() + " to " + f.getPath());
		}
	}

	/**
	 * Saves the content into the file by writing to a temporary file and then renaming the file.
	 */
	public static void saveFile(@NotNull File f,
	                            @NotNull InputStream is) throws IOException
	{
		File tmpF = new File(f.getParentFile(), f.getName() + "-" + System.currentTimeMillis() + ".tmp");
		try {
			Files.copy(is, tmpF.toPath());
		}
		catch (IOException ex) {
			if (!tmpF.delete())
				System.err.println("Couldn't delete tmp file: " + tmpF.getPath());
			throw ex;
		}
		// Must delete file and not just overwrite it using renameTo() as
		// IDEA doesn't detect filechanges otherwise...
		if (f.exists() && !f.delete())
			throw new IOException("Can't delete " + f.getPath());
		if (!tmpF.renameTo(f)) {
			if (!tmpF.delete())
				System.err.println("Couldn't delete tmp file: " + tmpF.getPath());
			throw new IOException("Can't rename " + tmpF.getPath() + " to " + f.getPath());
		}
	}

	/**
	 * Appends the content to the file.
	 */
	public static void appendToFile(@NotNull File f,
	                                @NotNull byte[] data) throws IOException
	{
		try (FileOutputStream fos = new FileOutputStream(f, true)) {
			fos.write(data);
		}
	}

	/**
	 * @return the complete file contents.
	 *         If the filename ends with ".gz", the contents is uncompressed.
	 */
	public static byte[] readFile(@NotNull File f) throws IOException
	{
		if (!f.isFile())
			throw new FileNotFoundException("Missing file: " + f.getPath());
		if (f.getPath().endsWith(".gz") || f.getPath().endsWith(".GZ"))
			return readGzippedFile(f);
		return Files.readAllBytes(f.toPath());
	}

	/**
	 * @param f File <i>must</i> be gzipped.
	 * @return The uncompressed content.
	 */
	public static byte[] readGzippedFile(@NotNull File f) throws IOException
	{
		if (!f.isFile())
			throw new IOException("Missing file: " + f.getPath());
		long size = f.length();
		if (size > 1024L * 1024L * 1024L)
			throw new IOException("File too big to read into memory: " + size + " bytes");
		try (GZIPInputStream is = new GZIPInputStream(new FileInputStream(f))) {
			return readFully(is);
		}
	}
}
