/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.misc;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public final class ByteBuffer
{
	private byte array[];
	private int count;

	public ByteBuffer(int capacity)
	{
		array = new byte[capacity];
	}

	public int length()
	{
		return count;
	}

	public int capacity()
	{
		return array.length;
	}

	public void ensureCapacity(int minimumCapacity)
	{
		if (minimumCapacity <= array.length)
			return;
		int newCapacity = Math.max(minimumCapacity, (count + 1) * 2);
		byte tmp[] = new byte[newCapacity];
		System.arraycopy(array, 0, tmp, 0, count);
		array = tmp;
	}

	public void ensureLastLineEndsWithNewline()
	{
		if (count == 0 || array[count - 1] == '\n')
			return;
		append('\n');
	}

	public void trimToSize()
	{
		if (count == array.length)
			return;
		byte[] tmp = new byte[count];
		System.arraycopy(array, 0, tmp, 0, count);
		this.array = tmp;
	}

	public void setLength(int newLength)
	{
		if (newLength < 0)
			throw new ArrayIndexOutOfBoundsException(newLength);
		if (newLength > array.length)
			ensureCapacity(newLength);
		if (count < newLength) {
			for (; count < newLength; count++)
				array[count] = 0;
		} else {
			count = newLength;
		}
	}

	public byte byteAt(int index)
	{
		if ((index < 0) || (index >= count))
			throw new ArrayIndexOutOfBoundsException(index);
		return array[index];
	}

	public byte lastByte()
	{
		if (count == 0)
			throw new ArrayIndexOutOfBoundsException(0);
		return array[count - 1];
	}

	public void writeTo(@NotNull OutputStream os) throws IOException
	{
		if (count > 0)
			os.write(array, 0, count);
	}

	public void readFully(@NotNull InputStream is) throws IOException
	{
		while (true) {
			int x = is.read();
			if (x < 0)
				break;
			if (count >= array.length)
				ensureCapacity(count + 1);
			array[count++] = (byte)x;
		}
	}

	public void getBytes(int srcBegin,
	                     int srcEnd,
	                     @NotNull byte dst[],
	                     int dstBegin)
	{
		if (srcBegin < 0)
			throw new ArrayIndexOutOfBoundsException(srcBegin);
		if ((srcEnd < 0) || (srcEnd > count))
			throw new ArrayIndexOutOfBoundsException(srcEnd);
		if (srcBegin > srcEnd)
			throw new ArrayIndexOutOfBoundsException("srcBegin > srcEnd");
		System.arraycopy(array, srcBegin, dst, dstBegin, srcEnd - srcBegin);
	}

	public void setByteAt(int index, int b)
	{
		if ((index < 0) || (index >= count))
			throw new ArrayIndexOutOfBoundsException(index);
		array[index] = (byte)b;
	}

	public ByteBuffer append(@NotNull byte data[])
	{
		ensureCapacity(count + data.length);
		System.arraycopy(data, 0, array, count, data.length);
		count += data.length;
		return this;
	}

	public ByteBuffer append(@NotNull byte data[], int offset, int len)
	{
		ensureCapacity(count + len);
		System.arraycopy(data, offset, array, count, len);
		count += len;
		return this;
	}

	@NotNull
	public ByteBuffer appendLatin1(@NotNull String s)
	{
		if (s == null)
			return this;
		int len = s.length();
		ensureCapacity(count + len);
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch > 255)
				ch = '?';
			array[count++] = (byte)ch;
		}
		return this;
	}

	@NotNull
	public ByteBuffer appendUtf8(@NotNull String s)
	{
		if (s == null)
			return this;
		return appendUtf8(s, 0, s.length());
	}

	@NotNull
	public ByteBuffer appendUtf8(@NotNull String s, int offset, int len)
	{
		if (s == null || len <= 0)
			return this;
		int minSize = count;
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(offset + i);
			if (ch < 0x80)
				minSize++;
			else if (ch < 0x0800)
				minSize += 2;
			else if (ch < 0x10000)
				minSize += 3;
			else
				minSize += 4;
		}
		ensureCapacity(minSize);
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(offset + i);
			if (ch < 0x80) {
				array[count++] = (byte)ch;
			} else if (ch < 0x0800) {
				array[count++] = (byte)(ch >> 6 | 0xC0);
				array[count++] = (byte)(ch & 0x3F | 0x80);
			} else if (ch < 0x10000) {
				array[count++] = (byte)(ch >> 12 | 0xE0);
				array[count++] = (byte)(ch >> 6 & 0x3F | 0x80);
				array[count++] = (byte)(ch & 0x3F | 0x80);
			} else {
				array[count++] = (byte)(ch >> 18 | 0xF0);
				array[count++] = (byte)(ch >> 12 & 0x3F | 0x80);
				array[count++] = (byte)(ch >> 6 & 0x3F | 0x80);
				array[count++] = (byte)(ch & 0x3F | 0x80);
			}
		}
		return this;
	}

	@NotNull
	public ByteBuffer append(int b)
	{
		if (count == array.length)
			ensureCapacity(count + 1);
		array[count++] = (byte)b;
		return this;
	}

	@NotNull
	public byte[] toBytes()
	{
		if (count == array.length)
			return array;
		byte[] b = new byte[count];
		System.arraycopy(array, 0, b, 0, count);
		return b;
	}

	@NotNull
	public String toString(@NotNull String encoding) throws UnsupportedEncodingException
	{
		return new String(array, 0, count, encoding);
	}

	@NotNull
	public String toString(int start, int end, @NotNull String encoding) throws UnsupportedEncodingException
	{
		return new String(array, start, end - start, encoding);
	}

	public void clear()
	{
		count = 0;
	}

	@NotNull
	public String toDebugString()
	{
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < count; i++) {
			if (sb.length() > 0)
				sb.append(", ");
			int ch = 255 & array[i];
			sb.append(ch);
			if (ch > 32 && ch <= 255) {
				sb.append("=");
				sb.append((char)ch);
			}
		}
		return sb.toString();
	}

	@NotNull
	public static String toUnsignedString(@Nullable byte[] b)
	{
		if (b == null)
			return "null";
		if (b.length == 0)
			return "[]";
		StringBuilder sb = new StringBuilder(3 + 3 * b.length);
		sb.append("[");
		for (int i = 0; i < b.length; i++) {
			if (i > 0)
				sb.append(", ");
			sb.append(255 & b[i]);
		}
		sb.append("]");
		return sb.toString();
	}

	public static void toUnsignedString(@NotNull StringBuilder sb, @Nullable byte[] b)
	{
		if (b == null) {
			sb.append("null");
			return;
		}
		sb.append("[");
		for (int i = 0; i < b.length; i++) {
			if (i > 0)
				sb.append(", ");
			sb.append(255 & b[i]);
		}
		sb.append("]");
	}

	@NotNull
	public static String toHexString(@Nullable byte[] bytes)
	{
		if (bytes == null || bytes.length == 0)
			return "";
		StringBuilder sb = new StringBuilder(bytes.length * 2);
		toHexString(sb, bytes, 0, bytes.length);
		return sb.toString();
	}

	public static void toHexString(@NotNull StringBuilder sb, @Nullable byte[] bytes)
	{
		toHexString(sb, bytes, 0, bytes.length);
	}

	public static void toHexString(@NotNull StringBuilder sb, @Nullable byte[] bytes, int offset, int len)
	{
		if (bytes == null || bytes.length == 0) {
			sb.append("");
			return;
		}
		while (len > 0) {
			len--;
			int b = bytes[offset++];
			sb.append(lcHexChars[(b >>> 4) & 15]);
			sb.append(lcHexChars[b & 15]);
		}
	}

	public static String toVerboseHexString(@Nullable byte[] bytes)
	{
		if (bytes == null)
			return "null";
		StringBuilder sb = new StringBuilder(bytes.length * 3 + 2);
		toVerboseHexString(sb, bytes, 0, bytes.length);
		return sb.toString();
	}

	public static void toVerboseHexString(@NotNull StringBuilder sb, @Nullable byte[] bytes)
	{
		toVerboseHexString(sb, bytes, 0, bytes.length);
	}

	public static void toVerboseHexString(@NotNull StringBuilder sb, @Nullable byte[] bytes, int offset, int len)
	{
		if (bytes == null) {
			sb.append("null");
			return;
		}
		sb.append("[");
		for (int i = 0; i < len; i++) {
			if (i > 0)
				sb.append(", ");
			byte b = bytes[offset++];
			sb.append(lcHexChars[(b >>> 4) & 15]);
			sb.append(lcHexChars[b & 15]);
		}
		sb.append("]");
	}

	private static final char[] lcHexChars = "0123456789abcdef".toCharArray();
}
