/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.misc;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.Semaphore;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public class KeyedMutexExecutor
{
	public void execute(@NotNull String key,
						@NotNull RunnableX exec) throws Exception
	{
		get(key).query(exec, null);
	}

	public <T> T query(@NotNull String key,
					   @NotNull Callable<T> exec) throws Exception
	{
		return get(key).query(null, exec);
	}

	@NotNull public final String mutexName;
	public boolean debugFlag;

	public KeyedMutexExecutor(@NotNull String mutexName)
	{
		this.mutexName = mutexName;
	}

	public KeyedMutexExecutor(@NotNull String mutexName, boolean debugFlag)
	{
		this.mutexName = mutexName;
		this.debugFlag = debugFlag;
	}

	public List<String> getActiveKeys()
	{
		List<String> list;
		synchronized (cache) {
			list = new ArrayList<>(cache.keySet());
		}
		return Collections.unmodifiableList(list);
	}

	private Entry get(@NotNull String key)
	{
		synchronized (cache) {
			Entry entry = cache.get(key);
			if (entry == null) {
				entry = new Entry(key);
				cache.put(key, entry);
			} else {
				entry.count++;
			}
			return entry;
		}
	}

	private class Entry
	{
		@NotNull final String key;
		@NotNull private final Semaphore semaphore;
		private volatile int count;
		private volatile Thread owner;

		private Entry(@NotNull String key)
		{
			this.key = key;
			this.count = 1;
			this.semaphore = new Semaphore(1, true);
		}

		public <T> T query(@Nullable RunnableX exec1,
						   @Nullable Callable<T> exec2) throws Exception
		{
			try {
				boolean nested = Thread.currentThread() == owner;
				if (!nested)
					acquireSemaphore();
				else if (debugFlag)
					logLn("Nested acquire of " + key);
				long st0 = System.nanoTime();
				try {
					if (exec2 != null)
						return exec2.call();
					exec1.run();
					return null;
				}
				finally {
					if (!nested)
						releaseSemaphore(st0);
					else if (debugFlag)
						logLn("Nested release of " + key);
				}
			}
			finally {
				releaseEntry();
			}
		}

		private void acquireSemaphore() throws InterruptedException
		{
			if (!semaphore.tryAcquire()) {
				if (debugFlag)
					logLn("Waiting for " + key);
				semaphore.acquire();
			}
			owner = Thread.currentThread();
			if (debugFlag)
				logLn("Got " + key);
		}

		private void releaseSemaphore(long st0) throws InterruptedException
		{
			if (debugFlag) {
				long usecs = (System.nanoTime() - st0) / 1000;
				logLn("Releases " + key + "  (" + usecs / 1000.0 + " msecs)");
			}
			owner = null;
			semaphore.release();
		}

		private void releaseEntry()
		{
			synchronized (cache) {
				if (--count == 0)
					cache.remove(key);
			}
		}
	}

	private void logLn(String s)
	{
		System.err.println("KeyedMutexExecutor." + mutexName + ": " + s);
	}

	private final Map<String, Entry> cache = new HashMap<>();
}
