/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <time.h>
#import <SystemConfiguration/SCPreferences.h>
#import <SystemConfiguration/SCNetworkConfiguration.h>

static CFStringRef bundleIdentifier = CFSTR("org.glimmerblocker.proxy.networkConfigurationListener");

/*
static void FailedAssert(const char *msg)
{
	printf("Failed Assert: %s\n", msg);
	exit(1);
}
#undef NSAssert
#define NSAssert(x, msg) if (!(x)) FailedAssert(msg)
*/

static NSMutableString* mstr;
static CFMutableStringRef cfmstr;
static int needComma = 0;
static int indent = 0;

static void printIndent()
{
	for (int i = 0; i < indent; i++)
		CFStringAppend(cfmstr, CFSTR("   "));
}

static void enterDict()
{
	CFStringAppend(cfmstr, CFSTR("{"));
	needComma <<= 1;
	indent++;
}

static void printItemIndent()
{
	if (needComma & 1)
		CFStringAppend(cfmstr, CFSTR(",\n"));
	else
		CFStringAppend(cfmstr, CFSTR("\n")); // after '{'
	needComma |= 1;
	printIndent();
}

static void leaveDict()
{
	int hasItems = needComma & 1;
	if (hasItems)
		CFStringAppend(cfmstr, CFSTR("\n"));
	indent--;
	needComma >>= 1;
	if (hasItems)
		printIndent();
	CFStringAppend(cfmstr, CFSTR("}"));
}

static void appendValue(const CFTypeRef value)
{
	if (!value) {
		CFStringAppend(cfmstr, CFSTR("null"));
		return;
	}
	CFTypeID type = CFGetTypeID(value);
	if (type == CFStringGetTypeID()) {
		CFStringAppend(cfmstr, CFSTR("'"));
		CFStringRef s = (CFStringRef)value;
		CFIndex len = CFStringGetLength(s);
		for (CFIndex i = 0; i < len; i++) {
			UniChar ch = CFStringGetCharacterAtIndex(s, i);
			if (ch == '\'') {
				[mstr appendString:@"\\'"];
				continue;
			}
			if (ch == '\\') {
				[mstr appendString:@"\\\\"];
				continue;
			}
			if (ch < 32) {
				[mstr appendFormat:@"\\x%02x", (int)ch];
				continue;
			}
			CFStringAppendCharacters(cfmstr, &ch, 1);
		}
		CFStringAppend(cfmstr, CFSTR("'"));
		return;
	}
	if (type == CFArrayGetTypeID()) {
		CFStringAppend(cfmstr, CFSTR("["));
		CFArrayRef a = (CFArrayRef)value;
		CFIndex num = CFArrayGetCount(a);
		for (CFIndex i = 0; i < num; i++) {
			if (i > 0)
				CFStringAppend(cfmstr, CFSTR(", "));
			appendValue(CFArrayGetValueAtIndex(a, i));
		}
		CFStringAppend(cfmstr, CFSTR("]"));
		return;
	}
	if (type == CFNumberGetTypeID()) {
		[mstr appendString:[(__bridge NSNumber*)value stringValue]];
		return;
	}
	if (type == CFBooleanGetTypeID()) {
		NSNumber* n = (__bridge NSNumber*)value;
		NSString* s = [n intValue] ? @"true" : @"false";
		[mstr appendString:s];
		return;
	}
	if (type == CFDataGetTypeID()) {
		//CFStringAppend(cfmstr, CFSTR("/* CFData */"));
		CFStringAppend(cfmstr, CFSTR("'0x"));
		CFDataRef a = (CFDataRef)value;
		CFIndex len = CFDataGetLength(a);
		for (CFIndex i = 0; i < len; i++)
			[mstr appendFormat:@"%02x", 255 & *(CFDataGetBytePtr(a) + i)];
		CFStringAppend(cfmstr, CFSTR("'"));
		return;
	}
	CFStringAppend(cfmstr, CFSTR("null"));
	CFStringRef s = CFCopyTypeIDDescription(type);
	printf("MSG: unhandled value type = %ld = %s\n", type, [(__bridge NSString*)s UTF8String]);
	CFRelease(s);
}

static void printLocationName()
{
	CFStringRef appName = CFSTR("GlimmerBlocker");
	SCPreferencesRef prefsR = SCPreferencesCreate(NULL, appName, NULL);
	if (!prefsR) {
		printf("MSG: SCPreferencesCreate returned NULL.");
		return;
	}
	SCNetworkSetRef networkSetR = SCNetworkSetCopyCurrent(prefsR);
	if (!networkSetR) {
		printf("MSG: SCNetworkSetCopyCurrent returned NULL.");
	} else {
		CFStringRef name = SCNetworkSetGetName(networkSetR);
		appendValue(name);
		CFRelease(networkSetR);
	}
	CFRelease(prefsR);
}


static void appendInterfaceData(SCDynamicStoreRef store, CFStringRef interfaceName, CFStringRef format,
								CFStringRef dictName, ...)
{
	CFStringRef path = CFStringCreateWithFormat(kCFAllocatorDefault, NULL, format, interfaceName);
	CFPropertyListRef propList = SCDynamicStoreCopyValue(store, path);
	CFRelease(path);
	path = NULL;
	if (!propList) {
		//printf("MSG: SCDynamicStoreCopyValue(%s) returned NULL.\n", [(NSString*)path UTF8String]);
		return;
	}
	printItemIndent();
	CFStringAppend(cfmstr, dictName);
	CFStringAppend(cfmstr, CFSTR(": "));
	enterDict();
	va_list arglist;
	va_start(arglist, dictName);
	CFStringRef attrName;
	while ((attrName = va_arg(arglist, CFStringRef)) != NULL) {
		CFStringRef jsName = va_arg(arglist, CFStringRef);
		const CFTypeRef value = CFDictionaryGetValue(propList, attrName);
		if (!value)
			continue;
		printItemIndent();
		CFStringAppend(cfmstr, jsName);
		CFStringAppend(cfmstr, CFSTR(": "));
		appendValue(value);
	}
	leaveDict();
	CFRelease(propList);
	propList = NULL;
	va_end(arglist);
}

static void appendInterfaceLinkStaus(SCDynamicStoreRef store, CFStringRef interfaceName)
{
	CFStringRef path = CFStringCreateWithFormat(kCFAllocatorDefault, NULL, CFSTR("State:/Network/Interface/%@/Link"), interfaceName);
	CFPropertyListRef propList = SCDynamicStoreCopyValue(store, path);
	CFRelease(path);
	path = NULL;
	if (!propList)
		return;
	CFTypeRef value = CFDictionaryGetValue(propList, CFSTR("Active"));
	if (value) {
		CFTypeID type = CFGetTypeID(value);
		if (type == CFBooleanGetTypeID()) {
			NSNumber* n = (__bridge NSNumber*)value;
			if (![n intValue])
				value = NULL; // don't print false, omit it to make json more terse
		}
	}
	if (value) {
		printItemIndent();
		CFStringAppend(cfmstr, CFSTR("link: "));
		appendValue(value);
	}
	CFRelease(propList);
	propList = NULL;
}

static void printOneInterface(SCDynamicStoreRef store, CFStringRef interfaceName, int pass)
{
	BOOL isLocalhost = CFStringGetLength(interfaceName) >= 3
		&& CFStringGetCharacterAtIndex(interfaceName, 0) == 'l'
		&& CFStringGetCharacterAtIndex(interfaceName, 1) == 'o'
		&& CFStringGetCharacterAtIndex(interfaceName, 2) <= '9';
	if (isLocalhost != (pass == 2)) {
		// print localhost interfaces at the bottom, as they're just too boring to be at the top.
		return;
	}
	printItemIndent();
	UniChar ch = CFStringGetCharacterAtIndex(interfaceName, 0);
	if (ch >= 'a' && ch <= 'z')
		CFStringAppend(cfmstr, interfaceName);
	else
		appendValue(interfaceName);
	CFStringAppend(cfmstr, CFSTR(": "));
	enterDict();
	appendInterfaceLinkStaus(store, interfaceName);
	appendInterfaceData(store, interfaceName, CFSTR("State:/Network/Interface/%@/AirPort"),
						CFSTR("AirPort"),
						CFSTR("Power Status"), CFSTR("power"),
						CFSTR("SSID_STR"), CFSTR("ssid"),
						CFSTR("SSID"), CFSTR("ssid_hex"),
						NULL);
	appendInterfaceData(store, interfaceName, CFSTR("State:/Network/Interface/%@/IPv4"),
						CFSTR("IPv4"),
						CFSTR("Addresses"), CFSTR("addr"),
						CFSTR("SubnetMasks"), CFSTR("mask"),
						CFSTR("BroadcastAddresses"), CFSTR("broadcast"),
						NULL);
	appendInterfaceData(store, interfaceName, CFSTR("State:/Network/Interface/%@/IPv6"),
						CFSTR("IPv6"),
						CFSTR("Addresses"), CFSTR("addr"),
						CFSTR("Flags"), CFSTR("flags"),
						CFSTR("PrefixLength"), CFSTR("prefixlen"),
						NULL);
	//  en1: { ssid:'Funny', ipv4: [ "127.0.0.1" ], ipv4mask: [ "255.255.255.0" ], ipv6: ["fe80::1"] }
	leaveDict();
}

static void printInterfaces()
{
	// http://developer.apple.com/documentation/Networking/Conceptual/SystemConfigFrameworks/SC_UnderstandSchema/SC_UnderstandSchema.html#//apple_ref/doc/uid/TP40001065-CH203-CHDFGAEF
	SCDynamicStoreContext context = { 0, NULL, NULL, NULL, NULL };
	CFPropertyListRef propList = NULL;
	SCDynamicStoreRef store = SCDynamicStoreCreate(NULL, bundleIdentifier, nil, &context);
	if (!store) {
		printf("MSG: SCDynamicStoreCreate returned NULL.");
		goto release;
	}
	propList = SCDynamicStoreCopyValue(store, CFSTR("State:/Network/Interface"));
	if (!propList) {
		printf("MSG: SCDynamicStoreCopyValue(State:/Network/Interface) returned NULL.");
		goto release;
	}
	CFArrayRef interfaceList = CFDictionaryGetValue(propList, CFSTR("Interfaces"));
	if (!interfaceList) {
		printf("MSG: CFDictionaryGetValue(State:/Network/Interface@Interfaces) returned NULL.");
		goto release;
	}
	CFIndex numInterfaces = CFArrayGetCount(interfaceList);
	//printf("DEBUG: got %d interfaces\n", numInterfaces);
	enterDict();
	for (int pass = 1; pass <= 2; pass++) {
		for (CFIndex interfaceIdx = 0; interfaceIdx < numInterfaces; interfaceIdx++) {
			CFStringRef interfaceName = CFArrayGetValueAtIndex(interfaceList, interfaceIdx);
			if (!interfaceName) {
				printf("DEBUG: interface[%ld] has no name?\n", interfaceIdx);
				continue;
			}
			//printf("DEBUG: interface = %s\n", [(NSString*)interfaceName UTF8String]);
			printOneInterface(store, interfaceName, pass);
		}
	}
	leaveDict();
release:
	if (store)
		CFRelease(store);
	if (propList)
		CFRelease(propList);
}

static void appendSoftNewline()
{
	if (![mstr hasSuffix:@"\n"])
		[mstr appendString:@"\n"];
}

static void printState(BOOL forDisplay)
{
	// reset globals:
	needComma = 0;
	indent = 0;
	if (!mstr) {
		mstr = [NSMutableString  stringWithCapacity:200];
		cfmstr = (__bridge CFMutableStringRef)mstr;
	}
	[mstr deleteCharactersInRange:NSMakeRange(0, [mstr length])];
	//
	if (forDisplay) {
		[mstr appendString:@"gb.network = "];
		enterDict();
	}
	printItemIndent();
	[mstr appendString:@"location: "];
	printLocationName();
	//
	printItemIndent();
	[mstr appendString:@"interfaces: "];
	printInterfaces();
	//
	if (forDisplay)
		leaveDict();
	//
	appendSoftNewline();
	if (!forDisplay) {
		time_t t = time(NULL);
		[mstr appendFormat:@"DEBUG: %s", ctime(&t)];
	}
	//
	fputs([mstr UTF8String], stdout);
	fflush(stdout);
}

static void scCallback(SCDynamicStoreRef dynStore, CFArrayRef changedKeys, void *info)
{
	CFIndex num = CFArrayGetCount(changedKeys);
	for (CFIndex i = 0; i < num; i++)
		printf("DEBUG: State changed[%ld]: %s\n", i, [(NSString*)CFArrayGetValueAtIndex(changedKeys, i) UTF8String]);
	// sometimes the callback is a wee bit faster than AirPort goes offline when selecting
	// "Turn AirPort Off" from the menubar,
	// so waiting a bit is sometimes required.
	sleep(1);
	printState(NO);
}

static void sourceCallBack(void *info)
{
	// dummy for CFRunLoop so it doesn't exit.  Never called.
}

static void addListener()
{
	SCDynamicStoreContext context = {
		.version			= 0,
		.info				= NULL,
		.retain				= NULL,
		.release			= NULL,
		.copyDescription	= NULL
	};
	SCDynamicStoreRef store = SCDynamicStoreCreate(NULL, bundleIdentifier, scCallback, &context);
	CFRunLoopSourceRef runLoopSrc = SCDynamicStoreCreateRunLoopSource(NULL, store, 0);
	CFRunLoopAddSource([[NSRunLoop currentRunLoop] getCFRunLoop], runLoopSrc, kCFRunLoopCommonModes);
	CFRelease(runLoopSrc);
	//
	CFMutableArrayRef fixedKeys = CFArrayCreateMutable(NULL, 0, NULL);
	//
	CFStringRef locationKey = SCDynamicStoreKeyCreateLocation(NULL);
	//NSLog(@"SCDynamicStoreKeyCreateLocation: %@", key);
	CFArrayAppendValue(fixedKeys, locationKey);
	CFRelease(locationKey);
	CFArrayAppendValue(fixedKeys, @"State:/Network/Interface");
	//
	CFMutableArrayRef regexpKeys = CFArrayCreateMutable(NULL, 0, NULL);
	CFArrayAppendValue(regexpKeys, @"State:/Network/Interface/.*");
	//
	SCDynamicStoreSetNotificationKeys(store, fixedKeys, regexpKeys);
	CFRelease(fixedKeys);
	CFRelease(regexpKeys);
}

int main (int argc, const char * argv[])
{
	@autoreleasepool {
		if (argc < 2 || strcmp(argv[1], "-continous")) {
			printState(YES);
			exit(0);
		}
		// Create a a sourceContext to be used by our source that makes
		// sure the CFRunLoop doesn't exit right away
		CFRunLoopSourceContext sourceContext;
		sourceContext.version = 0;
		sourceContext.info = NULL;
		sourceContext.retain = NULL;
		sourceContext.release = NULL;
		sourceContext.copyDescription = NULL;
		sourceContext.equal = NULL;
		sourceContext.hash = NULL;
		sourceContext.schedule = NULL;
		sourceContext.cancel = NULL;
		sourceContext.perform = &sourceCallBack;
		//
		CFRunLoopSourceRef sourceRef = CFRunLoopSourceCreate(NULL, 0, &sourceContext);
		CFRunLoopAddSource(CFRunLoopGetCurrent(), sourceRef, kCFRunLoopCommonModes);
		addListener();
		CFRelease(sourceRef);
		printState(NO);
		CFRunLoopRun();
		printf("main() done.\n");
	}
	return 0;
}

