/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "NetworkSettingsUpdater.h"
#import "GBDebug.h"

/*
 Getting settings:
 http://developer.apple.com/documentation/Networking/Reference/SysConfig/SCNetworkConfiguration/index.html

 Committing changes:
 http://developer.apple.com/documentation/Networking/Reference/SysConfig/SCPreferences/CompositePage.html
*/

@interface NetworkSettingsUpdater()

@property int proxyPort;
@property BOOL wantEnable;

- (BOOL)updateNetworkService:(SCNetworkServiceRef)serviceR; // returns YES if needs update.

@end

@implementation NetworkSettingsUpdater

BOOL cfstreql(CFStringRef a, CFStringRef b);
BOOL cfstreql(CFStringRef a, CFStringRef b)
{
	if ((a != NULL) != (b != NULL))
		return NO;
	return !a || !CFStringCompare(a, b, 0);
}

- (instancetype)initWithProxyPort:(int)proxyPort
		 withWantEnable:(BOOL)wantEnable
{
	if ((self = [super init])) {
		_proxyPort = proxyPort;
		_wantEnable = wantEnable;
		_actionDescription = [NSMutableArray arrayWithCapacity:20];
	}
	return self;
}


NSString *getDictString(CFDictionaryRef dict, CFStringRef key)
{
	const void *p = dict ? CFDictionaryGetValue(dict, key) : NULL;
	if (!p || CFGetTypeID(p) != CFStringGetTypeID())
		return NULL;
	return (__bridge NSString*)p;
}

int getDictNumber(CFDictionaryRef dict, CFStringRef key)
{
	const void *p = dict ? CFDictionaryGetValue(dict, key) : NULL;
	if (!p || CFGetTypeID(p) != CFNumberGetTypeID())
		return -1;
	int v;
	if (!CFNumberGetValue((CFNumberRef)p, kCFNumberIntType, &v))
		return -2;
	return v;
}

- (BOOL)updateNetworkService:(SCNetworkServiceRef)serviceR
{
	int wantedPort = _wantEnable ? _proxyPort : 0;
	NSString *wantedHost = _wantEnable ? @"127.0.0.1" : NULL;
	//
	SCNetworkProtocolRef protoR = SCNetworkServiceCopyProtocol(serviceR, kSCNetworkProtocolTypeProxies);
	if (!protoR)
		return NO; // this protocol has not been enabled.
	if (!SCNetworkProtocolGetEnabled(protoR))
		return NO;
	CFDictionaryRef proxyDictR = SCNetworkProtocolGetConfiguration(protoR); // returns NULL if no configuration.
	//NSLog(@"   proxyDict: %@", proxyDictR);
	NSString* host = getDictString(proxyDictR, kSCPropNetProxiesHTTPProxy);
	int port = getDictNumber(proxyDictR, kSCPropNetProxiesHTTPPort);
	BOOL enabled = getDictNumber(proxyDictR, kSCPropNetProxiesHTTPEnable) > 0;
	DebugNSLog(@"  Settings{%@} enable = %d, host = %@, port = %d", (proxyDictR ? @"has" : @"missing"), enabled, host, port);
	BOOL needsUpdateEnabled = NO;
	BOOL needsUpdateHost = NO;
	BOOL needsUpdatePort = NO;
	CFMutableDictionaryRef newProxyDictR;
	if (proxyDictR)
		newProxyDictR = CFDictionaryCreateMutableCopy(NULL, 0, proxyDictR);
	else
		newProxyDictR = CFDictionaryCreateMutable(NULL, 0, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
	// just keep old host/port when we disable proxy.
	if (wantedHost && ![wantedHost isEqualToString:host]) {
		DebugNSLog(@"  host needs update: %@ -> %@", host, wantedHost);
		needsUpdateHost = YES;
		CFDictionarySetValue(newProxyDictR, kSCPropNetProxiesHTTPProxy, (__bridge CFStringRef)wantedHost);
	}
	if (wantedPort && port != wantedPort) {
		DebugNSLog(@"  port needs update: %d -> %d", port, wantedPort);
		needsUpdatePort = YES;
		CFDictionarySetValue(newProxyDictR, kSCPropNetProxiesHTTPPort, (__bridge const void *)(@(wantedPort)));
	}
	if (enabled != _wantEnable) {
		DebugNSLog(@"  enabled needs update: %d -> %d", enabled, _wantEnable);
		needsUpdateEnabled = YES;
		CFDictionarySetValue(newProxyDictR, kSCPropNetProxiesHTTPEnable, (__bridge const void *)(@(_wantEnable ? 1 : 0)));
	}
	DebugNSLog(@"  needsUpdate = %d+%d+%d\n", needsUpdateEnabled, needsUpdateHost, needsUpdatePort);
	[_actionDescription addObject:(__bridge NSString*)SCNetworkServiceGetName(serviceR)];
	if (!needsUpdateEnabled && !needsUpdateHost && !needsUpdatePort) {
		[_actionDescription addObject:@"      No change."];
		CFRelease(newProxyDictR);
		return NO;
	}
	if (_wantEnable) {
		if (needsUpdateEnabled && needsUpdateHost && needsUpdatePort) {
			[_actionDescription addObject:@"      Enable proxy and set host/port."];
		} else if (!needsUpdateHost && !needsUpdatePort) {
			[_actionDescription addObject:@"      Enable proxy."];
		} else {
			NSMutableArray* a = [NSMutableArray arrayWithCapacity:3];
			if (needsUpdateEnabled)
				[a addObject:@"Enable proxy"];
			if (needsUpdateHost) {
				NSString* s = [NSString stringWithFormat:@"update host ('%@' -> '%@)", (host ? host : @""), wantedHost];
				[a addObject:s];
			}
			if (needsUpdatePort) {
				NSString* s = [NSString stringWithFormat:@"update port: %@ -> %d",
					(port ? [NSString stringWithFormat:@"%d", port] : @"none"),
					wantedPort];
				[a addObject:s];
			}
			if (![a count])
				[a addObject:@"???? no modifications?"];
			NSString* msg = [NSString stringWithFormat:@"    %@", [a componentsJoinedByString:@", "]];
			[_actionDescription addObject:msg];
		}
	} else {
		[_actionDescription addObject:@"      Disable proxy."];
	}
	// DebugNSLogNSLog(@" new dict: %@", newProxyDictR);
	BOOL ok = SCNetworkProtocolSetConfiguration(protoR, newProxyDictR);
	CFRelease(newProxyDictR);
	if (!ok)
		DebugNSLog(@"SCNetworkProtocolSetConfiguration failed!");
	return YES;
}

- (SCPreferencesRef)createSCPreferencesRef:(BOOL)commit
{
	CFStringRef appName = CFSTR("GlimmerBlocker");
	return SCPreferencesCreate(NULL, appName, NULL);
}

- (BOOL)updateSettingsWithCommit:(BOOL)commit
{
	[_actionDescription removeAllObjects];
	SCPreferencesRef prefsR = [self createSCPreferencesRef:commit];
	NSAssert(prefsR, @"prefsR");
	// can only get lock using auth.
	BOOL gotLock = commit ? SCPreferencesLock(prefsR, NO) : NO;
	if (commit && !gotLock)
		NSLog(@"Can't get preferences lock");
	SCNetworkSetRef networkSetR = SCNetworkSetCopyCurrent(prefsR);
	NSAssert(networkSetR, @"networkSetR");
	//
	BOOL needsUpdate = NO;
	CFArrayRef networkServicesArrayR = SCNetworkSetCopyServices(networkSetR);
	NSAssert(networkServicesArrayR, @"networkServicesArrayR");
	for (CFIndex i = 0; i < CFArrayGetCount(networkServicesArrayR); i++) {
		SCNetworkServiceRef serviceR = CFArrayGetValueAtIndex(networkServicesArrayR, i);
		BOOL enabled = SCNetworkServiceGetEnabled(serviceR);
		DebugNSLog(@"Service [%ld], name = %@%@", (long)i, SCNetworkServiceGetName(serviceR),
			  enabled ? @"" : @"  (disabled)");
		if (!enabled)
			continue;
		if ([self updateNetworkService:serviceR])
			needsUpdate = YES;
	}
	//
	CFRelease(networkServicesArrayR);
	CFRelease(networkSetR);
	BOOL ok = !needsUpdate;
	if (needsUpdate && commit) {
		ok = SCPreferencesCommitChanges(prefsR);
		if (!ok) {
			NSLog(@" SCPreferencesCommitChanges failed");
		} else {
			ok = SCPreferencesApplyChanges(prefsR);
			if (!ok)
				NSLog(@" SCPreferencesApplyChanges failed");
		}
		//
		if (gotLock) {
			BOOL unlocked = SCPreferencesUnlock(prefsR);
			if (!unlocked)
				NSLog(@"### Got problem releasing prefs lock");
		}
	}
	CFRelease(prefsR);
	return ok;
}

@end
