/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import <Security/AuthorizationTags.h>
#import <SecurityInterface/SFAuthorizationView.h>
#include <CommonCrypto/CommonDigest.h>
#import "Prefs.h"
#import "GBDebug.h"
#import "XmlDoc.h"
#import "Xml.h"
#import "AlertHelper.h"

static NSString* authHelperPath = @"/Library/PreferencePanes/GlimmerBlocker.prefPane/Contents/AuthHelper.app/Contents/MacOS/AuthHelper";

@interface PrefsHelper : NSObject {
}

@property NSFileManager *fm;
@property NSString* dirPath;
@property NSString* filePath;
@property AlertHelper* alertHelper;


- (instancetype)initWithAlertHelper:(AlertHelper*)alertHelper;

- (NSString*)filePath;
- (AlertHelper*)alertHelper;

- (XmlDoc*)load;

@end

@implementation PrefsHelper

- (instancetype)initWithAlertHelper:(AlertHelper*)alertHelper
{
	if (!(self = [super init]))
		return NULL;
	_alertHelper = alertHelper;
	_fm = [NSFileManager defaultManager];
	_dirPath = @"/Library/GlimmerBlocker";
	_filePath = [NSString stringWithFormat:@"%@/PanelSettings.xml", _dirPath];
	return self;
}


- (XmlDoc*)load
{
	if (![_fm fileExistsAtPath:_filePath]) {
		NSLog(@"GlimmerBlocker preferences file missing: %@", _filePath);
		return NULL;
	}
	NSURL* url = [NSURL fileURLWithPath:_filePath];
	NSData* data = [NSData dataWithContentsOfURL:url];
	if (!data)
		return NULL;
	XmlDoc* doc = [[XmlDoc alloc] initWithURL:url withData:data];
	if ([doc doc] && [doc rootE])
		return doc;
	NSLog(@"Failed loading data from %@", _filePath);
	return NULL;
}

@end

//==============================================================================================================
@interface Prefs ()

@property PrefsHelper *helper;
@property BOOL dirty;
//
@property SFAuthorizationView* authView;
@property BOOL hasGlimmerAuth;
@property BOOL hasWarnedInstallProblem;
@property NSData* authInExternalForm;
@property NSUInteger numberOfRunningHelpers;
@property (readonly) BOOL checkAutoHelperPermissions;

- (instancetype)initWithPrefsHelper:(PrefsHelper*)helper
					withRootElement:(NSXMLElement*)rootE
			withSFAuthorizationView:(SFAuthorizationView*)authView;

@end

@implementation Prefs

+ (Prefs*)prefsWithAlertHelper:(AlertHelper*)alertHelper
	   withSFAuthorizationView:(SFAuthorizationView*)authView
{
	NSAssert(alertHelper, @"alertHelper");
	NSAssert(authView, @"authView");
	PrefsHelper* helper = [[PrefsHelper alloc] initWithAlertHelper:alertHelper];
	if (!helper)
		return NULL;
	XmlDoc* doc = [helper load];
	NSXMLElement* rootE = doc ? [doc rootE] : [Xml createElementWithName:@"prefs"];
	return [[Prefs alloc] initWithPrefsHelper:helper
							   withRootElement:rootE
					   withSFAuthorizationView:authView];
}

- (instancetype)initWithPrefsHelper:(PrefsHelper*)helper
					withRootElement:(NSXMLElement*)rootE
			withSFAuthorizationView:(SFAuthorizationView*)authView
{
	if (!(self = [super initWithRootElement:rootE]))
		return NULL;
	_helper = helper;
	_authView = authView;
	return self;
}


- (void)forcedReload
{
	XmlDoc* doc = [_helper load];
	if (!doc) {
		NSLog(@"GB prefs can't reload xml panel prefs.");
		return;
	}
	NSXMLElement* rootE = [doc rootE];
	if (!rootE) {
		NSLog(@"GB prefs can't reload xml panel prefs: got doc without root element");
		return;
	}
	[self reloadWithRootElement:rootE];
	DebugNSLog(@"Prefs reloaded.");
}

/*
- (id)retain
{
	NSLog(@"#### Prefs retain: %d", [self retainCount]);
	[GBDebug debugDumpStackframe];
	return [super retain];
}

- (void)release
{
	NSLog(@"#### Prefs release: %d", [self retainCount]);
	[GBDebug debugDumpStackframe];
	[super release];
}
 */

#pragma mark -------------------------------- Globals
- (AlertHelper*)alertHelper
{
	return [_helper alertHelper];
}

#pragma mark -------------------------------- Filters

- (NSXMLElement*)filtersE
{
	return [Xml singletonElementForXPath:@"filters" forParent:[self rootE]];
}

- (NSXMLElement*)rulePreviewE
{
	return [Xml singletonElementForXPath:@"rule-preview" forParent:[self rootE]];
}

- (BOOL)zapRulePreviewElement
{
	NSArray* a = [[self rootE] elementsForName:@"rule-preview"];
	if (![a count])
		return NO;
	for (NSXMLElement* e in a)
		[e detach];
	[self markAsDirty];
	return YES;
}

#pragma mark -------------------------------- Save

#ifdef DEBUG_DISABLED
//#ifdef DEBUG
#define DirtyPrefsNSLog(fmt, ...) NSLog(fmt, ## __VA_ARGS__)
#else
#define DirtyPrefsNSLog(fmt, ...) do { } while (0)
#endif

- (BOOL)save
{
	if (![self hasGlimmerAuth]) {
		NSLog(@"Prefs.save without auth. Ignored.");
		return NO;
	}
	if (_dirty)
		DirtyPrefsNSLog(@"Saving dirty prefs, _numberOfRunningHelpers = %d", _numberOfRunningHelpers);
	else
		DirtyPrefsNSLog(@"Saves non-dirty prefs");
	NSData* fileContents = [self printToData];
	NSData* md5 = [GBDebug md5:fileContents];
	NSArray* inputData = @[fileContents, md5];
	BOOL beta = [self boolForXPath:@"updater/@use-beta-versions"];
	NSArray* args = @[@"--save-prefs", beta ? @"--beta" : @"--final"];
	int status = [self runAuthHelperWithArgs:(NSArray*)args withInputData:inputData withAuth:YES];
	if (!status) {
		_dirty = NO;
		DirtyPrefsNSLog(@"Save ok, _numberOfRunningHelpers = %d", _numberOfRunningHelpers);
		return YES;
	}
	DirtyPrefsNSLog(@"Save failed, _numberOfRunningHelpers = %d", _numberOfRunningHelpers);
	if (!_hasWarnedInstallProblem) {
		AlertHelper* ah = [self alertHelper];
		[ah prepareCriticalAlertWithTitle:@"Failed saving preferences"
					  withInformativeText:[_helper filePath]];
		[ah showPreparedAlert];
	}
	return NO;
}

- (void)saveIfDirty:(id)sender
{
	if (_numberOfRunningHelpers) // dunny why I get these nested calls.
		return;
	DirtyPrefsNSLog(@"saveIfDirty: %d, _numberOfRunningHelpers = %d", _dirty, _numberOfRunningHelpers);
	//[GBDebug debugDumpStackframe];
	if (_dirty) {
		[self save];
		DirtyPrefsNSLog(@"  dirty: %d, _numberOfRunningHelpers = %d", _dirty, _numberOfRunningHelpers);
	}
}

- (void)markAsDirty
{
	if (_dirty)
		return;
	DirtyPrefsNSLog(@"Prefs now dirty, _numberOfRunningHelpers = %d", _numberOfRunningHelpers);
	_dirty = YES;
	[self performSelectorOnMainThread:@selector(saveIfDirty:) withObject:self waitUntilDone:NO];
}


#pragma mark -------------------------------- XPath scalars

- (void)setString:(NSString*)value forXPath:(NSString*)xpath
{
	NSString* old = [self stringForXPath:xpath];
	if (![old length] && ![value length])
		return;
	if (old && value && [old isEqual:value])
		return;
	DebugNSLog(@"Changes global pref (%@): '%@' -> '%@'", xpath, old, value);
	[super setString:value forXPath:xpath];
}

#pragma mark -------------------------------- Auth

- (AuthorizationRef)authRef
{
	if (!_hasGlimmerAuth)
		return NULL;
	SFAuthorization* auth = [_authView authorization];
	if (!auth) {
		DebugNSLog(@"Has no SFAuthorization");
		return NULL;
	}
	return [auth authorizationRef];
}

- (void)notifyUpdatedAuthorizationState
{
	_hasGlimmerAuth = YES; // so [self authRef] doesn't short-circuit to NULL
	AuthorizationRef authRef = [self authRef];
	AuthorizationItem myItems = {glimmerAuthString, 0, NULL, 0};
	AuthorizationRights myRights = {1, &myItems};
	AuthorizationRights *authorizedRights = NULL;
	OSStatus stat = AuthorizationCopyRights(authRef, &myRights, NULL,
											kAuthorizationFlagDefaults,
											&authorizedRights);
	if (authorizedRights)
		AuthorizationFreeItemSet(authorizedRights);
	DebugNSLog(@"AuthorizationCopyRights returned %d", stat);
	_hasGlimmerAuth = (stat == errAuthorizationSuccess);
	_authInExternalForm = NULL;
	if (!_hasGlimmerAuth)
		return;
	AuthorizationExternalForm extAuth;
	stat = AuthorizationMakeExternalForm(authRef, &extAuth);
	if (stat == errAuthorizationSuccess) {
		_authInExternalForm = [NSData dataWithBytes:&extAuth length:sizeof(extAuth)];
	} else {
		NSLog(@"AuthorizationMakeExternalForm failed (%ld) after AuthorizationCopyRights returned success", (long)stat);
		_hasGlimmerAuth = NO;
	}
	if (_hasGlimmerAuth) {
		_hasWarnedInstallProblem = NO;
		[self checkAutoHelperPermissions];
	}
}

- (BOOL)has:(int)value withMask:(int)mask withKey:(NSString*)key inDict:(NSDictionary*)dict
{
	NSObject* obj = dict[key];
	if (!obj || ![obj isKindOfClass:[NSNumber class]])
		return NO;
	NSNumber* n = (NSNumber*)obj;
	BOOL ok = ([n intValue] & mask) == value;
	if (!ok)
		DebugNSLog(@"Missing file attr: %@ = %d, expected %d with mask %d", key, [n intValue], value, mask);
	return ok;
}

- (BOOL)checkAutoHelperPermissions
{
	NSDictionary* attrs = [[NSFileManager defaultManager] attributesOfItemAtPath:authHelperPath error:NULL];
	if (![self has:06555 withMask:06555 withKey:NSFilePosixPermissions inDict:attrs]
		|| ![self has:0 withMask:-1 withKey:NSFileOwnerAccountID inDict:attrs]) {
		if ([_authView authorizationState] == SFAuthorizationViewUnlockedState)
			NSLog(@"GB Deauthorizes due to invalid AuthHelper permissions/owner.");
		[_authView deauthorize:self];
		if (_hasWarnedInstallProblem)
			return NO;
		if (![[_helper alertHelper] canDisplayAlert])
			return NO;
		_hasWarnedInstallProblem = YES;
		AlertHelper* ah = [self alertHelper];
		[ah prepareCriticalAlertWithTitle:@"GlimmerBlocker needs to be reinstalled"
					  withInformativeText:@"Reason: The authentication tool does not have the proper permission assigned."];
		[ah showPreparedAlert];
		return NO;
	}
	_hasWarnedInstallProblem = NO;
	return YES;
}

- (void)writeData:(NSData*)data toMutableData:(NSMutableData*)mdata
{
	long long len = [data length];
	NSAssert(sizeof(len) == 8, @"longlong != 64");
	[mdata appendBytes:&len length:sizeof(len)];
	[mdata appendData:data];
}

- (int)runAuthHelperWithArgs:(NSArray*)args
			   withInputData:(NSArray*)inputData // array of NSData
					withAuth:(BOOL)withAuth
{
	return [self runAuthHelperWithArgs:args
						 withInputData:inputData
						withOutputData:NULL
							  withAuth:withAuth];
}

- (int)runAuthHelperWithArgs:(NSArray*)args
			   withInputData:(NSArray*)inputData
			  withOutputData:(NSMutableData*)outputData
					withAuth:(BOOL)withAuth
{
	NSAssert([args count], @"Must have --command");
	if (withAuth && ![self hasGlimmerAuth]) {
		NSLog(@"Can't run helper tool: required auth, but doesn't have it: %@", args[0]);
		return -1;
	}
	NSMutableData* allInput = (withAuth || inputData) ? [NSMutableData dataWithCapacity:250] : NULL;
	if (withAuth)
		[self writeData:_authInExternalForm toMutableData:allInput];
	if (inputData) {
		for (NSData* data in inputData)
			[self writeData:data toMutableData:allInput];
	}
	_numberOfRunningHelpers++;
	int status = [GBDebug executeCommandlineTool:authHelperPath withArgs:args optStdinData:allInput optStdoutData:outputData];
	_numberOfRunningHelpers--;
	if (status)
		DebugNSLog(@"Status from auto-helper %@: %d", args[0], status);
	return status;
}

@end
