/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "FilterPublisher.h"
#import "GBDebug.h"
#import "XmlDoc.h"
#import "Xml.h"
#import "ServerCommunication.h"
#import "AlertHelper.h"
#import "Prefs.h"
#import "Filter.h"
#import "FilterTabController.h"
#import "AutoZap.h"

#define CURRENT_VERSION 3

@interface FilterPublisher ()

@property Prefs* prefs;
@property AlertHelper* alertHelper;
@property ServerCommunication* comm;
@property FilterTabController* filterTabController;
@property BOOL hasAutoReleasedSelf;

@property BOOL suppressUserFeedback; // e.g. if user cancelled, but connection still returns error code.
@property NSURLConnection* connection;
@property NSInteger lastHttpStatusCode;
@property NSMutableData* responseBody;
@property NSURL* url;
@property Filter* filter;
@property NSArray *topLevelObjects;

@property NSURLAuthenticationChallenge* authChallenge;

@property (weak) IBOutlet NSWindow*				urlSheet;
@property (weak) IBOutlet NSTextField*			urlTextField;
@property (weak) IBOutlet NSTextField*			urlErrorLabel;
@property (weak) IBOutlet NSButton*				urlOkBtn;

@property (weak) IBOutlet NSWindow*				authSheet;
@property (weak) IBOutlet NSTextField*			authMessageTextField;
@property (weak) IBOutlet NSTextField*			authNameTextField;
@property (weak) IBOutlet NSTextField*			authPasswordTextField;
@property (weak) IBOutlet NSButton*				authRememberCB;

@property (weak) IBOutlet NSWindow*				progressSheet;
@property (weak) IBOutlet NSProgressIndicator*	processIndicator;

@property BOOL showsProgressSheet;

- (void)startWebDavPublish;

- (void)hideProgressSheet;
- (void)showProgressSheet;

- (BOOL)validateSimpleUrl:(BOOL)verbose;

@end

@implementation FilterPublisher

- (instancetype)initWithFilterTabController:(FilterTabController*)filterTabController
								 withFilter:(Filter*)filter
{
	if (!(self = [super init]))
		return nil;
	[AutoZap addSticky:self];
	_topLevelObjects = [AutoZap loadNibNamed:@"FilterPublisher" bundle:[filterTabController mainBundle] owner:self];
	NSAssert(_urlSheet, @"urlSheet is null");
	_filterTabController = filterTabController;
	_comm = [filterTabController comm];
	_prefs = [_comm prefs];
	_alertHelper = [_prefs alertHelper];
	_filter = filter;
	 // until we're done, released by [self dismiss].
	//
	if ([_filter url]) {
		_url = [NSURL URLWithString:[_filter url]];
		[self startWebDavPublish];
		return self;
	}
	[_urlTextField setDelegate:self];
	[self validateSimpleUrl:NO];
	[NSApp beginSheet: _urlSheet
	   modalForWindow: [_alertHelper window]
		modalDelegate: nil
	   didEndSelector: nil
		  contextInfo: nil];
	return self;
}

- (void)dealloc
{
	[self hideProgressSheet];
	[self dismiss];
	DebugNSLog(@"Publish.dealloc");
}

- (void)dismiss
{
	if (_hasAutoReleasedSelf)
		return;
	DebugNSLog(@"dismiss.commit");
	_hasAutoReleasedSelf = YES;
	[_connection cancel];
	[AutoZap add:self];
	[AutoZap zapSticky:self];
}

#pragma mark ------------------------ simple URL

- (BOOL)validateSimpleUrl:(BOOL)verbose
{
	NSString* s = [_urlTextField stringValue];
	NSString* error = NULL;
	if (![s length]) {
		error = @" ";
	} else if (![s hasPrefix:@"http://"] && ![s hasPrefix:@"https://"]) {
		error = @"URL must start with  http://  or  https://";
	} else if (![s hasSuffix:@".xml"]) {
		error = @"URL must end with  .xml";
	} else {
		NSString* host = [Filter extractHostFromStringUrl:s];
		if (![host length])
			error = @"URL has invalid hostname";
		else if (verbose && ([host isEqual:@"example.com"] || [host hasSuffix:@".example.com"]))
			error = @"example.com is not valid as host.";
	}
	if (!error) {
		@try {
			if (![NSURL URLWithString:s])
				error = @"Invalid URL (NSURL failed)";
		}
		@catch (NSException* ex) {
			error = @"Invalid URL (NSURL exception)";
		}
	}
	if (error)
		[_urlErrorLabel setStringValue:error];
	[_urlErrorLabel setHidden:!error];
	[_urlOkBtn setEnabled:!error];
	return !error;
}

- (void)controlTextDidChange:(NSNotification *)aNotification
{
	//DebugNSLog(@"controlTextDidChange: %@", aNotification);
	[self validateSimpleUrl:NO];
}

#pragma mark ------------------------ progress
- (void)hideProgressSheet
{
	if (!_showsProgressSheet)
		return;
	_showsProgressSheet = NO;
	[NSApp endSheet:_progressSheet];
	[_progressSheet orderOut:self];
}

- (void)showProgressSheet
{
	if (_showsProgressSheet || _hasAutoReleasedSelf)
		return;
	_showsProgressSheet = YES;
	[NSApp beginSheet: _progressSheet
	   modalForWindow: [_alertHelper window]
		modalDelegate: nil
	   didEndSelector: nil
		  contextInfo: nil];
	[_processIndicator startAnimation:self];
}

- (IBAction)uploadCancelAction:(id)sender
{
	[self dismiss];
}

#pragma mark ------------------------ url sheet

- (IBAction)publishCancelAction:(id)sender
{
	DebugNSLog(@"FilterPublisher.publishCancelAction");
	_suppressUserFeedback = YES;
	[NSApp endSheet:_urlSheet];
	[_urlSheet orderOut:self];
	[self dismiss];
}

- (IBAction)publishUrlConfirmAction:(id)sender
{
	DebugNSLog(@"FilterPublisher.publishUrlConfirmAction");
	if (![self validateSimpleUrl:YES])
		return;
	[NSApp endSheet:_urlSheet];
	[_urlSheet orderOut:self];
	@try {
		_url = [NSURL URLWithString:[_urlTextField stringValue]];
	}
	@catch (NSException* ex) {
		NSLog(@"Got exception in publishConfirmAction: %@", ex);
		[_alertHelper prepareCriticalAlertWithTitle:@"Invalid URL" withInformativeText:NULL];
		[_alertHelper showPreparedAlert];
		return;
	}
	NSString* u = [_url absoluteString];
	for (Filter* filter in [_filterTabController filterArray]) {
		if (![u isEqual:[filter url]])
			continue;
		NSString* msg = [filter isSubscription]
		? @"Cannot publish filter with same URL as a subscription as it would overwrite the subscription original."
		: @"Cannot publish filter with same URL as another published filter as it would overwrite the other published filter.";
		[_alertHelper prepareCriticalAlertWithTitle:@"Duplicate URL" withInformativeText:msg];
		[_alertHelper showPreparedAlert];
		return;
	}
	[self startWebDavPublish];
}

#pragma mark ------------------------ start uploading
- (void)startWebDavPublish
{
	DebugNSLog(@"Publish.start");
	NSXMLElement *filterDataE = [[_filter filterDataE] copy];
	NSString* comments = [_filter comments];
	if ([comments length]) {
		[Xml setAttribute:comments withName:@"description" inElement:filterDataE]; // storing in attribute is old format.
		[Xml setCData:comments forXPath:@"description" withRoot:filterDataE];
	}
	[Xml setAttribute:[NSString stringWithFormat:@"%d", CURRENT_VERSION]
			 withName:@"format-version"
			inElement:filterDataE];
	[Xml setAttribute:[[_filterTabController bundle] infoDictionary][@"CFBundleVersion"]
			 withName:@"gb-version"
			inElement:filterDataE];
	XmlDoc* doc = [[XmlDoc alloc] initWithRootElement:filterDataE];
	NSData* data = [doc printToData];
	NSLog(@"Publishes filter, %ld bytes", (long)[data length]);
	//DebugNSLog(@"XML: %@", [doc printToString]);
	NSMutableURLRequest* req = [NSMutableURLRequest requestWithURL:_url
													   cachePolicy:NSURLRequestReloadIgnoringLocalCacheData
												   timeoutInterval:60.0];
	[req setHTTPMethod: @"PUT"];
	[req setValue:@"application/xml" forHTTPHeaderField:@"Content-type"];
	[req setHTTPBody:data];
	_responseBody = [NSMutableData dataWithCapacity:0];
	[self showProgressSheet];
	_connection = [[NSURLConnection alloc] initWithRequest:req delegate:self];
	if (_connection)
		return;
	[_alertHelper prepareCriticalAlertWithTitle:@"Failed to publish"
							withInformativeText:@"unknown error: NSURLConnection.initWithRequest returned null"];
	[_alertHelper showPreparedAlert];
	[self dismiss];
}

#pragma mark ------------------------ authentication

- (IBAction)authCancelAction:(id)sender
{
	DebugNSLog(@"Publish.authCancelAction");
	_suppressUserFeedback = YES;
	[NSApp endSheet:_authSheet];
	[_authSheet orderOut:self];
	[[_authChallenge sender] cancelAuthenticationChallenge:_authChallenge];
	[self dismiss];
}

- (IBAction)authConfirmAction:(id)sender
{
	DebugNSLog(@"Publish.authConfirmAction");
	[NSApp endSheet:_authSheet];
	[_authSheet orderOut:self];
	[self showProgressSheet];
	NSURLCredentialPersistence persistence = [_authRememberCB state]
	/**/ ? NSURLCredentialPersistencePermanent : NSURLCredentialPersistenceForSession;
	NSURLCredential *credential = [NSURLCredential credentialWithUser:[_authNameTextField stringValue]
															 password:[_authPasswordTextField stringValue]
														  persistence:persistence];
	DebugNSLog(@"  Uses auth: %@", credential);
	[[_authChallenge sender] useCredential:credential forAuthenticationChallenge:_authChallenge];
	[AutoZap add:_authChallenge];
	_authChallenge = NULL;
}


-(void)connection:(NSURLConnection *)connection didReceiveAuthenticationChallenge:(NSURLAuthenticationChallenge*)challenge
{
	NSLog(@"Publish.didReceiveAuthenticationChallenge: %@", challenge);
	NSURLCredential* credential = [challenge proposedCredential];
	if (credential && [credential hasPassword] && ![challenge previousFailureCount]) {
		[[challenge sender] useCredential:credential forAuthenticationChallenge:challenge];
		return;
	}
	NSURLProtectionSpace* space = [challenge protectionSpace];
	NSString* msg = @"";
	if ([challenge previousFailureCount])
		msg = @"Previously specified username/password was NOT accepted.\n\n";
	msg = [NSString stringWithFormat:@"%@Enter your name and password to publish the filter at the URL:\n   %@\n",
								   msg, [_url absoluteURL]];
	if ([[space realm] length])
		msg = [NSString stringWithFormat:@"%@in the realm \"%@\".\n", msg, [space realm]];
	else
		msg = [msg stringByAppendingString:@"\n"];
	if ([[space authenticationMethod] isEqual:NSURLAuthenticationMethodHTTPDigest])
		msg = [msg stringByAppendingString:@"Your name and password will be sent securely."];
	else
		msg = [msg stringByAppendingString:@"Your name and password will be send in plain text and NOT securely."];
	[_authMessageTextField setStringValue:msg];
	NSString* s = [credential user];
	[_authNameTextField setStringValue:[s length] ? s : @""];
	[_authPasswordTextField setStringValue:@""];
	[_authRememberCB setState:!credential || ([credential persistence] == NSURLCredentialPersistencePermanent)];
	[self hideProgressSheet];
	[NSApp beginSheet: _authSheet
	   modalForWindow: [_alertHelper window]
		modalDelegate: nil
	   didEndSelector: nil
		  contextInfo: nil];
	_authChallenge = challenge;
}

#ifdef DEBUG_XXX
- (BOOL)respondsToSelector:(SEL)aSelector
{
	BOOL rts = [super respondsToSelector:aSelector];
	DebugNSLog(@"FilterPublisher, %@  respondsToSelector  %s", (rts ? @"does" : @"does not"), aSelector);
	return rts;
}
#endif

#pragma mark ------------------------ misc

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response
{
	// called for each response, i.e. for each redirect
	[_responseBody setLength:0];
	NSHTTPURLResponse* http = (NSHTTPURLResponse*)response;
	_lastHttpStatusCode = [http statusCode];
	DebugNSLog(@"didReceiveResponse: %ld   %@", _lastHttpStatusCode, [http allHeaderFields]);
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data
{
	[_responseBody appendData:data];
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error
{
	NSLog(@"Publish subscription request failed (suppress = %d): %@", _suppressUserFeedback, error);
	if (_suppressUserFeedback) {
		[self dismiss];
		return;
	}
	NSString* a = [error localizedDescription];
	NSString* b = [error userInfo][NSURLErrorFailingURLStringErrorKey];
	NSString* updateError = [NSString stringWithFormat:@"%@ %@", a, b];
	[self hideProgressSheet];
	[_alertHelper prepareCriticalAlertWithTitle:@"Failed publishing filter" withInformativeText:updateError];
	[_alertHelper showPreparedAlert];
	[self dismiss];
}

- (void)handlePublishSuccess
{
	DebugNSLog(@"Publish success.");
	if (![_filter url]) {
		DebugNSLog(@"Updates filter with published URL.");
		[_filter setUrl:[_url absoluteString]];
	}
	[_filter notifyPublished];
	//
	[_filter readAttributesFromXml];
	[_prefs markAsDirty];
	[self dismiss];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection
{
	DebugNSLog(@"connectionDidFinishLoading, http status = %ld, suppress = %d", _lastHttpStatusCode, _suppressUserFeedback);
#ifdef DEBUG
	DebugNSLog(@"Publish response: %@",
			   [[NSString alloc] initWithData:_responseBody encoding:NSISOLatin1StringEncoding]);
#endif
	[self hideProgressSheet];
	if (_lastHttpStatusCode == 200 || _lastHttpStatusCode == 201 || _lastHttpStatusCode == 204) {
		// Servers should respond with HTTP_NO_CONTENT = 204
		// but some responds with HTTP_CREATED = 201 or HTTP_OK = 200
		[self handlePublishSuccess];
		return;
	}
	[self dismiss];
	if (_suppressUserFeedback)
		return;
	NSString* msg = [NSHTTPURLResponse localizedStringForStatusCode:_lastHttpStatusCode];
	if (_lastHttpStatusCode == 405)
		msg = [msg stringByAppendingString:@"\n\nThis typically means that the web server doesn't support WebDav."];
	NSString* s = [NSString stringWithFormat:@"Got http error %ld: %@", _lastHttpStatusCode, msg];
	[_alertHelper prepareCriticalAlertWithTitle:@"Failed publishing filter" withInformativeText:s];
	[_alertHelper showPreparedAlert];
	DebugNSLog(@"Ended with error alert: %@", s);
}

@end
