/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "FilterRuleTableView.h"
#import "Prefs.h"
#import "FilterTabController.h"
#import "Filter.h"
#import "XmlDoc.h"
#import "AlertHelper.h"
#import "CodeEditHelper.h"
#import "FilterRuleEdit.h"
#import "Xml.h"
#import "GBDebug.h"

@implementation FilterRuleTableView

static NSString* RulesPboardType = @"GlimmerBlockerRules";
static NSString* rulesRootElement = @"glimmerblocker-rules";

- (instancetype)initWithFrame:(NSRect)frameRect
{
	if (!(self = [super initWithFrame:frameRect]))
		return nil;
	return self;
}


- (FilterTabController*)filterTabController
{
	return (FilterTabController*)[self delegate];
}

#pragma mark ------------------------ Rule data

- (NSArray*)getPasteboardTypes
{
	return @[RulesPboardType, NSStringPboardType];
}

- (NSData*)makePasteboardDataForRowsWithIndexes:(NSIndexSet*)set
{
	XmlDoc* doc = [[XmlDoc alloc] initWithRootElementTagName:rulesRootElement];
	FilterTabController *controller = [self filterTabController];
	NSArray* rules = [controller ruleElements];
	for (NSInteger row = [set firstIndex]; row != NSNotFound; row = [set indexGreaterThanIndex:row]) {
		NSXMLElement* ruleE = [rules[row] copy];
		[[doc rootE] addChild:ruleE];
		[Xml setAttribute:NULL withName:@"rule-id" inElement:ruleE]; // remove id, no need for it in clippings.
	}
	NSData* data = [doc printToData];
	return data;
}

- (void)copyRows:(NSIndexSet*)rowIndexes
	toPasteboard:(NSPasteboard*)pb
{
	[pb declareTypes:[self getPasteboardTypes] owner:nil];
	NSData* data = [self makePasteboardDataForRowsWithIndexes:rowIndexes];
	DebugNSLog(@"Pasteboard data size: %ld", [data length]);
	[pb setData:data forType:RulesPboardType];
	[pb setData:data forType:NSStringPboardType];
}

- (BOOL)pasteboardDataSeemsValid:(NSData*)data
{
	if (!data)
		return NO;
	@try {
		NSString* s = [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
		s = [CodeEditHelper trimmedString:s];
		return [s rangeOfString:@"<?xml"].location == 0
		&& [s rangeOfString:@"<glimmerblocker-rules"].location != NSNotFound
		&& [s rangeOfString:@"<rule"].location != NSNotFound
		&& [s rangeOfString:@"</rule"].location != NSNotFound
		&& [s rangeOfString:@"</glimmerblocker-rules>"].location != NSNotFound;
	}
	@catch (NSException* ex) {
		DebugNSLog(@"Got exception in pasteboardDataSeemsValid: %@", ex);
	}
	return NO;
}

- (BOOL)canModifyRules
{
	FilterTabController *controller = [self filterTabController];
	if (![[controller prefs] hasGlimmerAuth])
		return NO;
	Filter* filter = [controller selectedFilter];
	return [filter canEditRules];
}

- (NSData*)getValidDataFromPasteboard:(NSPasteboard*)pb
{
	NSData* data = [pb dataForType:RulesPboardType];
	if ([self pasteboardDataSeemsValid:data])
		return data;
	data = [pb dataForType:NSStringPboardType];
	if ([self pasteboardDataSeemsValid:data])
		return data;
	return NULL;
}

- (void)showCantPasteAlert:(NSString*)message
{
	FilterTabController *controller = [self filterTabController];
	AlertHelper* ah = [controller alertHelper];
	[ah prepareCriticalAlertWithTitle:@"Can't paste"
				  withInformativeText:message];
	[ah showPreparedAlert];
}

- (void)insertRuleData:(NSData*)data
{
	FilterTabController *controller = [self filterTabController];
	Filter* filter = [controller selectedFilter];
	if (!filter) {
		[self showCantPasteAlert:@"No filter is selected."];
		return;
	}
	if (![self pasteboardDataSeemsValid:data]) {
		[self showCantPasteAlert:@"The clipboard does not contain a valid rule clipping."];
		return;
	}
	NSURL* url = [NSURL URLWithString:@"dummy:"];
	XmlDoc* doc = [[XmlDoc alloc] initWithURL:url withData:data];
	if ([doc error]) {
		[self showCantPasteAlert:@"Cannot parse the rule clipping."];
		return;
	}
	NSArray* ruleList = [doc nodesForXPath:@"rule"];
	if (![ruleList count]) {
		// happends only if the <rule> was commented out and thereby fooling pasteboardDataSeemsValid.
		[self showCantPasteAlert:@"The rule clipping is empty, it has no rules."];
		return;
	}
	NSXMLElement* filterDataE = [filter filterDataE];
	for (NSXMLElement* ruleE in ruleList) {
		[ruleE detach];
		[Xml setAttribute:NULL withName:@"rule-id" inElement:ruleE]; // remove id, so it gets a new one.
		[filterDataE addChild:ruleE];
		[FilterRuleEdit getRuleId:ruleE]; // ensures it has a rule-id
	}
	[self reloadData];
	[controller setSavedRuleSelection:ruleList];
	[controller updateTabViewWithForceUpdate:YES];
}

#pragma mark ------------------------ Copy/paste

- (BOOL)validateMenuItem:(NSMenuItem *)item
{
   NSInteger row = [self selectedRow];
	if ([item action] == @selector(copy:))
		return row >= 0;
	if ([item action] == @selector(cut:))
		return row >= 0 && [self canModifyRules];
	if ([item action] == @selector(paste:))
		return [self canModifyRules] && [self getValidDataFromPasteboard:[NSPasteboard generalPasteboard]];
	if ([item action] == @selector(selectAll:))
		return [self numberOfRows] > 0;
	return YES;
}

- (void)copy:(id)sender
{
	NSPasteboard* pb = [NSPasteboard generalPasteboard];
	[self copyRows:[self selectedRowIndexes] toPasteboard:pb];
}

- (void)cut:(id)sender
{
	[self copy:sender];
	if (![self canModifyRules])
		return;
	FilterTabController *controller = [self filterTabController];
	[controller ruleRemoveAction:sender];
}

- (void)paste:(id)sender
{
	if (![self canModifyRules])
		return;
	NSData* data = [self getValidDataFromPasteboard:[NSPasteboard generalPasteboard]];
	if (data)
		[self insertRuleData:data];
}

- (void)selectAll:(id)sender
{
	NSIndexSet* indexes = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, [self numberOfRows])];
	[self selectRowIndexes:indexes byExtendingSelection:NO];
}

#pragma mark ------------------------ Dragging

- (BOOL)canDragRowsWithIndexes:(NSIndexSet *)rowIndexes atPoint:(NSPoint)mouseDownPoint
{
	return [rowIndexes count] > 0;
}

#pragma mark ------------------------ Misc

#if 0
- (BOOL)respondsToSelector:(SEL)aSelector
{
	BOOL rts = [super respondsToSelector:aSelector];
	DebugNSLog(@"FilterRuleTableView, %@ respondsToSelector  %s", (rts ? @"does   " : @"does not"), aSelector);
	return rts;
}
#endif

@end
