/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "FilterTabController.h"
#import "GBDebug.h"
#import "XmlDoc.h"
#import "Xml.h"
#import "ServerCommunication.h"
#import "AlertHelper.h"
#import "Prefs.h"
#import "Filter.h"
#import "FilterListView.h"
#import "FilterPublisher.h"
#import "FilterRuleEdit.h"
#import "FilterRuleTableView.h"
#import "Rule.h"
#import "HostTextFieldCell.h"
#import "AutoZap.h"

@interface FilterTabController ()

@property NSTabView* tabView;
@property NSTabViewItem* filtersTabViewItem;
@property NSBundle* bundle;
@property Prefs* prefs;
@property AlertHelper* alertHelper;
@property ServerCommunication* comm;
@property NSMutableArray* ruleElementArray;
@property Filter* rulesOwnerFilter;
@property NSArray *topLevelObjects;
@property NSImage* subscriptionGlobeImage;
@property NSImage* publishGlobeImage;
@property NSImage* publishGlobeDirtyImage;
@property NSImage* warningImage;
@property NSImage* updateAvailableImage;
@property BOOL hasPostInited;

@property NSMutableArray* rulesSortColumnIndexes;

@property int addSubFilterId;
@property NSString* addSubPathArgs;
@property NSURLConnection* addSubConnection;
@property NSMutableData* addSubData;
@property NSURL* addSubUrl;



@property (weak) IBOutlet NSView*				subView;
@property (weak) IBOutlet NSSplitView*			splitView;

@property (weak) IBOutlet FilterListView*		filterVertViewFlow;
@property (weak) IBOutlet NSButton*				filterAddBtn;
@property (weak) IBOutlet NSButton*				filterRemoveBtn;
@property (weak) IBOutlet NSButton*				filterActionBtn;
//
@property (weak) IBOutlet NSMenu*				filterActionMenu;
@property (weak) IBOutlet NSMenuItem*			filterAddByUrlMI;
@property (weak) IBOutlet NSMenuItem*			filterDuplicateMI;
@property (weak) IBOutlet NSMenuItem*			filterUpdateSelectedMI;
@property (weak) IBOutlet NSMenuItem*			filterUpdateAllMI;
@property (weak) IBOutlet NSMenuItem*			filterEditMI;
@property (weak) IBOutlet NSMenuItem*			filterPublishWebDavMI;
//
@property (weak) IBOutlet NSTextField*			filterCommentsLabelTV;
@property (weak) IBOutlet NSTextField*			filterCommentsTV;
//
@property (weak) IBOutlet NSTextField*			rulesWithoutServer;
@property (weak) IBOutlet NSTextField*			rulesTableHeader;
@property (weak) IBOutlet NSScrollView*			rulesTableScrollView;
@property (weak) IBOutlet FilterRuleTableView*	rulesTableView;
@property (weak) IBOutlet NSButton*				ruleAddBtn;
@property (weak) IBOutlet NSButton*				ruleRemoveBtn;
@property (weak) IBOutlet NSButton*				ruleActionBtn;
//
@property (weak) IBOutlet NSMenu*				ruleActionMenu;
@property (weak) IBOutlet NSMenuItem*			ruleDuplicateMI;
//
// Add subscription sheet:
@property (weak) IBOutlet NSWindow*				subAddSheet;
@property (weak) IBOutlet NSTextField*			subAddUrlTextField;
@property (weak) IBOutlet NSMatrix*				subAddUpdatesMatrix;
@property (weak) IBOutlet NSMatrix*				subAddSafetyMatrix;


- (void)loadRulesWithForceUpdate:(BOOL)forceUpdate;
- (void)sortRules;

- (void)ruleEditAction:(id)sender;

- (Filter*)createFilterForNewElement:(NSXMLElement*)filterE;
- (Filter*)createFilterForExistingElement:(NSXMLElement*)filterE;
- (Filter*)createFilterForExistingElement:(NSXMLElement*)filterE atIndex:(NSUInteger)idx;

- (void)releaseUpdateWithCancel:(BOOL)cancel;
@end

//=======================================================================================================

@implementation FilterTabController


+ (NSImage*)getPngImage:(NSString*)name inBundle:(NSBundle*)bundle
{
	NSString* path = [bundle pathForResource:name ofType:@"png"];
	NSAssert1(path, @"Missing Image: %@.png", name);
	return [[NSImage alloc] initWithContentsOfFile:path];
}

- (instancetype)initWithServerCommunication:(ServerCommunication*)comm
				withContainerView:(NSView*)containerView
					  withTabView:(NSTabView*)tabView
		   withFiltersTabViewItem:(NSTabViewItem*)filtersTabViewItem
					   withBundle:(NSBundle*)bundle
{
	if (!(self = [super init]))
		return nil;
	_topLevelObjects = [AutoZap loadNibNamed:@"FilterTabView" bundle:[comm mainBundle] owner:self];
	NSAssert(_filterVertViewFlow, @"filterVertViewFlow is null");
	[_filterVertViewFlow setDelegate:self];
	_tabView = tabView; // no retain: owned by superview
	_filtersTabViewItem = filtersTabViewItem; // no retain: owned by superview
	_comm = comm;
	_prefs = [comm prefs];
	_alertHelper = [_prefs alertHelper];
	_bundle = bundle;
	[_filterVertViewFlow setController:self];
	[containerView addSubview:_subView];
	_subscriptionGlobeImage = [FilterTabController getPngImage:@"SubscriptionGlobe" inBundle:bundle];
	_publishGlobeImage = [FilterTabController getPngImage:@"PublishGlobe" inBundle:bundle];
	_publishGlobeDirtyImage = [FilterTabController getPngImage:@"PublishGlobeDirty" inBundle:bundle];
	_warningImage = [FilterTabController getPngImage:@"WarningIcon" inBundle:bundle];
	_updateAvailableImage = [FilterTabController getPngImage:@"UpdateAvailableIcon" inBundle:bundle];
	_ruleElementArray = [NSMutableArray arrayWithCapacity:40];
	_rulesOwnerFilter = NULL;
	[_splitView setDelegate:self];
	[_rulesTableView registerForDraggedTypes:[_rulesTableView getPasteboardTypes]];
	[_rulesTableView setDraggingSourceOperationMask:NSDragOperationCopy forLocal:YES];
	[_rulesTableView setDraggingSourceOperationMask:NSDragOperationCopy forLocal:NO];
	return self;
}

- (void)updateRuleSortingIcon
{
	NSUInteger v = [_rulesSortColumnIndexes count] ? [_rulesSortColumnIndexes[0] intValue] : 0;
	NSUInteger selectedIndex = v / 2;
	BOOL ascending = (v & 1);
	//
	NSArray* cols = [_rulesTableView tableColumns];
	for (NSUInteger idx = 0; idx < [cols count]; idx++) {
		NSImage* img = NULL;
		if (idx == selectedIndex) {
			NSString* imageName = ascending ? @"NSAscendingSortIndicator" :  @"NSDescendingSortIndicator";
			img = [NSImage imageNamed:imageName];
		}
		[_rulesTableView setIndicatorImage:img inTableColumn:cols[idx]];
	}
}

- (void)addSortingColumn:(NSString*)columnId
			 isAscending:(BOOL)ascending
			   forUpdate:(BOOL)updating
{
	NSInteger columnIndex = [_rulesTableView columnWithIdentifier:columnId];
	if (columnIndex < 0)
		return;
	NSUInteger num = [_rulesSortColumnIndexes count];
	for (NSUInteger i = 0; i < num; i++) {
		if ([_rulesSortColumnIndexes[i] intValue] / 2 != columnIndex)
			continue;
		if (!updating)
			return; // already appended
		[_rulesSortColumnIndexes removeObjectAtIndex:i];
		break;
	}
	NSNumber* newMagic = @(columnIndex * 2 + ascending);
	if (updating) {
		[_rulesSortColumnIndexes insertObject:newMagic atIndex:0];
	} else {
		[_rulesSortColumnIndexes addObject:newMagic];
		return;
	}
	[_prefs removeElementsForXPath:@"iu/filters/rule-column"];
	NSXMLElement* parentE = [_prefs singletonElementForXPath:@"iu/filters"];
	NSArray* cols = [_rulesTableView tableColumns];
	for (NSUInteger i = 0; i < num; i++) {
		int v = [_rulesSortColumnIndexes[i] intValue];
		NSXMLElement* e = [Xml createElementWithName:@"rule-column"];
		NSTableColumn* col = cols[v/2];
		[Xml setAttribute:[col identifier] withName:@"column-id" inElement:e];
		if (v & 1)
			[Xml setAttribute:@"1" withName:@"ascending" inElement:e];
		[parentE addChild:e];
	}
	[_prefs markAsDirty];
}

- (void)restoreRuleSorting
{
	_rulesSortColumnIndexes = [NSMutableArray arrayWithCapacity:10];
	for (NSXMLElement* e in [_prefs nodesForXPath:@"iu/filters/rule-column"]) {
		NSString* columnId = [Xml getAttribute:e withName:@"column-id"];
		if (![columnId length])
			continue;
		BOOL ascending = [Xml getBoolAttribute:e withName:@"ascending"];
		[self addSortingColumn:columnId isAscending:ascending forUpdate:NO];
	}
	[self addSortingColumn:@"host" isAscending:YES forUpdate:NO]; // default to host if list is empty.
	NSArray* cols = [_rulesTableView tableColumns];
	for (NSUInteger idx = 0; idx < [cols count]; idx++) {
		NSTableColumn* col = cols[idx];
		[self addSortingColumn:[col identifier] isAscending:YES forUpdate:NO];
	}
	[self updateRuleSortingIcon];
}

- (void)restoreFilterSelection
{
	[self restoreRuleSorting];
	NSUInteger num = [_filterVertViewFlow rowCount];
	if (!num)
		return;
	NSInteger filterId = [_prefs intForXPath:@"iu/filters/@selected"];
	if (filterId) {
		for (NSUInteger i = 0; i < num; i++) {
			Filter* filter = [_filterVertViewFlow rowByIndex:i];
			if ([filter filterId] == filterId) {
				[_filterVertViewFlow setSelectionByIndex:i];
				return;
			}
		}
	}
	[_filterVertViewFlow setSelectionByIndex:0];
}

- (void)postInit
{
	if (_hasPostInited || ![_comm isServerRunning])
		return;
	_hasPostInited = YES;
	[_filterVertViewFlow setFreezeUpdates:YES];
	for (NSXMLElement* filterE in [[_prefs filtersE] elementsForName:@"filter"])
		[self createFilterForExistingElement:filterE];
	[_filterVertViewFlow setFreezeUpdates:NO];
	[_rulesTableView setDelegate:self];
	[_rulesTableView setDataSource:self];
	[_rulesTableView setTarget:self];
	[_rulesTableView setDoubleAction:@selector(ruleEditAction:)];
	[_rulesTableView reloadData];
	[self restoreFilterSelection];
#ifdef DEBUG
	[_subAddUrlTextField setStringValue:@"http://glimmerblocker.org/site/filters/Test.xml"];
#else
	[_subAddUrlTextField setStringValue:@""];
#endif
}

- (void)dealloc
{
	DebugNSLog(@"FilterTabController.dealloc");
	[self releaseUpdateWithCancel:YES];
	[_filterVertViewFlow setFreezeUpdates:YES];
	[_filterVertViewFlow setDelegate:NULL];
	//
	//
	[_subView removeFromSuperview];
	_tabView = NULL; // no retain: owned by superview
	_filtersTabViewItem = NULL; // no retain: owned by superview
}

#pragma mark ------------------------ filters

- (void)vertPatchTableChangedSelection:(id)sender
{
	[self updateTabViewWithForceUpdate:YES];
	Filter* filter = [self selectedFilter];
	if (!filter)
		return;
	if (![_prefs hasGlimmerAuth])
		return;
	[_prefs setInt:[filter filterId] forXPath:@"iu/filters/@selected"];
	[_prefs markAsDirty];
}

- (void)redrawRulesTable
{
	[_rulesTableView reloadData];
}

- (void)updateTabViewWithForceUpdate:(BOOL)forceUpdate
{
	BOOL hasAuth = [_prefs hasGlimmerAuth];
	Filter* filter = [self selectedFilter];
	[_filterAddBtn setEnabled:hasAuth];
	[_filterRemoveBtn setEnabled:hasAuth && [filter canDelete]];
	[_filterActionBtn setEnabled:hasAuth];
	[_filterCommentsTV setEditable:hasAuth && [filter canEditComments]];
	[_filterCommentsTV setDrawsBackground:[filter canEditComments]];
	[_filterCommentsTV setBordered:[filter canEditComments]];
	[_ruleAddBtn setEnabled:hasAuth && [filter canEditRules]];
	[_ruleRemoveBtn setEnabled:hasAuth && [filter canEditRules] && [_rulesTableView selectedRow] >= 0];
	[self loadRulesWithForceUpdate:forceUpdate];
	if (!filter) {
		[_filterCommentsLabelTV setStringValue:@"Comments for selected filter:"];
		[_filterCommentsTV setStringValue:@""];
		[_rulesTableHeader setStringValue:@"Rules in selected filter:"];
		return;
	}
	//
	[_filterCommentsTV setStringValue:([filter canEditComments] ? [filter comments] : [filter description])];
	if ([filter canEditComments])
		[_filterCommentsLabelTV setStringValue:[NSString stringWithFormat:@"Your comments for this filter:"]];
	else if ([[filter description] length])
		[_filterCommentsLabelTV setStringValue:[NSString stringWithFormat:@"Description:"]];
	else
		[_filterCommentsLabelTV setStringValue:@""];
	NSString* s = [filter enabled] ? @"Rules in “%@”:" : @"Rules in disabled filter “%@”:";
	[_rulesTableHeader setStringValue:[NSString stringWithFormat:s, [filter name]]];
}

- (void)loadRulesWithForceUpdate:(BOOL)forceUpdate
{
	Filter* filter = [self selectedFilter];
	if (!filter) {
		if (!_rulesOwnerFilter)
			return;
		DebugNSLog(@"Clears rules as no filter is selected anymore.");
		[_ruleElementArray removeAllObjects];
		[_rulesTableView reloadData];
		_rulesOwnerFilter = NULL;
		return;
	}
	BOOL hideRules = [filter isSubscription] && ![_comm isServerRunning];
	[_rulesWithoutServer setHidden:!hideRules];
	[_rulesTableHeader setHidden:hideRules];
	[_rulesTableScrollView setHidden:hideRules];
	[_ruleAddBtn setHidden:hideRules];
	[_ruleRemoveBtn setHidden:hideRules];
	if (filter == _rulesOwnerFilter && !forceUpdate)
		return;
	//DebugNSLog(@"Reloads rules, selected filter = %@", [filter name]);
	[_ruleElementArray removeAllObjects];
	_rulesOwnerFilter = filter;
	[_ruleElementArray setArray:[[filter filterDataE] elementsForName:@"rule"]];
	[self sortRules];
	[_rulesTableView reloadData];
	// Sometimes the 'enabled' checkbox in the first row isn't updated/redrawn when selecting another filter.
	// So do one more update async.
	[_rulesTableView performSelectorOnMainThread:@selector(reloadData)
									 withObject:NULL
								  waitUntilDone:NO];
	NSString* selectedRules = [Xml getAttribute:[_rulesOwnerFilter filterE] withName:@"selected-rules"];
	NSMutableSet* ruleSet = [NSMutableSet setWithCapacity:5];
	for (NSString* s in [selectedRules componentsSeparatedByString:@","])
		[ruleSet addObject:@([s intValue])];
	NSUInteger num = [_ruleElementArray count];
	NSMutableIndexSet* indexes = [[NSMutableIndexSet alloc] init];
	for (NSUInteger i = 0; i < num; i++) {
		NSXMLElement* ruleE = _ruleElementArray[i];
		NSNumber* key = @([FilterRuleEdit getRuleId:ruleE]);
		if ([ruleSet containsObject:key])
			[indexes addIndex:i];
	}
	[_rulesTableView selectRowIndexes:indexes byExtendingSelection:NO];
	[_rulesTableView scrollRowToVisible:[indexes firstIndex]];
}

- (BOOL)hasSubscription
{
	NSUInteger num = [_filterVertViewFlow rowCount];
	for (NSUInteger i = 0; i < num; i++) {
		Filter* filter = [_filterVertViewFlow rowByIndex:i];
		if ([filter isSubscription])
			return YES;
	}
	return NO;
}

- (Filter*)createFilterForExistingElement:(NSXMLElement*)filterE atIndex:(NSUInteger)idx
{
	Filter* filter = [[Filter alloc] initWithFilterElement:filterE
												  withComm:_comm
										 withTabController:self
										withFilterListView:_filterVertViewFlow
												   atIndex:idx];
	return filter;
}

// The element must already have been added to the DOM.
- (Filter*)createFilterForExistingElement:(NSXMLElement*)filterE
{
	NSUInteger idx = 0;
	for (NSXMLNode* e = [filterE previousSibling]; e; e = [e previousSibling]) {
		if ([e kind] == NSXMLElementKind)
			idx++;
	}
	return [self createFilterForExistingElement:filterE atIndex:idx];
}

- (Filter*)createFilterForNewElement:(NSXMLElement*)filterE
{
	NSXMLElement *filtersE = [_prefs filtersE];
	Filter* selectedFilter = [_filterVertViewFlow lastSelectedRow];
	// Cocoa DOM doesn't have insertBefore but only insertChild:atIndex
	NSUInteger idx = selectedFilter ? [[filtersE children] indexOfObject:[selectedFilter filterE]] : NSNotFound;
	if (idx == NSNotFound)
		[filtersE addChild:filterE];
	else if (idx >= [filtersE childCount])
		[filtersE addChild:filterE];
	else
		[filtersE insertChild:filterE atIndex:idx + 1];
	Filter* filter = [self createFilterForExistingElement:filterE];
	[_filterVertViewFlow setSelectionByView:filter];
	if ([filter canEditName])
		[filter startEditingName];
	return filter;
}

- (BOOL)hasFilterWithName:(NSString*)name
{
	for (Filter* filter in [_filterVertViewFlow rows]) {
		if ([[filter name] isEqual:name])
			return YES;
	}
	return NO;
}

- (Filter*)createNewUntitledFilter
{
	NSString* prefix = @"My untitled filter";
	NSString* name = prefix;
	if ([self hasFilterWithName:name]) {
		for (int i = 2; YES; i++) {
			name = [NSString stringWithFormat:@"%@-%d", prefix, i];
			if (![self hasFilterWithName:name])
				break;
		}
	}
	NSXMLElement *filterE = [Xml createElementWithName:@"filter"];
	[Xml setInt:[_prefs makeUniqueId] withName:@"id" inElement:filterE];
	NSXMLElement *filterDataE = [Xml createElementWithName:@"filter-data"];
	[filterE addChild:filterDataE];
	//
	[Xml setAttribute:name withName:@"name" inElement:filterDataE];
	[Xml setAttribute:@"1" withName:@"enabled" inElement:filterE];
	//
	return [self createFilterForNewElement:filterE];
	// don't save filter until changed.
}

- (IBAction)filterAddAction:(id)sender
{
	// If option-key is down, duplicate current filter instead of adding new empty filter.
	NSEvent* event = [NSApp currentEvent];
	NSUInteger modifiers = [event modifierFlags] & NSDeviceIndependentModifierFlagsMask;
	if (modifiers == NSAlternateKeyMask) {
		[self filterActionDuplicateAction:sender];
		return;
	}
	[self createNewUntitledFilter];
}

- (IBAction)filterActionDuplicateAction:(id)sender
{
	Filter* srcFilter = [self selectedFilter];
	if (!srcFilter) {
		NSLog(@"filterActionDuplicateAction without selected filter");
		return;
	}
	NSXMLElement *srcFilterDataE = [srcFilter filterDataE];
	if (!srcFilterDataE) {
		[_alertHelper prepareCriticalAlertWithTitle:@"Can't duplicate subscriptions when GlimmerBlocker is not running"
								withInformativeText:@"You must activate GlimmerBlocker before duplicating the filter"];
		[_alertHelper showPreparedAlert];
		return;
	}
	NSXMLElement *newFilterE = [Xml createElementWithName:@"filter"];
	[Xml setInt:[_prefs makeUniqueId] withName:@"id" inElement:newFilterE];
	NSXMLElement *newFilterDataE = [Xml createElementWithName:@"filter-data"];
	[newFilterE addChild:newFilterDataE];
	//
	NSString* name = [NSString stringWithFormat:@"%@ copy", [srcFilter name]];
	[Xml setAttribute:name withName:@"name" inElement:newFilterDataE];
	//
	NSString* comments = [srcFilter description];
	[Xml setAttribute:comments withName:@"comments" inElement:newFilterE];
	//
	[Xml setAttribute:[srcFilter enabled] ? @"1" : @"" withName:@"enabled" inElement:newFilterE];
	//
	for (NSXMLElement* ruleE in [[srcFilter filterDataE] elementsForName:@"rule"])
		[newFilterDataE addChild:[ruleE copy]];
	//
	[self createFilterForNewElement:newFilterE];
}

- (Filter*)selectedFilter
{
	return [_filterVertViewFlow firstSelectedRow];
}

- (Filter*)selectFilterById:(NSInteger)filterId
{
	for (Filter* filter in [_filterVertViewFlow rows]) {
		if ([filter filterId] == filterId) {
			[_filterVertViewFlow setSelectionByView:filter];
			return filter;
		}
	}
	return NULL;
}

- (Filter*)getFilterById:(NSInteger)filterId
{
	for (Filter* filter in [_filterVertViewFlow rows]) {
		if ([filter filterId] == filterId)
			return filter;
	}
	return NULL;
}

- (void)confirmFilterRemoveWithReturnCode:(NSInteger)returnCode
{
	if (returnCode != NSAlertFirstButtonReturn)
		return;
	NSUInteger idx = [[self selectedFilter] rowNumber];
	[[self selectedFilter] deleteFilter];
	if (idx < [_filterVertViewFlow rowCount])
		[_filterVertViewFlow setSelectionByIndex:idx];
	else if ([_filterVertViewFlow rowCount])
		[_filterVertViewFlow setSelectionByIndex:idx - 1];
	else
		[self updateTabViewWithForceUpdate:YES];
	[_comm saveDirtyPrefsAndNotify];
}

- (IBAction)filterRemoveAction:(id)sender
{
	[_alertHelper prepareWarningAlertWithTitle:@"Remove filter?"
							withInformativeText:[[self selectedFilter] name]];
	[_alertHelper addButtonWithTitle:@"OK"];
	[[_alertHelper addButtonWithTitle:@"Cancel"] setKeyEquivalent:@"\e"];
	[_alertHelper setDelegate:self
				 withSelector:@selector(confirmFilterRemoveWithReturnCode:)];
	[_alertHelper showPreparedAlert];
}

- (IBAction)filterEditAction:(id)sender
{
	if ([[self selectedFilter] isSubscription])
		[[self selectedFilter] startEditingSubscription:NO];
}

#pragma mark ------------------------ rules

- (BOOL)tableView:(NSTableView *)tableView shouldEditTableColumn:(NSTableColumn *)col row:(NSInteger)row
{
	if (tableView == _rulesTableView) {
		Filter* filter = [self selectedFilter];
		return filter && [filter canEditRules];
	}
	NSAssert(NO, @"Unknown table in shouldEditTableColumn");
	return NO;
}

- (void)setSavedRuleSelection:(NSArray*)ruleList
{
	NSMutableString* sb = [NSMutableString stringWithCapacity:(10 * [ruleList count])];
	for (NSXMLElement* ruleE in ruleList) {
		if ([sb length])
			[sb appendString:@","];
		[sb appendFormat:@"%d", [FilterRuleEdit getRuleId:ruleE]];
	}
	[Xml setAttribute:sb withName:@"selected-rules" inElement:[_rulesOwnerFilter filterE]];
	[_prefs markAsDirty];
}

- (BOOL)deleteRowData:(NSTableView *)tableView forRow:(NSUInteger)row
{
	if (tableView == _rulesTableView) {
		NSXMLElement* ruleE = _ruleElementArray[row];
		[ruleE detach];
		[_ruleElementArray removeObjectAtIndex:row];
		[_rulesOwnerFilter notifyDirty];
		return YES;
	}
	NSAssert(NO, @"Unknown table in deleteRowData");
	return NO;
}

- (IBAction)filterActionAction:(id)sender
{
	DebugNSLog(@"filterActionAction, t9DuplicateMI = %@, t9AddByUrlMI = %@", _filterDuplicateMI, _filterAddByUrlMI);
	[_filterActionMenu setAutoenablesItems:NO];
	NSRect frame = [_filterActionBtn frame];
	frame.origin.x = 0.0f;
	frame.origin.y = 0.0f;
	NSPopUpButtonCell* cell = [[NSPopUpButtonCell alloc] initTextCell:@"" pullsDown:YES];
	[cell setMenu:_filterActionMenu];
	Filter* filter = [self selectedFilter];
	[_filterDuplicateMI setEnabled:!!filter];
	[_filterEditMI setEnabled:[filter isSubscription]];
	[_filterUpdateSelectedMI setEnabled:[filter isSubscription]];
	[_filterUpdateAllMI setEnabled:[self hasSubscription]];
	if (filter && ![filter isSubscription]) {
		[_filterPublishWebDavMI setEnabled:![filter isMobileMePublished]];
	} else {
		[_filterPublishWebDavMI setEnabled:NO];
	}
	[cell performClickWithFrame:frame inView:_filterActionBtn];
}

- (IBAction)filterCommentsAction:(id)sender
{
	Filter* filter = [self selectedFilter];
	if (!filter || ![filter canEditComments])
		return;
	[filter setComments:[_filterCommentsTV stringValue]];
	[_comm saveDirtyPrefsAndNotify];
}

- (void)ruleEditCommit:(NSXMLElement*)ruleE
		  withFilter:(Filter*)filter
				isVirgin:(BOOL)virgin
{
	[_filterVertViewFlow setSelectionByView:filter];
	[self updateTabViewWithForceUpdate:YES]; // force to get sorting updated.
	NSInteger idx = [_ruleElementArray indexOfObject:ruleE];
	if (idx == NSNotFound) {
		NSLog(@"Added rule disappeared from _ruleElementArray");
	} else {
		NSIndexSet* indexes = [NSIndexSet indexSetWithIndex:idx];
		[_rulesTableView selectRowIndexes:indexes byExtendingSelection:NO];
		[_rulesTableView scrollRowToVisible:idx];
	}
	// Select the tab views we're inside.
	[_tabView selectTabViewItem:_filtersTabViewItem];
	[filter notifyDirty];
	[_prefs saveIfDirty:self];
	[_comm sendReloadPreferencesForModifiedFilter:[filter filterId] withRule:[FilterRuleEdit getRuleId:ruleE]];
	[[_tabView window] performSelectorOnMainThread:@selector(makeFirstResponder:)
										withObject:_rulesTableView
									 waitUntilDone:NO];
}

- (IBAction)ruleAddAction:(id)sender
{
	Filter* filter = [self selectedFilter];
	if (!filter) {
		NSLog(@"ruleAddAction without selected filter");
		return;
	}
	// If option-key is down, duplicate current rule instead of adding new empty rule.
	NSEvent* event = [NSApp currentEvent];
	NSUInteger modifiers = [event modifierFlags] & NSDeviceIndependentModifierFlagsMask;
	if (modifiers == NSAlternateKeyMask) {
		[self ruleActionDuplicateAction:sender];
		return;
	}
	FilterRuleEdit *editor = [[FilterRuleEdit alloc] initWithFilterTabController:self
										 withOwnerFilter:filter
										 withRuleElement:NULL
									  withSuspectElement:NULL];
	[AutoZap add:editor];
}

- (IBAction)ruleRemoveAction:(id)sender
{
	if ([_rulesTableView selectedRow] < 0)
		return;
	[self removeSelectedRows:_rulesTableView];
	[_comm saveDirtyPrefsAndNotify];
}

- (IBAction)ruleActionAction:(id)sender
{
	[_ruleActionMenu setAutoenablesItems:NO];
	NSRect frame = [_ruleActionBtn frame];
	frame.origin.x = 0.0f;
	frame.origin.y = 0.0f;
	NSPopUpButtonCell* cell = [[NSPopUpButtonCell alloc] initTextCell:@"" pullsDown:YES];
	[cell setMenu:_ruleActionMenu];
	NSInteger row = [_rulesTableView selectedRow];
	[_ruleDuplicateMI setEnabled:(row >= 0 && ![_rulesOwnerFilter isSubscription])];
	[cell performClickWithFrame:frame inView:_ruleActionBtn];
}

- (IBAction)ruleActionDuplicateAction:(id)sender
{
	NSIndexSet* set = [_rulesTableView selectedRowIndexes];
	for (NSInteger row = [set lastIndex]; row != NSNotFound; row = [set indexLessThanIndex:row]) {
		NSXMLElement* oldRuleE = _ruleElementArray[row];
		NSXMLElement* parentE = (NSXMLElement*)[oldRuleE parent];
		NSXMLElement* newRuleE = [oldRuleE copy];
		NSInteger idx = [[parentE children] indexOfObject:oldRuleE];
		[parentE insertChild:newRuleE atIndex:idx];
		[Xml setAttribute:NULL withName:@"rule-id" inElement:newRuleE]; // remove id, so it gets a new one.
		[FilterRuleEdit getRuleId:newRuleE]; // ensures it has a rule-id
	}
	[self updateTabViewWithForceUpdate:YES];
}

- (void)ruleEditAction:(id)sender
{
	NSInteger row = [_rulesTableView clickedRow];
	if (row < 0)
		return;
	Filter* filter = [self selectedFilter];
	if (!filter) {
		NSLog(@"ruleEditAction without selected filter");
		return;
	}
	FilterRuleEdit *editor = [[FilterRuleEdit alloc] initWithFilterTabController:self
																 withOwnerFilter:filter
																 withRuleElement:_ruleElementArray[row]
															  withSuspectElement:NULL];
	[AutoZap add:editor];
}

#pragma mark ------------------------ add subscription

- (void)subAddConfirmRedisplaySheet:(int)returnCode
{
	[self filterActionSubscribeAction:self];
}

- (void)showSimpleAddSubscriptionNetworkError:(NSString*)msg
{
	[_alertHelper dismissAlert];
	[_alertHelper prepareCriticalAlertWithTitle:@"Could not get subscription"
							withInformativeText:msg];
	[_alertHelper setDelegate:self
				 withSelector:@selector(subAddConfirmRedisplaySheet:)];
	[_alertHelper showPreparedAlert];
}

- (IBAction)filterActionSubscribeAction:(id)sender
{
	if (![_comm isServerRunning]) {
		[_alertHelper prepareCriticalAlertWithTitle:@"Can't add subscriptions when GlimmerBlocker is not running"
								withInformativeText:@"You must check activate GlimmerBlocker to add subscriptions."];
		[_alertHelper showPreparedAlert];
		return;
	}
	_addSubFilterId = [_prefs makeUniqueId];
	[NSApp beginSheet: _subAddSheet
	   modalForWindow: [_subView window]
		modalDelegate: nil
	   didEndSelector: nil
		  contextInfo: nil];
}

- (IBAction)subAddCancelAction:(id)sender
{
	[NSApp endSheet:_subAddSheet];
	[_subAddSheet orderOut:self];
}

- (void)cancelledAddSubscription:(int)returnCode
{
	DebugNSLog(@"cancelledAddSubscription called");
	[self releaseUpdateWithCancel:YES];
}

- (IBAction)subAddConfirmAction:(id)sender
{
	[NSApp endSheet:_subAddSheet];
	[_subAddSheet orderOut:self];
	NSString* s = [_subAddUrlTextField stringValue];
	while ([s length] && [s characterAtIndex:0] <= 32)
		s = [s substringFromIndex:1];
	while ([s length] && [s characterAtIndex:[s length] - 1] <= 32)
		s = [s substringToIndex:[s length] - 1]; // index included
	if (![s length]) {
		DebugNSLog(@"subAddConfirmAction: empty text field.");
		return;
	}
	NSString* urlSyntaxError = [Xml getHttpUrlSyntaxError:s];
	if (!urlSyntaxError && ![s hasSuffix:@".xml"])
		urlSyntaxError = @"The URL must end with '.xml'";
	if (urlSyntaxError) {
		[_alertHelper prepareCriticalAlertWithTitle:@"The URL you entered is not valid"
								withInformativeText:urlSyntaxError];
		[_alertHelper setDelegate:self
					 withSelector:@selector(subAddConfirmRedisplaySheet:)];
		[_alertHelper showPreparedAlert];
		return;
	}
	DebugNSLog(@"add subscription url: %@", s);
	[_alertHelper prepareInformativeAlertWithTitle:@"Adding subscription" withInformativeText:@"Downloading"];
	[_alertHelper setDelegate:self withSelector:@selector(cancelledAddSubscription:)];
	[[_alertHelper addButtonWithTitle:@"Cancel"] setKeyEquivalent:@"\e"];
	NSProgressIndicator* spinner = [[NSProgressIndicator alloc] initWithFrame:NSMakeRect(0, 0, 440, 20)];
	[spinner setStyle:NSProgressIndicatorBarStyle];
	[spinner setIndeterminate:YES];
	[spinner setDisplayedWhenStopped:NO];
	[_alertHelper setAccessoryView:spinner];
	[_alertHelper showPreparedAlert];
	[spinner startAnimation:self];
	//
	[self releaseUpdateWithCancel:NO];
	_addSubPathArgs = [NSString stringWithFormat:@"!/filters/subscription/add?url=%@&filter-id=%d",
						[GBDebug percentEscapeUrlParameterValue:s],
						_addSubFilterId];
	_addSubUrl = [NSURL URLWithString:[_comm getAdminUrl:_addSubPathArgs]];
	NSURLRequest* req = [NSURLRequest requestWithURL:_addSubUrl
										 cachePolicy:NSURLRequestReloadIgnoringLocalCacheData // want fresh data
									 timeoutInterval:60.0];
	_addSubData = [NSMutableData data];
	_addSubConnection = [[NSURLConnection alloc] initWithRequest:req delegate:self];
	if (!_addSubConnection) {
		[self releaseUpdateWithCancel:NO];
		[self showSimpleAddSubscriptionNetworkError:NULL];
		return;
	}
}

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response
{
	// called for each response, i.e. for each redirect
	[_addSubData setLength:0];
#ifdef DEBUG
	NSHTTPURLResponse* http = (NSHTTPURLResponse*)response;
	DebugNSLog(@"didReceiveResponse: %ld   %@", [http statusCode], [http allHeaderFields]);
#endif
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data
{
	[_addSubData appendData:data];
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error
{
	NSLog(@"Add subscription request failed: %@", error);
	NSString* a = [error localizedDescription];
	NSString* b = [error userInfo][NSURLErrorFailingURLStringErrorKey];
	NSString* updateError = [NSString stringWithFormat:@"%@ %@", a, b];
	[self showSimpleAddSubscriptionNetworkError:updateError];
	[self releaseUpdateWithCancel:NO];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection
{
	DebugNSLog(@"Got connectionDidFinishLoading, result size = %ld bytes", [_addSubData length]);
	[_alertHelper dismissAlert];
	XmlDoc* doc = [[XmlDoc alloc] initWithURL:_addSubUrl withData:_addSubData];
	[self releaseUpdateWithCancel:NO];
	if (![doc rootE]) {
		[self showSimpleAddSubscriptionNetworkError:@"Could not get proper xml response from server"];
		return;
	}
	NSString* error = [Xml getAttribute:[doc rootE] withName:@"error"];
	if (error) {
		[self showSimpleAddSubscriptionNetworkError:error];
		return;
	}
	NSString* pathArgs = [_addSubPathArgs stringByAppendingString:@"&commit=1"];
	doc = [_comm curlToXml:pathArgs withActionDescription:@"Could not add subscription"];
	if (!doc) {
		NSLog(@"AddSub: failed getting doc");
		return;
	}
	NSXMLElement* filterContainerE = [doc byXPath:@"filter-container"];
	if (!filterContainerE) {
		[_alertHelper prepareCriticalAlertWithTitle:@"Failed getting subscription"
								withInformativeText:@"Missing the <filter-container> element"];
		[_alertHelper setDelegate:self
					 withSelector:@selector(subAddConfirmRedisplaySheet:)];
		[_alertHelper showPreparedAlert];
		return;
	}
	NSXMLElement* filterDataE = [doc byXPath:@"filter-container/filter-data"];
	if (!filterDataE) {
		[_alertHelper prepareCriticalAlertWithTitle:@"Failed getting subscription"
								withInformativeText:@"Missing the <filter-data> element"];
		[_alertHelper setDelegate:self
					 withSelector:@selector(subAddConfirmRedisplaySheet:)];
		[_alertHelper showPreparedAlert];
		return;
	}
	DebugNSLog(@"AddSub success.");
	NSXMLElement* filterE = [Xml createElementWithName:@"filter"];
	[Xml setAttribute:[Xml getAttribute:filterContainerE withName:@"url"] withName:@"url" inElement:filterE];
	[Xml setAttribute:@"1" withName:@"enabled" inElement:filterE];
	[Xml setInt:_addSubFilterId withName:@"id" inElement:filterE];
	[Xml setAttribute:updatePolicies[[_subAddUpdatesMatrix selectedRow]] withName:@"update-policy" inElement:filterE];
	[Xml setAttribute:safetyPolicies[[_subAddSafetyMatrix selectedRow]] withName:@"safety-policy" inElement:filterE];
	[self createFilterForNewElement:filterE];
	DebugNSLog(@"AddSub: Saves settings");
	[_comm saveDirtyPrefsAndNotify];
}

- (void)releaseUpdateWithCancel:(BOOL)cancel
{
	if (!_addSubUrl)
		return;
	DebugNSLog(@"#### releaseUpdateWithCancel, data size = %ld", [_addSubData length]);
	if (cancel)
		[_addSubConnection cancel];
	_addSubConnection = NULL;
	_addSubData = NULL;
	_addSubUrl = NULL;
}

#pragma mark ------------------------ update subscription

- (IBAction)updateSelectedSubscriptionAction:(id)sender
{
	DebugNSLog(@"updateSelectedSubscriptionAction");
	Filter* filter = [self selectedFilter];
	if ([filter isSubscription])
		[filter startUpdateSubscription];
	else
		NSLog(@"Selected filter is not a subscription");
}

- (IBAction)updateAllSubscriptionsAction:(id)sender
{
	DebugNSLog(@"updateAllSubscriptionsAction");
	NSUInteger num = [_filterVertViewFlow rowCount];
	for (NSUInteger i = 0; i < num; i++) {
		Filter* filter = [_filterVertViewFlow rowByIndex:i];
		if ([filter isSubscription])
			[filter startUpdateSubscription];
	}
}

- (void)notifyUpdatedFilter:(Filter*)filter
{
	if (filter == [self selectedFilter])
		[self updateTabViewWithForceUpdate:YES];
}

#pragma mark ------------------------ Publish

- (IBAction)publishWebDavAction:(id)sender
{
	Filter* filter = [self selectedFilter];
	if (!filter)
		return;
	FilterPublisher *publisher = [[FilterPublisher alloc] initWithFilterTabController:self
																		   withFilter:filter];
	[AutoZap add:publisher];
}

#pragma mark ------------------------ Utils

- (NSArray*)filterArray
{
	return [_filterVertViewFlow rows];
}

- (NSArray*)ruleElements
{
	return [NSArray arrayWithArray:_ruleElementArray];
}

- (NSBundle*)mainBundle
{
	return [_comm mainBundle];
}

- (void)notifyUpdatedAuthorizationState
{
	[self updateTabViewWithForceUpdate:NO];
	for (Filter* filter in [_filterVertViewFlow rows])
		[filter notifyUpdatedAuthorizationState];
	[_rulesTableView reloadData];
}

#pragma mark ------------------------ Sorting

+ (NSString*)ruleActionText:(NSXMLElement*)ruleE
{
	switch ([Rule getRuleType:ruleE]) {
		case BlockRuleType:
			return @"block";
		case WhitelistRuleType:
			return @"whitelist";
		case ModifyErrorStatusRuleType:
			return @"fix!ok";
		case ModifyContentRuleType:
			return @"fix";
		case KeywordRuleType:
			return @"keyword";
		case RequestRuleType:
			return @"request";
		default:
			return @"???";
	}
}

static NSString* dotReverse(NSString* s)
{
	NSArray* a = [s componentsSeparatedByString:@"."];
	// missing reverse method.
	NSMutableString* sb = [NSMutableString stringWithCapacity:[s length]];
	for (NSUInteger i = [a count] - 1; true; i--) {
		[sb appendString:a[i]];
		if (i > 0)
			[sb appendString:@"."];
		else
			break;
	}
	return sb;
}

static int getPriority(NSXMLElement* ruleE)
{
	if (!ruleE)
		return 9;
	NSString* s = [Xml getAttribute:ruleE withName:@"priority"];
	return s ? [s intValue] : 2;
}

+ (NSString*)getHostSortingKey:(NSXMLElement*)ruleE
{
	NSString* s = [Xml getAttribute:ruleE withName:@"host"];
	if (s)
		return dotReverse(s);
	s = [Xml getAttribute:ruleE withName:@"keyword"];
	if (s)
		return [@" 0 " stringByAppendingString:s]; // make sort before normal rules.
	return @" 1";
}

- (NSInteger)compareRule:(NSXMLElement*)rule1E toRule:(NSXMLElement*)rule2E
{
	for (NSNumber* m in _rulesSortColumnIndexes) {
		NSInteger diff = 0;
		NSString* a;
		NSString* b;
		NSNumber* n1;
		NSNumber* n2;
		switch ([m intValue] / 2) {
			case 0: // enable-overwrite
				n1 = [_rulesOwnerFilter getRuleSubscriptionState:[Xml getAttribute:rule1E withName:@"rule-id"]];
				n2 = [_rulesOwnerFilter getRuleSubscriptionState:[Xml getAttribute:rule2E withName:@"rule-id"]];
				diff = [n1 intValue] - [n2 intValue];
				break;
			case 1: // priority
				diff = getPriority(rule1E) - getPriority(rule2E);
				break;
			case 2: // action
				a = [FilterTabController ruleActionText:rule1E];
				b = [FilterTabController ruleActionText:rule2E];
				diff = [a compare:b];
				break;
			case 3: // host
				a = [FilterTabController getHostSortingKey:rule1E];
				b = [FilterTabController getHostSortingKey:rule2E];
				diff = [a caseInsensitiveCompare:b];
				break;
			case 4: // path
				a = soft([Xml getAttribute:rule1E withName:@"path"]);
				b = soft([Xml getAttribute:rule2E withName:@"path"]);
				diff = [a caseInsensitiveCompare:b];
				break;
			case 5: // comments
				a = soft([Xml stringForXPath:@"comments" withRoot:rule1E]);
				b = soft([Xml stringForXPath:@"comments" withRoot:rule2E]);
				diff = [a caseInsensitiveCompare:b];
		}
		if (diff)
			return ([m intValue] & 1) ? diff : -diff;
	}
	return 0;
}


static NSInteger compareRules(id rule1E, id rule2E, void *context)
{
	FilterTabController* self = (__bridge FilterTabController*)context;
	return [self compareRule:(NSXMLElement *)rule1E toRule:(NSXMLElement *)rule2E];
}

- (void)sortRules
{
	//DebugNSLog(@"sortRules: %@", _rulesSortColumnIndexes);
	NSArray* a = [_ruleElementArray sortedArrayUsingFunction:&compareRules context:(__bridge void *)(self)];
	[_ruleElementArray setArray:a];
	[_rulesTableView reloadData];
}

#pragma mark ------------------------ Table Delegate


- (void)tableViewSelectionDidChange:(NSNotification *)aNotification
{
	[self updateTabViewWithForceUpdate:NO];
	if (!_rulesOwnerFilter)
		return;
	NSIndexSet* set = [_rulesTableView selectedRowIndexes];
	NSMutableArray* a = [NSMutableArray arrayWithCapacity:[set count]];
	for (NSInteger row = [set firstIndex]; row != NSNotFound; row = [set indexGreaterThanIndex:row])
		[a addObject:_ruleElementArray[row]];
	[self setSavedRuleSelection:a];
}

- (NSAttributedString*)textWithStrikeThrough:(NSString*)txt
{
	NSDictionary *attrsDictionary =@{NSStrikethroughStyleAttributeName: @1U};
	return [[NSAttributedString alloc] initWithString:txt attributes:attrsDictionary];
}

- (NSString *)tableView:(NSTableView *)aTableView
		 toolTipForCell:(NSCell *)aCell
				   rect:(NSRectPointer)rect
			tableColumn:(NSTableColumn *)col
					row:(NSInteger)row
		  mouseLocation:(NSPoint)mouseLocation
{
	if (aTableView != _rulesTableView || ![_rulesOwnerFilter isSubscription])
		return @"";
	NSString* cid = [col identifier];
	if (![cid isEqual:@"enabled"])
		return @"";
	NSXMLElement* ruleE = _ruleElementArray[row];
	NSString* ruleId = [Xml getAttribute:ruleE withName:@"rule-id"];
	if (!ruleId)
		return @"";
	int state = [[_rulesOwnerFilter getRuleSubscriptionState:ruleId] intValue];
	BOOL enabled = ![Xml getBoolAttribute:ruleE withName:@"disabled"];
	if (state == NSMixedState) {
		if (enabled)
			return @"Using authors default, which is to enable the rule.";
		else
			return @"Using authors default, which is to disable the rule.";
	}
	if (state == NSOffState) {
		if (enabled)
			return @"You have forced the rule to always be disabled and the authors default to enable the rule is ignored.";
		else
			return @"You have forced the rule to always be disabled and the authors default to disable the rule is ignored.";
	} else {
		if (enabled)
			return @"You have forced the rule to always be enabled and the authors default to enable the rule is ignored.";
		else
			return @"You have forced the rule to always be enabled and the authors default to disable the rule is ignored.";
	}
}

#pragma mark ------------------------ SplitView Delegate

- (BOOL)splitView:(NSSplitView *)sender canCollapseSubview:(NSView *)subview
{
	return NO;
}

- (CGFloat)splitView:(NSSplitView *)sender constrainMaxCoordinate:(CGFloat)proposedMax ofSubviewAt:(NSInteger)offset
{
	//DebugNSLog(@"constrainMaxCoordinate: proposedMax = %f,  ofSubviewAt = %d", proposedMax, offset);
	return [sender frame].size.height - 144;
}

- (CGFloat)splitView:(NSSplitView *)sender constrainMinCoordinate:(CGFloat)proposedMin ofSubviewAt:(NSInteger)offset
{
	//DebugNSLog(@"constrainMinCoordinate: proposedMax = %f,  ofSubviewAt = %d", proposedMin, offset);
	return 119;
}

/*
- (CGFloat)splitView:(NSSplitView *)sender constrainSplitPosition:(CGFloat)proposedPosition ofSubviewAt:(NSInteger)offset
{
	NSLog(@"constrainSplitPosition: proposedPosition = %f,  ofSubviewAt = %d", proposedPosition, offset);
	return proposedPosition;
}
*/

@end

@implementation FilterTabController(NSTableDataSource)

- (NSInteger)numberOfRowsInTableView:(NSTableView *)tableView
{
	if (tableView == _rulesTableView)
		return [_ruleElementArray count];
	NSLog(@"numberOfRowsInTableView called for unknown NSTableView: %@", tableView);
	return 0;
}

- (id)tableView:(NSTableView *)tableView objectValueForTableColumn:(NSTableColumn *)col row:(NSInteger)row
{
	NSString* cid = [col identifier];
	if (tableView == _rulesTableView) {
		//DebugNSLog(@"  rule.oVfTc(%@), row %d", cid, row);
		if ((NSUInteger)row >= [_ruleElementArray count]) {
			NSLog(@"Invalid row = %ld in objectValueForTableColumn, has only %ld rules", (long)row, (long)[_ruleElementArray count]);
			return NULL;
		}
		NSXMLElement* ruleE = _ruleElementArray[row];
		NSString* txt;
		if ([cid isEqual:@"enabled"]) {
			NSString* ruleId = [Xml getAttribute:ruleE withName:@"rule-id"];
			if (!_rulesOwnerFilter || !ruleId)
				return @(NSMixedState);
			if ([_rulesOwnerFilter isSubscription])
				return [_rulesOwnerFilter getRuleSubscriptionState:ruleId];
			return @(![Xml getBoolAttribute:ruleE withName:@"disabled"]);
		} else if ([cid isEqual:@"priority"]) {
			return [NSString stringWithFormat:@"%d", getPriority(ruleE)]; // NSAttributedString doesn't work with alignment.
		} else if ([cid isEqual:@"action"]) {
			txt = [FilterTabController ruleActionText:ruleE];
			if ([Xml getBoolAttribute:ruleE withName:@"disabled"])
				return [self textWithStrikeThrough:txt];
		} else if ([cid isEqual:@"host"]) {
			return @"dummy";
		} else if ([cid isEqual:@"path"]) {
			txt = [Xml getAttribute:ruleE withName:@"path"];
		} else if ([cid isEqual:@"comments"]) {
			txt = [Xml stringForXPath:@"comments" withRoot:ruleE];
			if (!txt)
				txt = @"";
			NSRange range = [txt rangeOfCharacterFromSet:[NSCharacterSet newlineCharacterSet]];
			if (range.location != NSNotFound)
				txt = [txt substringToIndex:range.location];
		} else {
			NSLog(@"objectValueForTableColumn: Unknown rule column: %@", cid);
			return NULL;
		}
		if (txt && [_rulesOwnerFilter safetySimpleRulesOnly] && ![Rule ruleIsSafe:ruleE])
			return [self textWithStrikeThrough:txt];
		else
			return txt;
	}
	NSLog(@"objectValueForTableColumn called for unknown NSTableView: %@", tableView);
	return NULL;
}

static NSColor* green;
static NSColor* red;
static NSColor* black;

- (void)tableView:(NSTableView*)aTableView
  willDisplayCell:(id)aCell
   forTableColumn:(NSTableColumn *)aTableColumn
			  row:(NSInteger)row
{
	if (aTableView != _rulesTableView)
		return;
	if ([aCell isKindOfClass:[NSButtonCell class]]) {
		[aCell setEnabled:[_prefs hasGlimmerAuth]];
		[aCell setAllowsMixedState:[_rulesOwnerFilter isSubscription]];
		return;
	}
	if ([aCell isKindOfClass:[HostTextFieldCell class]]) {
		HostTextFieldCell* htfc = (HostTextFieldCell*)aCell;
		NSXMLElement* ruleE = _ruleElementArray[row];
		[htfc setRuleE:ruleE filter:_rulesOwnerFilter];
	}
	if (![aCell isKindOfClass:[NSTextFieldCell class]])
		return;
	if (!green) {
		green = [NSColor colorWithDeviceRed:0 green:0.75f blue:0 alpha:1];
		red = [NSColor redColor];
		black = [NSColor blackColor];
	}
	NSColor* color = NULL;
	if (_rulesOwnerFilter && [_rulesOwnerFilter isSubscription]) {
		NSXMLElement* ruleE = _ruleElementArray[row];
		NSString* ruleId = [Xml getAttribute:ruleE withName:@"rule-id"];
		if ([_rulesOwnerFilter safetySimpleRulesOnly] && ![Rule ruleIsSafe:ruleE])
			color = [NSColor grayColor];
		else if (ruleId) {
			switch ([[_rulesOwnerFilter getRuleSubscriptionState:ruleId] intValue]) {
				case NSOnState:
					color = green;
					break;
				case NSOffState:
					color = red;
					break;
			}
		}
	}
	[aCell setTextColor:(color ? color : black)];
}


- (void)tableView:(NSTableView *)aTableView
   setObjectValue:(id)value
   forTableColumn:(NSTableColumn *)aTableColumn
			  row:(NSInteger)row
{
	//DebugNSLog(@"setObjectValue(%@) for col %@ row %d", value, [aTableColumn identifier], row);
	NSXMLElement* ruleE = _ruleElementArray[row];
	NSString* ruleId = [Xml getAttribute:ruleE withName:@"rule-id"];
	//DebugNSLog(@"  ruleId = %@, owner = %x, class = %d", ruleId, _rulesOwnerFilter, [value isKindOfClass:[NSNumber class]]);
	if (!_rulesOwnerFilter || !ruleId || ![value isKindOfClass:[NSNumber class]])
		return;
	if ([_rulesOwnerFilter isSubscription]) {
		[_rulesOwnerFilter setRuleSubscriptionState:value forRuleId:ruleId];
	} else {
		[Xml setAttribute:([value intValue] ? NULL : @"1") withName:@"disabled" inElement:ruleE];
		[_prefs markAsDirty];
		[_rulesOwnerFilter notifyDirty];
		[_comm saveDirtyPrefsAndNotify];
	}
	[_rulesTableView setNeedsDisplayInRect:[_rulesTableView rectOfRow:row]];
}

- (void)tableView:(NSTableView *)tableView didClickTableColumn:(NSTableColumn *)col
{
	if (tableView != _rulesTableView)
		return;
	BOOL ascending = YES;
	if ([_rulesSortColumnIndexes count]) {
		int v = [_rulesSortColumnIndexes[0] intValue];
		NSTableColumn* currentCol = [_rulesTableView tableColumns][v/2];
		if (currentCol == col && (v & 1))
			ascending = NO;
	}
	[self addSortingColumn:[col identifier] isAscending:ascending forUpdate:YES];
	[self updateRuleSortingIcon];
	[self sortRules];
	[_prefs markAsDirty];
}

- (BOOL)tableView:(NSTableView*)tableView writeRowsWithIndexes:(NSIndexSet*)rowIndexes toPasteboard:(NSPasteboard*)pb
{
	if (tableView != _rulesTableView)
		return NO;
	DebugNSLog(@"writeRowsWithIndexes for rulesTableView");
	// Copy the row numbers to the pasteboard.
	[_rulesTableView copyRows:rowIndexes toPasteboard:pb];
	return YES;
}

- (NSDragOperation)tableView:(NSTableView*)tableView
				validateDrop:(id <NSDraggingInfo>)info
				 proposedRow:(NSInteger)row
	   proposedDropOperation:(NSTableViewDropOperation)operation
{
	if (tableView != _rulesTableView || operation != NSTableViewDropAbove || [info draggingSource] == _rulesTableView)
		return NSDragOperationNone;
	//DebugNSLog(@"validate Drop, info: %@", info);
	if (![_rulesTableView canModifyRules])
		return NSDragOperationNone;
	NSData* data = [_rulesTableView getValidDataFromPasteboard:[info draggingPasteboard]];
	return data ? NSDragOperationCopy : NSDragOperationNone;
}

- (BOOL)tableView:(NSTableView *)tableView
	   acceptDrop:(id <NSDraggingInfo>)info
			  row:(NSInteger)row
	dropOperation:(NSTableViewDropOperation)operation
{
	if (tableView != _rulesTableView || operation != NSTableViewDropAbove || [info draggingSource] == _rulesTableView)
		return NSDragOperationNone;
	DebugNSLog(@"acceptDrop: %@", info);
	if (![_rulesTableView canModifyRules])
		return NSDragOperationNone;
	NSData* data = [_rulesTableView getValidDataFromPasteboard:[info draggingPasteboard]];
	if (!data)
		return NO;
	[_rulesTableView insertRuleData:data];
	return YES;
}

@end
