/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "ServerCommunication.h"
#import "GBDebug.h"
#import "Prefs.h"
#import "XmlDoc.h"
#import "Xml.h"
#import "AlertHelper.h"

#ifdef DEBUG
//#define DEBUG_COMM 1
#endif

static NSString* javaServerLogPath = @"/Library/Logs/GlimmerBlocker/GlimmerBlocker.log";

@interface ServerCommunication()

@property NSBundle* mainBundle;
@property Prefs* prefs;
@property AlertHelper* alertHelper;
@property NSString* publicAdminKey;
@property NSString* secureAdminHost;

@end

@implementation ServerCommunication

- (instancetype)initWithPrefs:(Prefs*)prefs
			   withMainBundle:(NSBundle*)mainBundle
{
	if ((self = [super init])) {
		_mainBundle = mainBundle;
		_prefs = prefs;
		_alertHelper = [prefs alertHelper];
		_publicAdminKey = [_prefs stringForXPath:@"proxy-server/@info-admin-key"];
		if (!_publicAdminKey) {
			NSMutableString *sb = [NSMutableString stringWithCapacity:40];
			for (int i = 0; i < 10; i++)
				[sb appendFormat:@"%04x", arc4random() & 0xFFFF];
			_publicAdminKey = [NSString stringWithString:sb];
			[_prefs setString:_publicAdminKey forXPath:@"proxy-server/@info-admin-key"];
			[_prefs markAsDirty];
			DebugNSLog(@"adminHost: %@", _publicAdminKey);
		}
	}
	return self;
}


- (void)performPostLoadPreferencePanelCheck
{
/*
	if (![self isServerRunning] || ![_launchctl needsUpdatePlist])
		return;
	[_alertHelper prepareCriticalAlertWithTitle:@"GlimmerBlocker needs to be restarted"
							withInformativeText:@"Please deactivate and reactivate GlimmerBlocker."];
	[_alertHelper showPreparedAlert];
*/
}

#pragma mark ------------------------------------- URLs

- (BOOL)getSecureAdminKey
{
	if (_secureAdminHost)
		return YES;
	NSMutableData* data = [NSMutableData dataWithCapacity:200];
	int stat = [_prefs runAuthHelperWithArgs:@[@"--java-settings"]
							   withInputData:NULL
							  withOutputData:data
									withAuth:YES];
	if (stat) {
		DebugNSLog(@"Got error status %d from AuthHelper --java-settings", stat);
		return NO;
	}
	if (![data length]) {
		NSLog(@"##### Didn't get any output from --java-settings");
		return NO;
	}
	NSURL* url = [NSURL URLWithString:@"x-file:java-settings.xml"];
	XmlDoc* doc = [[XmlDoc alloc] initWithURL:url withData:data];
	if (![doc doc] || ![doc rootE]) {
		NSLog(@"#### Can't parse java-settings xml.");
		return NO;
	}
	_secureAdminHost = [doc stringForXPath:@"proxy/@secure-admin-key"];
	DebugNSLog(@"Did get _secureAdminHost");
	return YES;
}

- (NSString*)getAdminUrl:(NSString*)pathArgs
{
	NSString* key = _publicAdminKey;
	if ([pathArgs hasPrefix:@"!"]) {
		pathArgs = [pathArgs substringFromIndex:1];
		if ([self getSecureAdminKey]) {
			key = _secureAdminHost;
		} else {
			NSLog(@"Can't get secure-admin host for: %@", pathArgs);
			key = @"failed-secure-admin";
		}
	}
	return [NSString stringWithFormat:@"http://127.0.0.1:%d/=gb=/%@%@", [self getPortNumber], key, pathArgs];
}

#pragma mark ------------------------------------- updater

- (void)updateUpdaterLaunchd
{
	[self saveDirtyPrefsAndNotify];
	BOOL run = [_prefs boolForXPath:@"updater/@check-in-background"];
	if (!run) {
		NSArray* args = @[@"--updater-stop"];
		[_prefs runAuthHelperWithArgs:(NSArray*)args withInputData:NULL withAuth:YES];
		[GBDebug executeCommandlineTool:@"launchctl" withArgs:@[@"unload", [GBDebug updaterPlistPath]]];
		return;
	}
	NSArray* args = @[@"--updater-start"];
	[_prefs runAuthHelperWithArgs:(NSArray*)args withInputData:NULL withAuth:YES];
	[GBDebug executeCommandlineTool:@"launchctl" withArgs:@[@"unload", [GBDebug updaterPlistPath]]];
	[GBDebug executeCommandlineTool:@"launchctl" withArgs:@[@"load", [GBDebug updaterPlistPath]]];
}

- (void)launchUpdaterApp
{
	NSString *appPath = [GBDebug updaterBundlePath];
	[GBDebug executeCommandlineTool:@"open" withArgs:@[@"-a", appPath]];
}


#pragma mark ------------------------------------- server management
- (BOOL)isServerRunning
{
	// a real check is more expensive.
	NSString* path = @"/Library/LaunchDaemons/org.glimmerblocker.proxy.plist";
	return [[NSFileManager defaultManager] fileExistsAtPath:path];
	// We could do
	//   sudo launchctl list org.glimmerblocker.proxy
	// but the id-param is totally undocumented.
	// So we'd have to check if anything runs at the port,
	// or using a full "launchctl list".
}

- (void)startServer
{
	[_prefs saveIfDirty:self];
	NSArray* args = @[@"--start"];
	[_prefs runAuthHelperWithArgs:(NSArray*)args withInputData:NULL withAuth:YES];
	//
	// The server might take some time, and loading filters will fail with an error dialog if it's
	// not ready in time.
	// So keep waiting a bit until the server responds to http.
	NSString* pathArgs = @"/test-running";
	NSURLResponse *response;
	NSError *__autoreleasing errorRef = NULL;
	int i = 0;
	while (true) {
		if ([self curlToData:pathArgs withResponse:&response withErrorRef:&errorRef]) {
			DebugNSLog(@"Server now ready after %d retries.", i);
			break;
		}
		if (++i > 20) {
			NSLog(@"Proxy server doesn't seem to come online.");
			break;
		}
		[NSThread sleepForTimeInterval:0.25];
	}
	[NSThread sleepForTimeInterval:0.25];
	//
	[self updateUpdaterLaunchd];
}

- (void)stopServer
{
	[_prefs saveIfDirty:self];
	[self updateUpdaterLaunchd];
	//
	if ([self isServerRunning]) {
		NSError* __autoreleasing error = NULL;
		[self curl:@"!/stop-notification" withErrorRef:&error];
		[NSThread sleepForTimeInterval:0.1]; // give server some time to log the event.
	}
	NSArray* args = @[@"--stop"];
	[_prefs runAuthHelperWithArgs:(NSArray*)args withInputData:NULL withAuth:YES];
}

- (int)getPortNumber
{
	return [_prefs intForXPath:@"proxy-server/@listen-port"];
}

- (void)openServerLogInConsole
{
	[GBDebug executeCommandlineTool:@"open" withArgs:@[@"-a", @"Console", javaServerLogPath]];
}

- (void)sendReloadPreferencesForModifiedFilter:(int)filterId withRule:(int)ruleId;
{
	if (![self isServerRunning])
		return;
	NSString* url = @"!/reload-panel-settings";
	if (filterId > 0)
		url = [NSString stringWithFormat:@"%@?filter=%d&rule=%d", url, filterId, ruleId];
	[self curl:url];
}

- (void)saveDirtyPrefsAndNotify
{
	[_prefs saveIfDirty:self];
	[self sendReloadPreferencesForModifiedFilter:0 withRule:0];
}

#pragma mark ------------------------------------- Show error sheet

- (void)showServerErrorAlertWithUrl:(NSURL*)url
						  withError:(__autoreleasing NSError*)error
{
	NSString* msg = [NSString stringWithFormat:@"%@\n\nURL: %@",
					 @"Try deactivate and reactivate GlimmerBlocker",
					 [url path]];
	if (error)
		msg = [NSString stringWithFormat:@"%@\n\n%@", msg, error];
	NSLog(@"GB server error: %@", msg);
	[_alertHelper prepareCriticalAlertWithTitle:@"Problem performing the request"
							withInformativeText:msg];
	[_alertHelper showPreparedAlert];
	NSLog(@"GlimmerBlocker download of URL failed: %@", url);
	NSLog(@"Error: %@", error);
}

#pragma mark ------------------------------------- curl, NSData

- (NSData*)curlToData:(NSString*)pathArgs
		 withResponse:(NSURLResponse**)response
		 withErrorRef:(__autoreleasing NSError**)errorRef
{
	*response = NULL;
	if (errorRef)
		*errorRef = NULL;
	NSData* postData = NULL;
	NSRange r = [pathArgs rangeOfString:@"?"];
	if (r.location != NSNotFound && [pathArgs length] - r.location > 100) {
		postData = [[pathArgs substringFromIndex:r.location + 1] dataUsingEncoding:NSUTF8StringEncoding];
		pathArgs = [pathArgs substringToIndex:r.location];
	}
	NSURL* url = [NSURL URLWithString:[self getAdminUrl:pathArgs]];
	NSMutableURLRequest* req = [[NSMutableURLRequest alloc ] initWithURL:url];
	[req setCachePolicy:NSURLRequestReloadIgnoringLocalCacheData];
	double timeout = [pathArgs hasPrefix:@"/test-running"] ? 1.0 : 60.0;
	[req setTimeoutInterval:timeout];
	if (postData) {
		[req setHTTPMethod:@"POST"];
		[req setValue:@"application/x-www-form-urlencoded; charset=UTF-8" forHTTPHeaderField:@"Content-Type"];
		[req setHTTPBody:postData];
	}
	NSError* __autoreleasing error = NULL;
	NSError *__autoreleasing* tmpRef = (errorRef ? errorRef : &error);
	NSData* data = [NSURLConnection sendSynchronousRequest:req returningResponse:response error:tmpRef];
	if (!data || !*response || *tmpRef) {
		if (errorRef)
			return NULL;
		[self showServerErrorAlertWithUrl:url withError:error];
		if (!response)
			NSLog(@"  didn't get any response object.");
		return NULL;
	}
#ifdef DEBUG_COMM
	DebugNSLog(@"Did download URL: %@", url);
#endif
	return data;
}

#pragma mark ------------------------------------- curl, NSString

- (NSString*)curl:(NSString*)pathArgs
{
	return [self curl:pathArgs withErrorRef:NULL];
}

- (NSString*)curl:(NSString*)pathArgs
	 withErrorRef:(__autoreleasing NSError**)errorRef
{
	NSURLResponse *response = NULL;
	NSData* data = [self curlToData:pathArgs withResponse:&response withErrorRef:errorRef];
	if (!data)
		return NULL;
#ifdef DEBUG_COMM
	DebugNSLog(@"Did download URL: %@", pathArgs);
	DebugNSLog(@"  textEncodingName: %@", [response textEncodingName]);
#endif
	NSString* s = [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
#if DEBUG_COMM
	DebugNSLog(@"Result: {{{%@}}}", s);
#endif
	return s;
}

#pragma mark ------------------------------------- curl, XmlDoc

- (XmlDoc*)curlToXml:(NSString*)pathArgs withActionDescription:(NSString*)actionDesc
{
	return [self curlToXml:pathArgs withActionDescription:actionDesc withErrorRef:NULL];
}

- (XmlDoc*)curlToXml:(NSString*)pathArgs withActionDescription:(NSString*)actionDesc
		withErrorRef:(__autoreleasing NSError**)errorRef;
{
	NSURLResponse *response = NULL;
	NSData* data = [self curlToData:pathArgs withResponse:&response withErrorRef:errorRef];
	if (!data)
		return NULL;
	NSURL* url = [NSURL URLWithString:[self getAdminUrl:pathArgs]];
	XmlDoc* doc = [[XmlDoc alloc] initWithURL:url withData:data];
	if (![doc rootE])
		return NULL;
	NSString* error = [Xml getAttribute:[doc rootE] withName:@"error"];
	if (error) {
		if (errorRef) {
			NSDictionary* dict = @{NSLocalizedDescriptionKey: error};
			NSString* errorDomain = @"dk.vitality.glimmerblocker";
			*errorRef = [NSError errorWithDomain:errorDomain code:-1 userInfo:dict];
			return NULL;
		}
		if (actionDesc) {
			[_alertHelper prepareCriticalAlertWithTitle:actionDesc
									withInformativeText:error];
			[_alertHelper showPreparedAlert];
		} else {
			NSLog(@"Got http error in curlXml without actionDesc: %@", error);
		}
		return NULL;
	}
	return doc;
}

@end
