/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy;

import java.io.File;
import java.io.FileOutputStream;
import java.util.HashSet;
import java.util.Set;
import org.glimmerblocker.proxy.util.DocumentPrinter;
import org.glimmerblocker.proxy.util.XmlParserHelper;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import dk.vitality.lib.foundation.misc.RandomUtils;
import dk.vitality.lib.foundation.misc.StringUtils;
import dk.vitality.lib.foundation.xml.xpath.XmlArgs;
import static dk.vitality.lib.foundation.exception.ExceptionUtils.getExMsg;

public class PanelPreferencesUpgrader
{
	public static void main(String[] args)
	{
		System.err.println("PanelPreferencesUpgrader launched.");
		try {
			new PanelPreferencesUpgrader();
			System.err.println("PanelPreferencesUpgrader done.");
		}
		catch (Exception ex) {
			System.err.println("PanelPreferencesUpgrader got exception in main thread: " + getExMsg(ex));
			ex.printStackTrace();
			System.exit(1);
		}
	}

	XmlArgs docXA;
	Document doc;

	PanelPreferencesUpgrader() throws Exception
	{
		File f = new File("/Library/GlimmerBlocker/PanelSettings.xml");
		docXA = new XmlArgs(XmlParserHelper.parse(f));
		doc = docXA.getDocument();
		String old = new DocumentPrinter(doc).printToString();
		int version = docXA.getReqInt("@version");
		if (version < 3) {
			setAttr(true, "suspects/@enabled");
			setAttr(true, "suspects/@show-details");
		}
		if (version < 5) {
			setAttr(true, "updater/@check-in-background");
			setAttr(false, "updater/@use-beta-versions");
		}
		if (version < 6)
			setAttr(null, "debug/@plugins");
		if (version < 7)
			setAttr(null, "debug/@enable"); // renamed to "enabled" just like rest of file
		if (version < 8)
			setAttr(null, "proxy-server/@public-admin-key"); // renamed to "info-admin-key"
		if (version < 9) {
			removeFilter("http://glimmerblocker.org/site/filters/default/WorldWideAdProviders.xml");
			removeFilter("http://glimmerblocker.org/site/filters/danish-sites.xml");
			//
			addFilter(true, "http://glimmerblocker.org/site/filters/default/ad-networks.xml");
			addFilter(true, "http://glimmerblocker.org/site/filters/default/site-specific.xml");
			addFilter(false, "http://glimmerblocker.org/site/filters/default/intrusive.xml");
			addFilter(false, "http://glimmerblocker.org/site/filters/default/site-enhancements.xml");
		}
		//if (version < 10) {
			//addFilter(true, "http://glimmerblocker.org/site/filters/default/keywords.xml");
		//}
		if (version < 11) {
			String s = docXA.getOptString("network-settings/@site-proxy");
			if (s != null && s.equals("1"))
				setAttr("static", "network-settings/@site-proxy");
		}
		if (version < 12) {
			boolean beta = docXA.getOptBoolean("updater/@use-beta-versions");
			File betaF = new File("/Library/GlimmerBlocker/UpdaterUseBetaChannel.txt");
			if (beta)
				new FileOutputStream(betaF).close();
			else if (betaF.exists() && !betaF.delete())
				System.err.println("Can't delete beta marker file: " + betaF);
		}
		int currentVersion = 12;
		//--------
		upgradeFilterElements();
		upgradeRuleElements(docXA.getElementArray("filters/filter/filter-data/rule"));
		docXA.getContextElement().setAttribute("version", Integer.toString(currentVersion));
		// can't just check version number to find out if the prefs have been modified
		// as <filter> or <rule> elements might have been modified,
		// so we need to do full check of serialized xml.
		if (old.equals(new DocumentPrinter(doc).printToString())) {
			System.err.println("  Did not modify " + f.getName());
			return;
		}
		System.err.println("  Saves modified " + f.getName());
		XmlParserHelper.saveDocument(doc, f);
	}

	void addFilter(boolean enabled, String url) throws Exception
	{
		if (docXA.exists("filters/filter[@url='" + url + "']"))
			return;
		Element filtersE = docXA.getReqElement("filters");
		Node beforeN = filtersE.getFirstChild();
		for (XmlArgs xa : docXA.getXmlArgsList("filters/filter")) {
			String u = xa.getOptString("@url");
			if (u != null && u.startsWith("http://glimmerblocker.org"))
				beforeN = xa.getContextNode().getNextSibling();
		}
		Element e = doc.createElement("filter");
		e.setAttribute("url", url);
		e.setAttribute("id", Integer.toString(getNewFilterId()));
		e.setAttribute("safety-policy", "modification");
		e.setAttribute("update-policy", "auto");
		if (enabled)
			e.setAttribute("enabled", "1");
		filtersE.insertBefore(e, beforeN);
	}

	void upgradeFilterElements() throws Exception
	{
		for (XmlArgs filterXA : docXA.getXmlArgsList("filters/filter")) {
			filterXA.getContextElement().removeAttribute("selected-rule"); // renamed to 'selected-rules'
			if (filterXA.getOptInt("@id", 0) == 0)
				filterXA.getContextElement().setAttribute("id", Integer.toString(getNewFilterId()));
		}
	}

	int getNewFilterId() throws Exception
	{
		int id = 1001;
		while (docXA.exists("filters/filter[@id='" + id + "']"))
			id++;
		return id;
	}

	void removeFilter(String url) throws Exception
	{
		for (Element e : docXA.getElementArray("filters/filter[@url='" + url + "']"))
			e.getParentNode().removeChild(e);
	}

	void setAttr(boolean b, String xpath) throws Exception
	{
		setAttr(b ? "1" : null, xpath);
	}

	void setAttr(String value, String xpath) throws Exception
	{
		XmlArgs xa = docXA;
		for (String s : StringUtils.splitByChar(xpath, '/')) {
			if (!s.startsWith("@")) {
				XmlArgs subXA = xa.getOptXmlArgs(s);
				if (subXA == null) {
					Element e = doc.createElement(s);
					xa.getContextNode().appendChild(e);
					subXA = XmlArgs.create(e);
				}
				xa = subXA;
				continue;
			}
			Element e = xa.getContextElement();
			String name = s.substring(1);
			if (value != null)
				e.setAttribute(name, value);
			else
				e.removeAttribute(name);
		}
	}

	public static void upgradeRuleElements(Element[] ruleElements) throws Exception
	{
		Set<Integer> usedRuleIds = new HashSet<>();
		for (Element e : ruleElements) {
			String id = e.getAttribute("id");
			if (id != null && id.length() > 0)
				usedRuleIds.add(Integer.parseInt(id));
		}
		for (Element e : ruleElements) {
			XmlArgs ruleXA = XmlArgs.create(e);
			if (!ruleXA.exists("@type"))
				addRuleType(ruleXA);
			if (ruleXA.getOptInt("@rule-id", 0) > 0)
				continue;
			int id;
			do {
				id = RandomUtils.randomizer.nextInt(); // see Xml.m
			} while (id < 1234567 || usedRuleIds.contains(id));
			usedRuleIds.add(id);
			ruleXA.getContextElement().setAttribute("rule-id", Integer.toString(id));
		}
	}

	static void addRuleType(XmlArgs ruleXA) throws Exception
	{
		Element ruleE = ruleXA.getContextElement();
		if (!ruleXA.exists("@whitelist")) { // old, now obsolete attribute.
			ruleE.setAttribute("type", "block");
			return;
		}
		if (ruleXA.exists("css") || ruleXA.exists("js") || ruleXA.exists("transform"))
			ruleE.setAttribute("type", "modify");
		else
			ruleE.setAttribute("type", "whitelist");
	}
}
