/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals;

import java.io.File;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.util.XmlParserHelper;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.io.FileContentsUtils;
import dk.vitality.lib.foundation.misc.HexUtils;
import dk.vitality.lib.foundation.xml.xpath.XmlArgs;

public class JavaSettings implements Logger
{
	public final GlimmerGlobals globals;

	private final Logger logger;
	private String secureAdminKey;
	private String attachmentDownloadHostPrefix;
	private boolean dirty;
	private static final int codeVersion = 1;

	private static final String javaPrefsPath = "/Library/GlimmerBlocker/JavaSettings.xml";

	public JavaSettings(GlimmerGlobals globals) throws Exception
	{
		this.globals = globals;
		this.logger = new PrefixedLogger(globals, "java-settings");
		ln("Loads from disk.");
		try {
			File f = new File(javaPrefsPath);
			XmlArgs docXA = null;
			if (f.exists())
				docXA = new XmlArgs(XmlParserHelper.parse(f));
			else
				ln("Missing file. Using defaults.");
			if (docXA != null && docXA.getContextElement() != null)
				postLoadSettings(docXA);
		}
		catch (Exception ex) {
			ln("FAILED TO LOAD JAVA SETTINGS: " + ex.getMessage());
			ln(ExceptionUtils.getStackTraceAsString(ex));
			ln("");
			ln("OVERRIDES WITH DEFAULT SETTINGS.");
		}
		if (secureAdminKey == null) {
			secureAdminKey = HexUtils.toLcHex(FileContentsUtils.getDevRandomBytes(24));
			notifyDirty();
		}
		if (attachmentDownloadHostPrefix == null) {
			attachmentDownloadHostPrefix = "glimmerblocker.attachment." +
					HexUtils.toLcHex(FileContentsUtils.getDevRandomBytes(10)) + ".";
			notifyDirty();
		}
		saveNowIfDirty();
	}

	public String getSecureAdminKey()
	{
		return secureAdminKey;
	}

	public String getAttachmentDownloadHostPrefix()
	{
		return attachmentDownloadHostPrefix;
	}

	private void postLoadSettings(@NotNull XmlArgs docXA) throws Exception
	{
		int oldVersion = docXA.getReqInt("@version");
		if (oldVersion > codeVersion) {
			ln("Settings on disk is newer version (" + oldVersion + ")" +
					" than the server code (" + codeVersion + ")");
		}
		secureAdminKey = docXA.getOptString("proxy/@secure-admin-key");
		attachmentDownloadHostPrefix = docXA.getOptString("proxy/@attachment-host-prefix");
	}

	public synchronized void saveNowIfDirty()
	{
		if (!dirty)
			return;
		if (globals.debugFlag)
			ln("Saves preferences.");
		dirty = false;
		// if we clear the dirty-flag after saving, it won't be cleared if saving fails,
		// and might keep trying to save.
		try {
			Document doc = XmlParserHelper.createDocument();
			Element rootE = doc.createElement("java-settings");
			doc.appendChild(rootE);
			rootE.setAttribute("version", Integer.toString(codeVersion));
			//
			Element proxyE = doc.createElement("proxy");
			rootE.appendChild(proxyE);
			proxyE.setAttribute("secure-admin-key", secureAdminKey);
			proxyE.setAttribute("attachment-host-prefix", attachmentDownloadHostPrefix);
			//
			XmlParserHelper.saveDocument(doc, new File(javaPrefsPath));
		}
		catch (Exception ex) {
			ln("Failed saving java settings: " + ex.getMessage());
			ln(ExceptionUtils.getStackTraceAsString(ex));
		}
	}

	synchronized void notifyDirty()
	{
		if (dirty)
			return;
		dirty = true;
		if (globals.debugFlag)
			ln("Now dirty.");
	}


	public void ln(@Nullable String s)
	{
		logger.ln(s);
	}
}
