/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals.network.dns;

import java.net.InetAddress;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.ref.FatRef;

public class DnsResolver implements Logger
{
	@NotNull private final Logger logger;
	@NotNull private FatRef<Boolean> debugNetworkConfig;

	public DnsResolver(@NotNull Logger parentLogger,
	                   @NotNull FatRef<Boolean> debugNetworkConfig)
	{
		this.logger = new PrefixedLogger(parentLogger, "DNS");
		this.debugNetworkConfig = debugNetworkConfig;
	}

	public InetAddress[] getAllByName(@NotNull DnsResolverConfig config,
	                                  @NotNull final String hostname) throws Exception
	{
		if (config.isNativeResolver)
			return useNative(hostname, "config");
		else if (hostname.length() >= 7 && isDottedVersion4(hostname))
			return useNative(hostname, "dotted IPv4");
		else if (hostname.indexOf(':') >= 0)
			return useNative(hostname, "colon IPv6");
		DnsResolverCache drc;
		synchronized (caches) {
			drc = caches.get(config);
			if (drc == null) {
				drc = new DnsResolverCache(config, this, debugNetworkConfig);
				caches.put(config, drc);
			}
		}
		return drc.query(dnsResolver -> dnsResolver.lookupHost(hostname));
	}

	private InetAddress[] useNative(@NotNull final String hostname, @NotNull String why) throws Exception
	{
		try {
			InetAddress[] result = InetAddress.getAllByName(hostname);
			if (debugNetworkConfig.get())
				logger.ln("Native (" + why + ") DNS resolved '" + hostname + "' to: " + Arrays.toString(result));
			return result;
		}
		catch (Exception ex) {
			if (debugNetworkConfig.get())
				logger.ln("Native DNS failed to resolve '" + hostname + "'.");
			throw ex;
		}
	}

	public void ln(@Nullable String s)
	{
		logger.ln(s);
	}

	/**
	 * Typical case for simple dotted address, e.g. "130.225.220.16"
	 */
	private static boolean isDottedVersion4(String s)
	{
		int len = s.length();
		int dots = 0;
		int minAllowedDotIndex = 1;
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch == '.') {
				if (dots == 3 || i < minAllowedDotIndex)
					return false;
				dots++;
				minAllowedDotIndex = i + 2; // dots must be separated by digit.
				continue;
			}
			if (ch < '0' || ch > '9')
				return false;
		}
		return dots == 3 && minAllowedDotIndex <= len;
	}

	private static final Map<DnsResolverConfig, DnsResolverCache> caches = new HashMap<>();
}
