/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals.subscriptions;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.TimeUnit;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.xml.xpath.XmlArgs;

public class SubscriptionsAutoUpdater implements Runnable, Logger
{
	public final GlimmerGlobals globals;
	private final Logger logger;
	private final ArrayBlockingQueue<Long> sleepQueue = new ArrayBlockingQueue<>(5);
	public static final long hour = 60 * 60 * 1000;
	private static final long updateInterval = 23 * hour;

	public SubscriptionsAutoUpdater(GlimmerGlobals globals)
	{
		this.globals = globals;
		this.logger = new PrefixedLogger(globals, "SubscriptionsAutoUpdater");
		Thread t = new Thread(this, "SubscriptionsAutoUpdater");
		t.setDaemon(true);
		t.start();
		if (hasGBSkipAutoUpdaters())
			ln("Has GBSkipAutoUpdaters");
		else
			sleepQueue.offer(-1L);
	}

	public static boolean hasGBSkipAutoUpdaters()
	{
		return System.getProperty("GBSkipAutoUpdaters") != null;
	}

	public void runForcedCheck()
	{
		sleepQueue.offer(Long.MAX_VALUE);
	}

	/**
	 * Called when a filter is enabled.  In this case GB shouldn't wait between updating
	 * each filter because it then might take many seconds before the enabled filter is updated.
	 */
	public void runSoftCheck()
	{
		sleepQueue.offer(System.currentTimeMillis() - updateInterval);
	}

	public void run()
	{
		while (true) {
			Long earliestAllowedUpdateTime = null;
			try {
				long sleep;
				if (hasGBSkipAutoUpdaters())
					sleep = Long.MAX_VALUE;
				else
					sleep = hour;
				earliestAllowedUpdateTime = sleepQueue.poll(sleep, TimeUnit.MILLISECONDS);
			}
			catch (InterruptedException ex) {
				ln("Was interrupted: " + ExceptionUtils.getExMsg(ex));
			}
			boolean sleepBetweenUpdates;
			if (earliestAllowedUpdateTime == null) {
				// Normal check every hour.
				earliestAllowedUpdateTime = System.currentTimeMillis() - updateInterval;
				sleepBetweenUpdates = true;
			} else if (earliestAllowedUpdateTime == -1) {
				// Check just after launch of GB
				earliestAllowedUpdateTime = System.currentTimeMillis() - hour;
				sleepBetweenUpdates = true;
			} else {
				// Forced "check all now"
				sleepBetweenUpdates = false;
			}
			try {
				boolean skippedLast = false;
				for (XmlArgs filterXA : globals.panelSettingsXA.getXmlArgsList("filters/filter")) {
					if (sleepBetweenUpdates && !skippedLast)
						sleepUntilNetworkIsStable();
					skippedLast = !updateSubscription(filterXA, earliestAllowedUpdateTime);
				}
			}
			catch (Exception ex) {
				ln("Got exception: " + ex.getMessage());
				ln(ExceptionUtils.getStackTraceAsString(ex));
			}
		}
	}

	/**
	 * @return true if the method did something processor intensive.
	 */
	private boolean updateSubscription(XmlArgs filterXA,
	                                   long earliestAllowedUpdateTime) throws Exception
	{
		String url = filterXA.getOptString("@url");
		if (url == null || filterXA.exists("filter-data"))
			return false;
		int filterId = filterXA.getReqInt("@id");
		try {
			if ("manual".equals(filterXA.getOptString("@update-policy"))) {
				if (globals.debugFlag)
					ln("Skips checking subscription for updates: update-policy = manual.");
				return false;
			}
			globals.filterSubscriptions.updateSubscription(filterId, false, earliestAllowedUpdateTime);
		}
		catch (Exception ex) {
			ln("Failed updating subscription " + filterId + ": " + url);
			ln("  " + ExceptionUtils.getExMsg(ex));
			if (globals.debugFlag)
				ln(ExceptionUtils.getStackTraceAsString(ex));
		}
		return true;
	}

	private void sleepUntilNetworkIsStable()
	{
		// If network configuration has just changed, DNS might not work yet, and it will
		// cause a lot of failures.
		while (true) {
			long st0 = System.currentTimeMillis();
			try {
				Thread.sleep(1000);
			}
			catch (InterruptedException ex) {
				ln("Network state check sleep was interrupted: " + ExceptionUtils.getExMsg(ex));
				continue;
			}
			long st1 = System.currentTimeMillis();
			if (st1 - st0 > 2000)
				continue; // mac did sleep while sleep()
			long changed = globals.networkManager.getConfig().date;
			if (st1 - changed > 30 * 1000)
				return;
		}
	}

	public void ln(@Nullable String s)
	{
		logger.ln(s);
	}
}
