/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals.suspect;

import java.net.URL;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.http.data.Headers;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import static java.util.regex.Pattern.CASE_INSENSITIVE;
import static java.util.regex.Pattern.compile;

class SuspectCandidate implements Logger
{
	@NotNull final Logger parentLogger;
	@NotNull final GlimmerGlobals globals;
	@NotNull final String requestMethod;
	@NotNull final URL requestURL;
	@NotNull final Headers requestHeaders;
	@NotNull final Headers responseHeaders;
	@NotNull final String requestUrlPath;
	@NotNull final SuspectsManager mgr;

	@Nullable final String referer;
	final boolean crossDomain;
	@NotNull final String requestDomain;

	@Nullable private Logger logger;

	@Nullable Suspect suspect;

	private static final boolean debugSuspectCandidates = false;

	SuspectCandidate(@NotNull Logger parentLogger,
	                 @NotNull GlimmerGlobals globals,
	                 @NotNull SuspectsManager mgr,
	                 @NotNull String requestMethod,
	                 @NotNull URL requestURL,
	                 @NotNull Headers requestHeaders,
	                 @NotNull Headers responseHeaders)
	{
		this.parentLogger = parentLogger;
		this.globals = globals;
		this.requestMethod = requestMethod;
		this.requestURL = requestURL;
		this.requestHeaders = requestHeaders;
		this.responseHeaders = responseHeaders;
		this.mgr = mgr;
		this.referer = requestHeaders.getLC("referer");
		String s = requestURL.getPath();
		this.requestUrlPath = (s == null || s.length() == 0) ? "/" : s;
		this.requestDomain = getDomain(requestURL.getHost());
		this.crossDomain = isDifferentDomain();
		if (debugSuspectCandidates)
			ln("Suspect.process: " + requestURL);
		if (requestDomain.equals("youtube.com") || requestDomain.equals("google.com")) {
			if (debugSuspectCandidates)
				ln("Whitelist domain: " + requestDomain);
			return;
		}
		String ct = responseHeaders.contentTypeBaseLC;
		if (ct.equals("application/x-shockwave-flash")) {
			create(SuspectType.Flash, "Content-Type is Flash™");
			return;
		}
		if (requestUrlPath.contains(".swf")) {
			create(SuspectType.Flash, "Flash: path ends with .swf but with content-type: " + ct);
			if (globals.debugFlag || debugSuspectCandidates)
				ln(".swf with base content-type = '" + ct + "'");
			return;
		}
		boolean js = requestUrlPath.endsWith(".js") || ct.endsWith("javascript");
		if (js) {
			if (requestURL.getHost().endsWith("googlesyndication.com"))
				return; // don't mind text ads
			if (crossDomain) {
				create(SuspectType.Javascript, "Javascript from different domain");
				return;
			}
		}
	}

	private void create(SuspectType type, String why)
	{
		String query = requestURL.getQuery();
		if (query == null)
			query = "";
		this.suspect = new Suspect(requestURL.getHost(), requestUrlPath, query,
				crossDomain, referer, type, why);
	}

	private boolean isDifferentDomain()
	{
		if (referer == null || !referer.startsWith("http"))
			return false;
		Matcher m = urlHostRE.matcher(referer);
		if (!m.find()) {
			if (globals.debugFlag)
				ln("Referer doesn't match host-RE: '" + referer + "'");
			return false;
		}
		String r = getDomain(m.group(1));
		if (debugSuspectCandidates)
			ln("Domains: req = '" + requestDomain + "', referer = '" + r + "' from " + referer);
		return !requestDomain.equals(r);
	}

	private static final Pattern urlHostRE = compile("^https?://([^:/]+)", CASE_INSENSITIVE);

	private String getDomain(String host)
	{
		String r = host;
		int a = host.lastIndexOf('.');
		if (a > 0) {
			int b = host.lastIndexOf('.', a - 1);
			if (b > 0) {
				String s = host.substring(b + 1, a);
				if (s.equals("co") || s.equals("edu") || s.equals("gov") || s.equals("com") || s.equals("org")) {
					int c = host.lastIndexOf('.', b - 1);
					r = (c < 0) ? host : host.substring(c + 1);
				} else {
					r = host.substring(b + 1);
				}
			}
		}
		if (debugSuspectCandidates)
			ln("Domain(" + host + ") -> '" + r + "'");
		return r;
	}

	public void ln(@Nullable String s)
	{
		if (logger == null)
			logger = new PrefixedLogger(parentLogger, "suspects");
		logger.ln(s);
	}
}
