/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.util.zip.GZIPInputStream;
import org.glimmerblocker.proxy.http.client.connection.HttpResponseInfo;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.misc.ArrayUtils;

public class BodyCollector
{
	@NotNull final HttpResponseInfo rhi;
	@Nullable protected byte[] responseContent;

	public BodyCollector(@NotNull HttpResponseInfo rhi)
	{
		this.rhi = rhi;
	}

	public void readBody(@Nullable InputStream is) throws Exception
	{
		if (rhi.contentLength == 0) {
			responseContent = ArrayUtils.emptyByteArray;
			return;
		}
		if (is == null)
			return;
		int bufferSize = 4096;
		if (rhi.contentLength >= 0)
			bufferSize = (int)Math.max(16, Math.min(16384, rhi.contentLength));
		byte[] buffer = new byte[bufferSize];
		ByteArrayOutputStream baos = new ByteArrayOutputStream(bufferSize);
		if (isGzipped())
			is = new GZIPInputStream(is);
		while (true) {
			int num = is.read(buffer);
			if (num < 0)
				break;
			baos.write(buffer, 0, num);
		}
		responseContent = baos.toByteArray();
	}

	private boolean isGzipped() throws Exception
	{
		String ct = rhi.headers.contentEncoding;
		if (ct == null || ct.isEmpty())
			return false;
		ct = ct.trim().toLowerCase();
		if (ct.length() == 0 || ct.equals("identity"))
			return false;
		if (ct.equals("none")) // buggy http://www.sparitual.de
			return false;
		if (ct.equals("gzip") || ct.contains("gzip"))
			return true;
		throw new Exception("Can't decode content-encoding = '" + ct + "'");
	}

	@NotNull
	public ResultData getResultData()
	{
		// make headers object without content-encoding, as it has been decoded.
		return new ResultData(rhi.headers.copyWithContentEncoding(null), responseContent);
	}
}
