/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client.curl;

import java.util.List;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.HeadersBuilder;
import org.glimmerblocker.proxy.rhino.callback.CallbackObjectBase;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.NativeArray;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;
import org.mozilla.javascript.UniqueTag;
import static org.glimmerblocker.proxy.http.data.HeadersBuilder.acceptEncodingHeaderNameLC;

/**
 * Not valid for caching, one shot only!
 * The name used for the properties are always mixed-case.
 */
public class HeadersMapObject extends ScriptableObject implements Logger
{
	@NotNull CallbackObjectBase parent;
	@NotNull final Scriptable scope;
	@NotNull final Context ctx;

	public HeadersMapObject(@NotNull CallbackObjectBase parent,
	                        @NotNull Context ctx,
	                        @NotNull Scriptable scope,
	                        @NotNull Headers h)
	{
		this.parent = parent;
		this.scope = scope;
		this.ctx = ctx;
		setPrototype(parent.prototype);
		for (String nameLC : h.headerMap.keySet()) {
			List<String> values = h.headerMap.get(nameLC);
			if (values.isEmpty())
				continue;
			Object a = ctx.newArray(scope, values.toArray(new Object[values.size()]));
			String nameMX = Headers.getMixedCaseHeaderName(nameLC);
			super.put(nameMX, this, a);
		}
		setHeaderLC("content-type", h.contentType);
		setHeaderLC("content-encoding", h.contentEncoding);
	}

	@Nullable
	public String getHeaderLC(@NotNull String nameLC)
	{
		Object a = get(Headers.getMixedCaseHeaderName(nameLC), this);
		if (a == UniqueTag.NOT_FOUND || a == UniqueTag.NULL_VALUE)
			return null;
		if (!(a instanceof NativeArray))
			return Context.toString(a);
		NativeArray na = (NativeArray)a;
		if (na.getLength() <= 0)
			return null;
		return Context.toString(na.get(0, na));
	}

	@Nullable
	public String getBaseContentType()
	{
		String s = getHeaderLC("content-type");
		if (s == null)
			return null;
		else
			return Headers.makeBaseLcContentType(s);
	}

	public void setHeaderLC(@NotNull String nameLC, @Nullable String value)
	{
		if (value == null || value.length() == 0) {
			delete(nameLC);
			return;
		}
		Object a = ctx.newArray(scope, new Object[]{value});
		put(nameLC, this, a);
	}

	@NotNull
	public Headers extract(int status, String statusMessage)
	{
		HeadersBuilder hb = HeadersBuilder.makeEmpty();
		hb.status = status;
		hb.statusMessage = statusMessage;
		for (Object k : getIds()) {
			if (!(k instanceof String))
				continue;
			String name = ((String)k);
			Object a = get(name, this);
			if (a == null)
				continue;
			String nameLC = name.toLowerCase();
			if (a instanceof NativeArray) {
				NativeArray na = (NativeArray)a;
				int num = (int)na.getLength();
				for (int i = 0; i < num; i++)
					extractHeader(hb, nameLC, Context.toString(na.get(i, na)));
			} else {
				extractHeader(hb, nameLC, Context.toString(a));
			}
		}
		return hb.toHeaders();
	}

	private void extractHeader(@NotNull HeadersBuilder hb,
	                           @NotNull String nameLC,
	                           @Nullable String value)
	{
		if (nameLC.equals("content-type"))
			hb.contentType = value;
		else if (nameLC.equals("content-encoding"))
			hb.contentEncoding = value;
		else if (value != null && value.length() > 0)
			hb.add(nameLC, value);
		else
			hb.removeAll(nameLC);
	}

	@Override
	public String getClassName()
	{
		return getClass().getSimpleName();
	}

	public void put(@NotNull String name, @NotNull Scriptable start, @Nullable Object value)
	{
		String nameLC = name.toLowerCase();
		String nameMX = Headers.getMixedCaseHeaderName(nameLC);
		checkAllowedToModifyHeader(nameLC, nameMX, value);
		super.put(nameMX, start, value);
	}

	public Object get(String name, Scriptable start)
	{
		name = Headers.getMixedCaseHeaderName(name);
		return super.get(name, start);
	}

	public void delete(@NotNull String name)
	{
		String nameLC = name.toLowerCase();
		String nameMX = Headers.getMixedCaseHeaderName(nameLC);
		checkAllowedToModifyHeader(nameLC, nameMX, null);
		super.delete(nameMX);
	}

	private void checkAllowedToModifyHeader(@NotNull String nameLC, @NotNull String nameMX, @Nullable Object value)
	{
		if (nameLC.equals(acceptEncodingHeaderNameLC)) {
			String s = (value != null) ? Context.toString(value) : null;
			if (s == null || s.length() <= 0 || s.equalsIgnoreCase("gzip"))
				return;
			throw Context.reportRuntimeError("The value '" + s + "' is not acceptable" +
					" for the Accept-Encoding header." +
					" It must be null, empty or 'gzip' as the proxy must be able to decode the content.");
		}
		// Allow to set host, as setting the url automatically updates the host header.
		if (!nameLC.equals("host") && HeadersBuilder.neverCopyHeadersLC.contains(nameLC)) {
			throw Context.reportRuntimeError("You must not modify the '" + nameMX + "' http header as" +
					" the proxy must control its value.");
		}
	}

	public void ln(@Nullable String s)
	{
		parent.ln(s);
	}
}
