/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client.curl;

import java.net.ConnectException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.UnknownHostException;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.proxyRequest.errorPage.ErrorPage;
import org.glimmerblocker.proxy.rhino.callback.CallbackObjectBase;
import org.glimmerblocker.proxy.rhino.callback.CallbackRhinoBinder;
import org.glimmerblocker.proxy.rhino.callback.CallbackSubObject;
import org.glimmerblocker.proxy.rhino.callback.ContentGlobalObjectBase;
import org.glimmerblocker.proxy.rhino.callback.RhinoCallback;
import org.glimmerblocker.proxy.util.HttpException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.mozilla.javascript.Context;
import dk.vitality.lib.foundation.misc.StringUtils;
import static java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT;
import static java.net.HttpURLConnection.HTTP_NOT_FOUND;

public class ResponseSubObject extends CallbackSubObject
{
	StaticRequestSubObject requestObject;
	Headers origHeaders;
	int newStatus;
	String newStatusMessage;
	ContentGlobalObjectBase contentObject;
	HeadersMapObject headers;

	private static final CallbackRhinoBinder classBinder = new CallbackRhinoBinder(ResponseSubObject.class);

	public ResponseSubObject(@NotNull CallbackObjectBase parent)
	{
		super(parent, classBinder);
	}

	public void prepareResponseSubObject(@NotNull Context ctx,
	                                     @NotNull StaticRequestSubObject requestObject,
	                                     @NotNull Headers origHeaders,
	                                     @NotNull ContentGlobalObjectBase contentObject) throws Exception
	{
		prepareObjectBase(ctx);
		this.requestObject = requestObject;
		this.origHeaders = origHeaders;
		this.contentObject = contentObject;
		this.newStatus = origHeaders.status;
		this.newStatusMessage = origHeaders.statusMessage;
		this.headers = null;
	}

	@Override
	public void notifyLeftContext()
	{
		super.notifyLeftContext();
		this.origHeaders = null;
		this.contentObject = null;
		this.newStatusMessage = null;
		this.headers = null;
	}

	public boolean dirty()
	{
		return newStatus != origHeaders.status
				|| !StringUtils.triEquals(newStatusMessage, origHeaders.statusMessage)
				|| headers != null;
	}

	@NotNull
	public Headers extract()
	{
		return jsGet_headers().extract(newStatus, newStatusMessage);
	}
	//===========================================================================

	@RhinoCallback
	public HeadersMapObject jsGet_headers()
	{
		if (headers != null)
			return headers;
		headers = new HeadersMapObject(this, ctx, scope, origHeaders);
		return headers;
	}

	//===========================================================================

	@RhinoCallback
	public String jsGet_status()
	{
		return Integer.toString(newStatus);
	}

	@RhinoCallback
	@SuppressWarnings({"UnusedCatchParameter"})
	public void jsSet_status(String s)
	{
		try {
			int x = Integer.parseInt(s);
			if (x < 100 || x > 999)
				throw Context.reportRuntimeError("Invalid http status: " + x);
			newStatus = x;
		}
		catch (Exception ex) {
			throw Context.reportRuntimeError("Invalid http status: '" + s + "'");
		}
	}

	@RhinoCallback
	public String jsGet_statusMessage()
	{
		return newStatusMessage;
	}

	@RhinoCallback
	public void jsSet_statusMessage(String s)
	{
		newStatusMessage = (s != null) ? s : "";
	}

	@Nullable
	@RhinoCallback
	public String jsGet_baseContentType()
	{
		return jsGet_headers().getBaseContentType();
	}

	@Nullable
	@RhinoCallback
	public String jsGet_contentType()
	{
		return jsGet_headers().getHeaderLC("content-type");
	}

	@RhinoCallback
	public void jsSet_contentType(String s)
	{
		jsGet_headers().setHeaderLC("content-type", s);
	}

	@RhinoCallback
	public void jsFunction_sendRedirect(String url)
	{
		newStatus = HttpURLConnection.HTTP_SEE_OTHER;
		newStatusMessage = "Redirected to " + url;
		HeadersMapObject h = jsGet_headers();
		h.setHeaderLC("location", url);
		h.setHeaderLC("content-type", "text/plain");
		contentObject.setContent("Redirected to: " + url + "\n");
	}

	@RhinoCallback
	public void jsFunction_sendHtml(String html)
	{
		set(html, "text/html");
	}

	@RhinoCallback
	public void jsFunction_sendText(String text)
	{
		set(text, "text/plain");
	}

	@RhinoCallback
	@SuppressWarnings({"ThrowableInstanceNeverThrown"})
	public void jsFunction_sendUnknownHostErrorPage(String host) throws Exception
	{
		sendErrorPage(host, new UnknownHostException(host), HTTP_NOT_FOUND, "Host not found: " + host);
	}

	@RhinoCallback
	@SuppressWarnings({"ThrowableInstanceNeverThrown"})
	public void jsFunction_sendCantConnectErrorPage(String host) throws Exception
	{
		sendErrorPage(host, new ConnectException(host), HTTP_GATEWAY_TIMEOUT, "Server denies connection: " + host);
	}

	@SuppressWarnings({"ThrowableInstanceNeverThrown"})
	private void sendErrorPage(String host, Exception ex, int status, String statusMessage) throws Exception
	{
		URL url = new URL("http", host, "");
		String ua = requestObject.jsGet_headers().getHeaderLC("user-agent");
		HttpException he = new HttpException(new Headers(status, statusMessage), ex);
		ErrorPage ep = new ErrorPage(globals, he, ua, url);
		newStatus = status;
		newStatusMessage = statusMessage;
		HeadersMapObject hmo = jsGet_headers();
		hmo.setHeaderLC("content-type", "text/html");
		hmo.delete("location");
		contentObject.setContent(ep.getResultData().getContentAsText(this, false));
	}

	private void set(String html, String contentType)
	{
		newStatus = HttpURLConnection.HTTP_OK;
		newStatusMessage = "OK";
		HeadersMapObject h = jsGet_headers();
		h.setHeaderLC("content-type", contentType);
		h.delete("location");
		contentObject.setContent(html);
	}
}
