/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.data;

import java.io.Serializable;
import java.net.HttpURLConnection;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.misc.StringUtils;

/**
 * Immutable HTTP headers (and status). It is used for both the request and the response headers.
 * <p/>
 * {@link HeadersBuilder} is to Headers as StringBuilder is to String.
 */
public class Headers implements Serializable
{
	public final int status; // http://httpstatus.es/
	@NotNull public final String statusMessage;
	public final boolean icyProtocol;
	public final boolean isBodylessStatusCode;


	/**
	 * BEWARE: If not sent by the server, it is null.
	 */
	@Nullable public final String contentType;

	/**
	 * The base content-type without any options and in lowercase.
	 * contentType = "TeXt/CsS; charset=utf-8" results in contentTypeBaseLC="text/css"
	 */
	@NotNull public final String contentTypeBaseLC;

	/**
	 * The compression (not charset) such as "gzip"
	 */
	@Nullable public final String contentEncoding;

	/**
	 * Maps from lower-case header name to ordered list of values.
	 * Does not contain Content-Type, Content-Encoding, Content-Length and a few other special headers.
	 */
	@NotNull public final Map<String, List<String>> headerMap;

	public static final Headers simpleTextHtml = new Headers("text/html; charset=UTF-8");

	/**
	 * Initializes with the specified status and message.
	 */
	public Headers(int status, @NotNull String statusMessage)
	{
		this(status, statusMessage, "text/plain; charset=UTF-8", null, null);
	}

	/**
	 * Initializes with the specified content-type,
	 * status = 200 OK and no content-encoding.
	 */
	public Headers(@Nullable String contentType)
	{
		this(HttpURLConnection.HTTP_OK, "OK", contentType, null, null);
	}

	/**
	 * @param status          HTTP status (e.g. 200)
	 * @param statusMessage   HTTP status message (e.g. "OK")
	 * @param contentType     Content-type header (e.g. "text/html")
	 * @param contentEncoding Content-Encoding header (e.g. "gzip")
	 * @param headerMap       A deep copy is made. Maps from lower-case header name to ordered list of values.
	 */
	public Headers(int status, @NotNull String statusMessage,
	               @Nullable String contentType,
	               @Nullable String contentEncoding,
	               @Nullable Map<String, List<String>> headerMap)
	{
		this(status, statusMessage, contentType, contentEncoding, headerMap, false);
	}

	/**
	 * @param status          HTTP status (e.g. 200)
	 * @param statusMessage   HTTP status message (e.g. "OK")
	 * @param contentType     Content-type header (e.g. "text/html")
	 * @param contentEncoding Content-Encoding header (e.g. "gzip")
	 * @param headerMap       A deep copy is made. Maps from lower-case header name to ordered list of values.
	 * @param icyProtocol     If the response is using the ICY procotol instead of HTTP
	 */
	public Headers(int status, @NotNull String statusMessage,
	               @Nullable String contentType,
	               @Nullable String contentEncoding,
	               @Nullable Map<String, List<String>> headerMap,
	               boolean icyProtocol)
	{
		this.status = status;
		this.statusMessage = statusMessage;
		this.isBodylessStatusCode = isBodylessStatusCode(status);
		this.contentType = contentType;
		this.contentTypeBaseLC = makeBaseLcContentType(contentType);
		this.contentEncoding = contentEncoding;
		this.icyProtocol = icyProtocol;
		if (headerMap == null || headerMap.isEmpty()) {
			this.headerMap = Collections.emptyMap();
		} else {
			Map<String, List<String>> map = new HashMap<>(headerMap.size());
			for (String key : headerMap.keySet()) {
				map.put(key, Collections.unmodifiableList(new ArrayList<>(headerMap.get(key))));
			}
			this.headerMap = Collections.unmodifiableMap(map);
		}
	}

	private Headers(@NotNull Map<String, List<String>> headerMap,
	                int status,
	                @NotNull String statusMessage,
	                @Nullable String contentType,
	                @Nullable String contentEncoding,
	                boolean icyProtocol)
	{
		this.status = status;
		this.statusMessage = statusMessage;
		this.isBodylessStatusCode = isBodylessStatusCode(status);
		this.contentType = contentType;
		this.contentTypeBaseLC = makeBaseLcContentType(contentType);
		this.contentEncoding = contentEncoding;
		this.headerMap = headerMap;
		this.icyProtocol = icyProtocol;
	}

	public Headers copyWithContentType(@Nullable String contentType)
	{
		if (StringUtils.triEquals(contentType, this.contentType) && contentEncoding == null)
			return this;
		return new Headers(headerMap, status, statusMessage, contentType, null, icyProtocol);
	}

	public Headers copyWithContentEncoding(@Nullable String contentEncoding)
	{
		if (StringUtils.triEquals(contentEncoding, this.contentEncoding))
			return this;
		return new Headers(headerMap, status, statusMessage, contentType, contentEncoding, icyProtocol);
	}

	/**
	 * Compares content-type to the specified value by ignoring charset and other sub-parameters.
	 * This makes "text/plain" equal to "text/plain; charset=UTF-8"
	 */
	public boolean isContentTypeBaseLC(String ctLC)
	{
		return contentTypeBaseLC.equals(ctLC);
	}

	@Nullable
	public String getLC(String nameLC)
	{
		List<String> list = headerMap.get(nameLC);
		return (list == null || list.isEmpty()) ? null : list.get(0);
	}

	public boolean hasMobileUserAgent()
	{
		String ua = getLC("user-agent");
		// iPod 2 contains "iPhone OS".  iPod 1 doesn't and need explicit iPod.
		return (ua != null) && (ua.contains("iPhone") || ua.contains("(iPod;"));
	}

	/**
	 * Transforms a lowercase name into the proper mixedcase name.
	 * "content-type" is returned as "Content-Type".
	 *
	 * @param nameLC The <b>lowercase</b> header name.
	 */
	@NotNull
	public static String getMixedCaseHeaderName(@NotNull String nameLC)
	{
		String mx = mixedCaseCache.get(nameLC);
		if (mx != null)
			return mx;
		int len = nameLC.length();
		if (nameLC.startsWith("icy-")) {
			// 2009-07-23 email from RGG <rgg@gagafactory.com>: ICY headers must be completely lowercase.
			mx = nameLC.toLowerCase();
		} else {
			StringBuilder sb = new StringBuilder(len);
			boolean uc = true;
			for (int i = 0; i < len; i++) {
				char ch = nameLC.charAt(i);
				if (uc)
					ch = Character.toUpperCase(ch);
				uc = (ch == '-');
				sb.append(ch);
			}
			mx = sb.toString();
		}
		mixedCaseCache.put(nameLC, mx);
		return mx;
	}

	private static final Map<String, String> mixedCaseCache = new ConcurrentHashMap<>();
	static {
		mixedCaseCache.put("x-sessioncookie", "x-sessioncookie"); // QuickTime RTSP.
	}

	private static boolean isBodylessStatusCode(int status)
	{
		return status <= 199
				|| status == HttpURLConnection.HTTP_NOT_MODIFIED
				|| status == HttpURLConnection.HTTP_NO_CONTENT; // rfc 2616, 4.3
	}

	@NotNull
	public static String makeBaseLcContentType(@Nullable String ct)
	{
		if (ct == null)
			return "-none-";
		int idx = ct.indexOf(';');
		if (idx > 0)
			ct = ct.substring(0, idx).trim();
		return ct.toLowerCase();
	}
}
