/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.server;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.InetAddress;
import java.net.SocketException;
import javax.net.ssl.HttpsURLConnection;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.http.client.cache.HttpClientCache;
import org.glimmerblocker.proxy.http.client.stream.DevNullInputStream;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.proxyRequest.ProxyRequestHandler;
import org.glimmerblocker.proxy.util.HttpException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.misc.StringUtils;

public class HttpServerRequest implements Logger
{
	@NotNull public final GlimmerGlobals globals;

	/**
	 * The HTTP request method (from the client) such as "GET".
	 */
	@NotNull public final String method;

	/**
	 * The request PATH, e.g. "/"  or "http://www.ruc.dk" in case of proxy requests.
	 */
	@NotNull public final String path;

	/**
	 * The request headers such as User-Agent
	 */
	@NotNull public final Headers headers;

	/**
	 * Request is using HTTP/1.1  (or later)
	 */
	public final boolean http11;

	/**
	 * Size of POST data.
	 */
	public final long requestBodyContentLength;

	public final String requestTransferEncoding;

	// package-private variables.
	final boolean hasNonEmptyRequestBody;
	volatile boolean hasReadFullRequestBody; // can be updated by another thread.

	Headers responseHeaders;
	long responseBodyContentLength = -1;
	private OutputStream responseOutputStream;
	private boolean hasSentFullResponseBody;

	@NotNull final HttpServerConnection conn;
	private final int requestCount;
	private boolean hasSent100Continue;

	HttpServerRequest(@NotNull HttpServerConnection conn, int requestCount) throws Exception
	{
		this.globals = conn.globals;
		this.conn = conn;
		this.requestCount = requestCount;
		HttpRequestHeaderReader r = new HttpRequestHeaderReader(this);
		this.method = r.method;
		this.path = r.uri;
		this.headers = r.requestHeaders;
		this.http11 = r.http11;
		this.requestBodyContentLength = r.contentLength;
		this.requestTransferEncoding = r.transferEncoding;
		this.hasNonEmptyRequestBody = (requestBodyContentLength > 0) || (requestTransferEncoding != null);
	}

	public void ln(@Nullable String s)
	{
		conn.ln("[" + requestCount + "] " + (s != null ? s : ""));
	}

	void main() throws Exception
	{
		try {
			if (!hasNonEmptyRequestBody)
				hasReadFullRequestBody = true;
			if (method.equals("CONNECT")) {
				ConnectMethodHandler handler = new ConnectMethodHandler(this);
				Headers h = handler.getBlockedHeaders();
				if (h == null) {
					handler.main();
					return;
				}
				conn.disableKeepalive("No keep-alive after blocked CONNECT");
				sendResponseHeaders(h, 0);
			} else {
				ProxyRequestHandler handler = new ProxyRequestHandler(this, globals);
				handler.main();
			}
			if (globals.debugHttpServer && globals.debugVerbose)
				ln("RequestHandler done.");
			if (!hasReadFullRequestBody)
				disableKeepalive("Has not read all of request body when done serving request.");
			if (responseHeaders != null && responseOutputStream == null)
				getOutputStream();
			if (responseOutputStream != null)
				responseOutputStream.close();
		}
		catch (Exception ex) {
			boolean isSocketException = (ex instanceof SocketException);
			if (globals.debugFlag || !isSocketException)
				ln("Got exception handling request: " + ExceptionUtils.getExMsg(ex));
			if (globals.debugVerbose || (globals.debugFlag && !isSocketException))
				ln(ExceptionUtils.getStackTraceAsString(ex));
			conn.disableKeepalive("Got exception handling request");
			if (responseHeaders != null)
				return;
			// we've got some of the request but not yet sent response, send something back.
			Headers h;
			if (ex instanceof HttpException)
				h = ((HttpException)ex).headers;
			else
				h = new Headers(HttpsURLConnection.HTTP_BAD_REQUEST, "Can't read request");
			sendResponseHeaders(h, -1);
		}
		conn.os.flush();
		if (responseHeaders == null)
			conn.disableKeepalive("Has not sent response headers");
		else if (!hasSentFullResponseBody)
			conn.disableKeepalive("Has not sent full response body");
		if (!hasReadFullRequestBody)
			conn.disableKeepalive("Has not read full request body");
		if (globals.debugHttpServer && globals.debugVerbose)
			ln("ServerRequest done.");
	}

	public boolean hasSentResponseHeaders()
	{
		return responseHeaders != null;
	}

	public void disableKeepalive(@NotNull String why)
	{
		conn.disableKeepalive(why);
	}

	public boolean keepaliveUsable()
	{
		return conn.useKeepAlive;
	}

	public InetAddress getClientAddress()
	{
		return conn.skt.getInetAddress();
	}

	void notifyHasSentFullResponseBody()
	{
		if (globals.debugHttpServer && globals.debugVerbose && !hasSentFullResponseBody)
			ln("notifyHasSentFullResponseBody() called.");
		hasSentFullResponseBody = true;
	}

	public void send100Continue(Headers h) throws IOException
	{
		if (h.status != 100)
			throw new IOException("Invalid status = " + h.status + " for sending 100-continue");
		if (hasSent100Continue)
			throw new IOException("Has already sent 100-continue.");
		if (responseHeaders != null)
			throw new IOException("Wanted to send 100-continue after having sent response headers");
		hasSent100Continue = true;
		long contentLength = -1;
		sendResponseHeaders(h, contentLength, globals.debugHttpServer, http11, conn.useKeepAlive, conn.os, this);
	}

	/**
	 * Send the specified HTTP headers to the client.
	 */
	public void sendResponseHeaders(@NotNull Headers h, long contentLength) throws IOException
	{
		if (responseHeaders != null) {
			conn.disableKeepalive("Wanted to send response headers twice.");
			throw new IOException("Has already sent response headers");
		}
		responseHeaders = h;
		if (h.icyProtocol)
			conn.disableKeepalive("ICY response");
		String s = h.getLC("content-disposition");
		if (s != null && s.toLowerCase().contains("attachment")) {
			conn.disableKeepalive("has 'Content-Disposition: attachment': " +
					" disallowed as workaround for semi-hang bug in Cocoa's URLDownloader" +
					" for multiple concurrent downloads from the same server.");
		}
		if (!hasReadFullRequestBody)
			disableKeepalive("Sends response headers before full request body has been read.");
		sendResponseHeaders(h, contentLength, globals.debugHttpServer, http11, conn.useKeepAlive, conn.os, this);
		if (!h.isBodylessStatusCode && contentLength >= 0)
			responseBodyContentLength = contentLength;
		if (contentLength == 0 || h.isBodylessStatusCode)
			hasSentFullResponseBody = true;
	}

	private static void sendResponseHeaders(Headers h,
	                                        long contentLength,
	                                        boolean debugHttpServer,
	                                        boolean http11,
	                                        boolean useKeepAlive,
	                                        OutputStream os,
	                                        Logger logger) throws IOException
	{
		PrintWriter pw;
		StringWriter sw;
		if (debugHttpServer) {
			sw = new StringWriter(400);
			pw = new PrintWriter(sw);
		} else {
			sw = null;
			pw = new PrintWriter(new OutputStreamWriter(new NonFlushingOutputStream(os), "ISO-8859-1"));
		}
		if (h.icyProtocol)
			pw.print("ICY ");
		else if (http11)
			pw.print("HTTP/1.1 ");
		else
			pw.print("HTTP/1.0 ");
		pw.print(h.status);
		pw.print(' ');
		pw.print(h.statusMessage);
		pw.print("\r\n");
		if (h.contentType != null) {
			pw.print("Content-Type: ");
			pw.print(h.contentType);
			pw.print("\r\n");
		}
		if (h.contentEncoding != null) {
			pw.print("Content-Encoding: ");
			pw.print(h.contentEncoding);
			pw.print("\r\n");
		}
		if (!h.isBodylessStatusCode) {
			if (contentLength >= 0) {
				pw.print("Content-Length: ");
				pw.print(contentLength);
				pw.print("\r\n");
			} else if (http11 && !h.icyProtocol && useKeepAlive) {
				pw.print("Transfer-Encoding: chunked\r\n");
			}
		}
		for (String nameLC : h.headerMap.keySet()) {
			String name = Headers.getMixedCaseHeaderName(nameLC);
			for (String value : h.headerMap.get(nameLC)) {
				pw.print(name);
				pw.print(": ");
				pw.print(value);
				pw.print("\r\n");
			}
		}
		if (useKeepAlive) {
			pw.print("Keep-Alive: timeout=5\r\n");
			pw.print("Connection: Keep-Alive\r\n");
			pw.print("Proxy-Connection: Keep-Alive\r\n");
		} else {
			pw.print("Connection: close\r\n");
			pw.print("Proxy-Connection: close\r\n");
		}
		pw.print("\r\n");
		pw.flush();
		if (sw != null) {
			String hdr = sw.toString();
			os.write(hdr.getBytes("ISO-8859-1"));
			logger.ln("sendResponseHeaders:");
			for (String s : StringUtils.splitByChar(hdr.trim(), '\n'))
				logger.ln("  " + s.trim());
		}
	}

	public OutputStream getOutputStream() throws IOException
	{
		if (responseHeaders == null)
			throw new IOException("Can't get body OutputStream when response headers aren't sent yet.");
		if (responseOutputStream != null)
			throw new IOException("Has already created response body OutputStream");
		if (responseHeaders.isBodylessStatusCode)
			responseOutputStream = new HttpServerContentLengthOutputStream(this, 0);
		else if (responseBodyContentLength >= 0)
			responseOutputStream = new HttpServerContentLengthOutputStream(this, responseBodyContentLength);
		else if (http11 && !responseHeaders.icyProtocol && conn.useKeepAlive)
			responseOutputStream = new HttpServerChunkedOutputStream(this);
		else
			responseOutputStream = new HttpServerInfiniteOutputStream(this);
		return responseOutputStream;
	}

	public InputStream getInputStream() throws IOException
	{
		if (!hasNonEmptyRequestBody)
			return new DevNullInputStream();
		if (hasReadFullRequestBody)
			throw new IOException("Has already created request body InputStream.");
		if (requestTransferEncoding != null) {
			if (!requestTransferEncoding.equalsIgnoreCase("chunked"))
				throw new IOException("Can't handle requests with Transfer-Encoding: " + requestTransferEncoding);
			return new HttpServerChunkedInputStream(this);
		}
		if (requestBodyContentLength <= 0)
			return new DevNullInputStream();
		else
			return new HttpServerContentLengthInputStream(this, requestBodyContentLength);
	}

	private static class NonFlushingOutputStream extends OutputStream
	{
		@NotNull final OutputStream os;

		private NonFlushingOutputStream(@NotNull OutputStream os)
		{
			this.os = os;
		}

		public void write(int i) throws IOException
		{
			os.write(i);
		}

		@Override
		public void write(@NotNull byte[] data, int offset, int len) throws IOException
		{
			os.write(data, offset, len);
		}

		@Override
		public void flush() throws IOException
		{
		}
	}

	public HttpClientCache getHttpClientCache()
	{
		return conn.getHttpClientCache();
	}
}
