/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.proxyRequest;

import java.net.URL;
import java.text.Normalizer;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.http.client.UrlStreamer;
import org.glimmerblocker.proxy.http.client.connection.HttpResponseInfo;
import org.glimmerblocker.proxy.http.client.stream.RequestData;
import org.glimmerblocker.proxy.http.client.stream.TimeoutableInputStream;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.HeadersBuilder;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.http.data.UrlParams;
import org.glimmerblocker.proxy.util.HtmlUtils;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.html.URLUtils;
import dk.vitality.lib.foundation.io.FilenameUtils;
import dk.vitality.lib.foundation.misc.StringUtils;

class ForwardStreamAttachmentDownloader extends ForwardStreamerBase
{
	String attachmentFilename;
	Map<String, String> cookies = new HashMap<>();
	boolean hasFilenameInUrl;
	final boolean browserUnderstandsCharsetInContentDisposition;

	ForwardStreamAttachmentDownloader(@NotNull ProxyRequestHandler rh,
	                                  @NotNull Headers headers,
	                                  @NotNull String requestPath) throws Exception
	{
		super(rh, headers);
		browserUnderstandsCharsetInContentDisposition = decodeUA();
		if (!requestPath.equals("/download") && !requestPath.startsWith("/download/"))
			return;
		//
		// Redirects should be followed by the proxy server as it otherwise will be the
		// browser and thereby loosing the filename information as this code wouldn't
		// be executed for the destination URLs.
		followRedirects = true;
		UrlParams sp = new UrlParams();
		int idx = rh.request.path.indexOf('?');
		if (idx >= 0)
			sp.addFromQueryString(rh.request.path.substring(idx + 1));
		attachmentFilename = HtmlUtils.cleanFilename(sp.reqString("filename"));
		hasFilenameInUrl = rh.request.path.endsWith(attachmentFilename);
		// set the real URL we're going to stream instead of the magic blocker URL.
		String s = sp.reqString("url");
		if (globals.debugFlag) {
			ln("attachment url: " + s);
			ln("attachment filename: " + attachmentFilename);
		}
		url = new URL(s);
		String host = url.getHost();
		String cookieHeader = requestHeaders.getLC("cookie");
		if (cookieHeader != null) {
			for (String c : StringUtils.splitByChar(cookieHeader, ';')) {
				int eq = c.indexOf('=');
				String name = (eq < 0) ? c : c.substring(0, eq).trim();
				if (name.startsWith("watched_video_id_list_") && host.endsWith("youtube.com"))
					continue; // ignore: way too big.
				String value = (eq < 0) ? "" : c.substring(eq + 1).trim();
				if (globals.debugFlag)
					ln("Saves initial cookie: '" + name + "' = '" + value + "'");
				cookies.put(name, value);
			}
		}
	}

	private boolean decodeUA()
	{
		String ua = prh.request.headers.getLC("user-agent");
		ln("UA: " + ua);
		if (ua == null)
			return true;
		if (ua.contains("Firefox/")) {
			ln("Firefox => can handle charset in content-disposition");
			return true;
		}
		if (!ua.contains("AppleWebKit/")) {
			ln("Not WebKit nor Firefox => assumes it's able to handle charset in content-disposition");
			return true;
		}
		Matcher m = versionRE.matcher(ua);
		if (!m.find()) {
			ln("WebKit but no Version/xxx => very old version which can't handle charset in content-disposition");
			return false;
		}
		int version = Integer.parseInt(m.group(1));
		boolean ok = version >= 6;
		ln("WebKit version " + version + " => " + (ok ? "able" : "NOT able") + " to handle charset in content-disposition");
		return ok;
	}

	private static final Pattern versionRE = Pattern.compile("\\bVersion/(\\d+)");

	public void execute() throws Exception
	{
		if (attachmentFilename == null) { // favicon etc.
			prh.sendResultData(ResultData.checkerboardPatternGif);
			return;
		}
		super.execute();
	}

	protected void executeOneRequest() throws Exception
	{
		if (globals.debugFlag)
			ln("-----------------------------------------------------");
		super.executeOneRequest();
	}

	protected RequestData makeRequestData() throws Exception
	{
		HeadersBuilder hb = new HeadersBuilder(requestHeaders);
		hb.removeAll("host");
		hb.add("host", url.getHost()); // override the .glimmerblocker.test  host name
		StringBuilder sb = new StringBuilder();
		for (String name : cookies.keySet()) {
			if (sb.length() > 0)
				sb.append("; ");
			sb.append(name).append("=").append(cookies.get(name));
		}
		hb.removeAll("cookie");
		if (globals.debugFlag)
			ln("Request cookies: " + sb);
		if (sb.length() > 0) {
			hb.add("cookie", sb.toString());
		}
		return new RequestData(method, url, hb.toHeaders(), getRequestBody());
	}

	protected boolean handleRedirect() throws Exception
	{
		boolean r = super.handleRedirect();
		List<String> list = r ? responseHeaders.headerMap.get("set-cookie") : null;
		if (list != null) {
			for (String c : list) {
				int semicolon = c.indexOf(';');
				if (semicolon >= 0)
					c = c.substring(0, semicolon).trim();
				int eq = c.indexOf('=');
				String name = (eq < 0) ? c : c.substring(0, eq).trim();
				String value = (eq < 0) ? "" : c.substring(eq + 1).trim();
				if (globals.debugFlag)
					ln("set-cookie in response: '" + name + "' = '" + value + "'");
				cookies.put(name, value);
			}
		}
		return r;
	}

	protected void sendResponse(@NotNull HttpResponseInfo hri,
	                            @Nullable TimeoutableInputStream bodyStream) throws Exception
	{
		AttachmentStreamer streamer = new AttachmentStreamer(globals, prh, hri);
		streamer.streamResponse(bodyStream);
	}

	class AttachmentStreamer extends UrlStreamer
	{
		AttachmentStreamer(@NotNull GlimmerGlobals globals,
		                   @NotNull ProxyRequestHandler prh,
		                   @NotNull HttpResponseInfo rhi)
		{
			super(globals, prh, rhi);
		}

		@Override
		protected Headers getResponseHeaders()
		{
			Headers h = super.getResponseHeaders();
			if (h.status != 200)
				return h;
			if (globals.debugFlag) {
				ln("Attachment source url: " + url);
				ln("  source content-type: " + h.contentType);
			}
			HeadersBuilder hb = new HeadersBuilder(h);
			String reqPath = url.getPath();
			if (reqPath == null || reqPath.length() == 0)
				reqPath = "/";
			String fn = attachmentFilename;
			fn = StringUtils.replaceAll(fn, "&amp;", "&");
			if (h.isContentTypeBaseLC("video/mp4")) {
				fn = FilenameUtils.changeSuffixTo(fn, ".mp4");
			} else if (h.isContentTypeBaseLC("video/flv") || h.isContentTypeBaseLC("video/x-flv")) {
				fn = FilenameUtils.changeSuffixTo(fn, ".flv");
			} else if (h.isContentTypeBaseLC("application/octet-stream") && reqPath.endsWith(".flv")) {
				hb.contentType = "video/x-flv";
				fn = FilenameUtils.changeSuffixTo(fn, ".flv");
			}
			//--------------------------------------------
			// Firefox 3 supports the charset parameter in the C-D header, but Safari 3 doesn't.
			// http://greenbytes.de/tech/tc2231/
			hb.removeAll("content-disposition");
			String cd;
			if (browserUnderstandsCharsetInContentDisposition) {
				if (globals.debugFlag)
					ln("Browser understands charset in content-disposition.");
				cd = "attachment; filename*=UTF-8''" + URLUtils.encodeURL(fn);
			} else {
				ln("fn before strip accents: '" + fn + "'");
				try {
					fn = Normalizer.normalize(fn, Normalizer.Form.NFC);
					ln("fn after normalize accents: '" + fn + "'");
				}
				catch (Throwable ex) {
					// Gets class-not-found when running under java 1.5
					ln("Can't normalize composed chars in download filename: " + ExceptionUtils.getExMsg(ex));
				}
				fn = stripAccents(fn);
				fn = fn.replace('"', ' ').trim();
				ln("fn after strip accents: '" + fn + "'");
				cd = "attachment; filename=\"" + fn + "\"";
				//boolean ascii = printableAsciiRE.matcher(fn).matches();
				//if (globals.debugFlag)
				//	ln("hasFilenameInUrl = " + hasFilenameInUrl + ", fn pure ascii = " + ascii);
				/*
				if (hasFilenameInUrl && !ascii)
					cd = "attachment"; // has filename in URL in this case.
				else
					cd = "attachment; filename=\"" + fn + "\"";
				*/
			}
			hb.add("content-disposition", cd);
			if (globals.debugFlag) {
				ln("   final content-type: " + hb.contentType);
				ln("  content-disposition: " + cd);
				ln("       final filename: " + fn);
			}
			return hb.toHeaders();
		}
	}

	//private static Pattern printableAsciiRE = Pattern.compile("^\\p{ASCII}+$");

	@NotNull
	private static String stripAccents(@NotNull String s)
	{
		int len = s.length();
		StringBuilder sb = new StringBuilder(len);
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch < ' ' || ch == '\'' || ch == '"' || ch > 255 || ch == 127)
				continue;
			sb.append(ch);
		}
		return sb.toString();
	}
}
