/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.proxyRequest;

import java.io.BufferedInputStream;
import java.io.InputStream;
import org.glimmerblocker.proxy.globals.history.HistoryEntryBuilder;
import org.glimmerblocker.proxy.http.client.UrlStreamer;
import org.glimmerblocker.proxy.http.client.cache.HttpClientCache;
import org.glimmerblocker.proxy.http.client.connection.HttpResponseInfo;
import org.glimmerblocker.proxy.http.client.curl.CurlBase;
import org.glimmerblocker.proxy.http.client.stream.DevNullInputStream;
import org.glimmerblocker.proxy.http.client.stream.RequestBody;
import org.glimmerblocker.proxy.http.client.stream.TimeoutableInputStream;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.http.server.HttpServerRequest;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;

abstract class ForwardStreamerBase extends CurlBase
{
	@NotNull final ProxyRequestHandler prh;
	@NotNull final HttpServerRequest request;
	final InputStream requestBodyStream;
	final HistoryEntryBuilder history;

	protected ForwardStreamerBase(@NotNull ProxyRequestHandler prh,
	                              @NotNull Headers requestHeaders) throws Exception
	{
		super(prh.globals, prh, prh.networkConfig, prh.request.getClientAddress(),
				prh.request.method, prh.requestURL, requestHeaders);
		this.prh = prh;
		this.request = prh.request;
		this.history = prh.makeHistoryEntryBuilder();
		InputStream is = request.getInputStream();
		if (is instanceof DevNullInputStream) {
			requestBodyStream = null;
			return;
		}
		int sz;
		if (request.requestBodyContentLength < 0) // chunked, unknown size.
			sz = 4096;
		else
			sz = Math.max(16, (int)Math.min(65536, request.requestBodyContentLength));
		requestBodyStream = new BufferedInputStream(is, sz);
		requestBodyStream.mark(sz);
	}

	@Override
	public void execute() throws Exception
	{
		try {
			super.execute();
		}
		finally {
			history.commit();
		}
	}

	protected void executeWithoutExceptionHandling(HttpClientCache clientCache) throws Exception
	{
		String auth = requestHeaders.getLC("authorization");
		if (auth != null && (auth.startsWith("NTLM ") || auth.startsWith("Negotiate "))) {
			// Yet another Microsoft Extend and Embrace incompability:
			// NTML + Negotiate authentication schemes authenticate the *connection*
			// instead of the client, and the authentication is not valid for requests
			// sent using other connections. So these requests must be sent using
			// a server-connection specific http-client cache to ensure we use the same
			// client-connection for further keep-alive requests.
			// http://www.innovation.ch/personal/ronald/ntlm.html
			// http://davenport.sourceforge.net/ntlm.html
			clientCache = request.getHttpClientCache();
			if (globals.debugVerbose)
				ln("Uses server-connection specific client-connection cache.");
		}
		super.executeWithoutExceptionHandling(clientCache);
	}

	protected RequestBody getRequestBody() throws Exception
	{
		if (redirectCount > 0)
			return RequestBody.NONE;
		if (requestBodyStream != null)
			requestBodyStream.reset();
		return new RequestBody(requestBodyStream, request.requestBodyContentLength);
	}

	protected void readBody(@NotNull HttpResponseInfo hri,
	                        @Nullable TimeoutableInputStream bodyStream) throws Exception
	{
		if (!globals.filtersEnabled) {
			streamResponse(hri, bodyStream);
			return;
		}
		if (bodyStream == null && !hri.isContentLessResponse) {
			ln("Missing bodyStream for non-body-less response: " + responseHeaders.status);
			if (globals.debugFlag)
				ln(ExceptionUtils.getStackTraceAsString());
		}
		sendResponse(hri, bodyStream);
	}

	protected abstract void sendResponse(@NotNull HttpResponseInfo hri,
	                                     @Nullable TimeoutableInputStream bodyStream) throws Exception;

	protected void sendRequestRuleData(@NotNull ResultData data) throws Exception
	{
		prh.sendResultData(data);
	}

	protected void streamResponse(@NotNull HttpResponseInfo hri,
	                              @Nullable TimeoutableInputStream bodyStream) throws Exception
	{
		UrlStreamer streamer = new UrlStreamer(globals, prh, hri);
		streamer.streamResponse(bodyStream);
	}

	@Nullable
	protected HistoryEntryBuilder getHistoryEntryBuilderForRequestRules()
	{
		return history;
	}

	protected void handle100ContinueResponse(@NotNull HttpResponseInfo hri) throws Exception
	{
		if (globals.debugHttpServer)
			ln("Interjects 100-Continue header.");
		request.send100Continue(hri.headers);
	}

	protected boolean isMobileUserAgent()
	{
		return prh.request.headers.hasMobileUserAgent();
	}
}
