/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.proxyRequest.transform;

import java.util.ArrayList;
import java.util.List;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.history.HistoryEntryBuilder;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.http.client.curl.CurlBase;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.proxyRequest.ProxyRequestHandler;
import org.glimmerblocker.proxy.rhino.callback.HtmlHelper;
import org.glimmerblocker.proxy.rhino.runtime.ExecutorDelegate;
import org.glimmerblocker.proxy.rule.ModificationRule;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.RhinoException;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;
import dk.vitality.lib.foundation.exception.ExceptionUtils;

public class Transformer
{
	@NotNull public final GlimmerGlobals globals;
	@NotNull public ResultData rd;
	public final boolean didTransformContent;

	@NotNull private final Logger logger;
	@NotNull private final CurlBase curlBase;
	@NotNull private ProxyRequestHandler prh;
	@NotNull private final List<TransformErrorInfo> errorList = new ArrayList<>(0);
	@NotNull private final HistoryEntryBuilder history;

	public Transformer(@NotNull GlimmerGlobals globals,
	                   @NotNull Logger parentLogger,
	                   @NotNull List<ModificationRule> transformerList,
	                   @NotNull ProxyRequestHandler prh,
	                   @NotNull CurlBase curlBase,
	                   @NotNull Headers headers,
	                   @NotNull String text,
	                   @NotNull HistoryEntryBuilder history) throws Exception
	{
		this.globals = globals;
		this.logger = new PrefixedLogger(parentLogger, "transform");
		this.prh = prh;
		this.curlBase = curlBase;
		this.rd = new ResultData(forceCharset(headers), text);
		this.history = history;
		boolean any = false;
		for (ModificationRule rule : transformerList) {
			if (rule.transformContentTypeMatcher == null) {
				if (globals.debugFilters && globals.debugVerbose)
					ln("Rule ignored, has no content-type matcher: " + rule.toString());
				continue;
			}
			if (!rule.transformContentTypeMatcher.matches(headers.status, headers.contentTypeBaseLC)) {
				if (globals.debugFilters && globals.debugVerbose) {
					ln("Rule doesn't match content-type '" + headers.contentType + "'" +
							" using " + rule.transformContentTypeMatcher.name() + ": " + rule.toString());
				}
				continue;
			}
			any = true;
			new Delegate(rule);
		}
		if (errorList.isEmpty()
				|| !globals.showFilterTransformErrorHtmlBox
				|| !MimeTypeMatcher.Html.matches(rd.headers.status, rd.headers.contentTypeBaseLC)) {
			didTransformContent = any;
			return;
		}
		HtmlHelper hc = new HtmlHelper(rd.getContentAsText(logger, globals.debugVerbose));
		hc.insertAtBodyEnd(new TransformErrorReport(errorList).getErrorHtml(), true);
		this.rd = new ResultData(forceCharset(headers), hc.html);
		didTransformContent = true;
	}

	private Headers forceCharset(@NotNull Headers headers)
	{
		//return headers;
		return headers.copyWithContentType(headers.contentTypeBaseLC + "; charset=UTF-8");
	}

	public void ln(@Nullable String s)
	{
		logger.ln(s);
	}

	class Delegate implements ExecutorDelegate<TransformerGlobalObject>
	{
		@NotNull final ModificationRule rule;

		Delegate(@NotNull ModificationRule rule)
		{
			this.rule = rule;
			try {
				if (globals.debugFilters)
					ln("Runs transformer script: " + rule.toString());
				globals.rhinoRuntimeCache.execute(this, rule.runtimeId, rule.transform, rule.toString());
			}
			catch (Throwable ex) {
				history.addAction(rule, "transform-failed");
				TransformErrorInfo info = new TransformErrorInfo(rule, ex);
				errorList.add(info);
				ln("Transform filter failed with exception: " + ExceptionUtils.getExMsg(ex));
				ln("   " + ex.getClass().getName());
				ln("   filter: " + rule.filterName);
				if (ex instanceof RhinoException) {
					RhinoException rex = (RhinoException)ex;
					ln("  line " + rex.lineNumber() + ", col " + rex.columnNumber());
				}
				ln("   " + rule);
				if (globals.debugFilters)
					ln(ExceptionUtils.getStackTraceAsString(ex));
			}
		}

		@NotNull
		public TransformerGlobalObject create(@NotNull ScriptableObject scope,
		                                      @NotNull Scriptable defaultPrototype) throws Exception
		{
			return new TransformerGlobalObject(globals, scope, defaultPrototype);
		}

		public void prepare(@NotNull TransformerGlobalObject callback, @NotNull Context ctx) throws Exception
		{
			String t = rd.getContentAsText(logger, globals.debugVerbose);
			callback.prepareTransformerCallback(prh, ctx, t, prh, curlBase, rd.headers);
		}

		public void executionDone(@NotNull TransformerGlobalObject callback) throws Exception
		{
			ResultData rd2 = callback.getOptResultData();
			if (rd2 != null)
				rd = rd2;
			history.addAction(rule, "transformed");
		}
	}
}
