/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.rhino.callback;

import java.io.InputStreamReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.glimmerblocker.proxy.globals.log.DatePrefixedLogger;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.jetbrains.annotations.NotNull;
import dk.vitality.lib.foundation.misc.StringUtils;

public class SimpleHtmlParser extends PrefixedLogger
{
	@NotNull final String html;
	int idx;
	final List<StackElement> stack = new ArrayList<>();

	static final boolean debugFlag = false;

	public SimpleHtmlParser(@NotNull Logger logger, @NotNull String html, int idx)
	{
		super(logger, "SHP");
		this.html = html;
		this.idx = idx;
	}

	public String parseSubTree()
	{
		int start = idx;
		skipSubTree();
		int end = Math.min(idx, html.length());
		return html.substring(start, end);
	}

	public void skipSubTree()
	{
		while (idx < html.length()) {
			while (eat() != '<') {
				if (idx >= html.length())
					return;
			}
			skipWhitespace();
			char ch = eat();
			switch (ch) {
				case '/':
					boolean hadStack = !stack.isEmpty();
					parseElementCloseTag();
					if (hadStack && stack.isEmpty())
						return;
					continue;
				case '!':
					parseMarkupDeclaration();
					continue;
			}
			if ((ch < 'a' || ch > 'z') && (ch < 'A' || ch > 'Z'))
				continue; // invalid tag name.
			idx--; // un-eat so the char becomes part of tagName.
			int start = idx - 1;
			String s = parseTagName();
			parseElementAttributes();
			boolean empty = (peek() == '/') || emptyTags.contains(s);
			if (idx < html.length() && html.charAt(idx) == '>') // soft-skip
				idx++;
			if (debugFlag) {
				ln("Got element (" + s + (empty ? ".empty" : "") +
						"): {{{" + html.substring(start, idx) + "}}}");
			}
			if (!s.equals("script") && !s.equals("style")) {
				if (!empty)
					stack.add(new StackElement(s));
				continue;
			}
			// special-case <script> and <style> by simple text-search for end-tag.
			int x = html.indexOf("</" + s, idx);
			if (x < 0) {
				ln("Didn't find end-tag for <" + s + ">");
				idx = html.length();
			} else {
				idx = x + 2 + s.length();
				skipWhitespace();
				if (peek() == '>')
					idx++;
				if (debugFlag)
					ln("Skips " + (idx - start) + " bytes <" + s + ">...</" + s + ">");
			}
		}
	}

	private void parseElementCloseTag()
	{
		int start = idx - 2;
		skipWhitespace();
		String s = parseTagName();
		while (idx < html.length() && html.charAt(idx) != '>')
			idx++;
		idx++;
		if (debugFlag)
			ln("Got end tag: '" + s + "' = {{{" + html.substring(start, idx) + "}}}");
		if (!emptyTags.contains(s)) {
			StackElement se = new StackElement(s);
			while (!stack.isEmpty()) {
				int p = stack.get(stack.size() - 1).closePriority;
				if (p < se.closePriority)
					break;
				StackElement popped = stack.remove(stack.size() - 1);
				if (debugFlag) {
					ln("     </" + s + " = " + se.closePriority + "> popped " +
							"<" + popped.tag + " = " + popped.closePriority + ">");
				}
				if (p == se.closePriority)
					break;
			}
		}
	}

	private void parseMarkupDeclaration()
	{
		int start = idx - 2;
		if (peek() == '[' && html.substring(idx).startsWith("[CDATA[")) {
			int x = html.indexOf("]]>", idx);
			if (x < 0) {
				ln("Didn't find end of <![CDATA[");
				idx = html.length();
			} else {
				idx = x + 3;
			}
		} else if (peek() == '-' && peek(1) == '-') { // <!--  -->  comment.
			idx += 2;
			while (idx < html.length() && (peek() != '-' || peek(1) != '-' || peek(2) != '>'))
				idx++;
			idx += 3; // eat end tag.
		} else {
			while (idx < html.length() && peek() != '>')
				idx++;
		}
		if (debugFlag)
			ln("Skipped <!..>: {{{" + html.substring(start, idx) + "}}}");
	}

	private String parseTagName()
	{
		skipWhitespace();
		StringBuilder sb = new StringBuilder();
		while (idx < html.length()) {
			char ch = html.charAt(idx);
			if (ch == '>' || ch == '/' || Character.isWhitespace(ch))
				break;
			idx++;
			sb.append(Character.toLowerCase(ch));
		}
		return sb.toString();
	}

	private void parseElementAttributes()
	{
		while (idx < html.length()) {
			char ch = html.charAt(idx);
			if (ch == '/' || ch == '>')
				return;
			idx++;
			if (ch != '=')
				continue;
			skipWhitespace();
			if (idx >= html.length())
				break;
			ch = html.charAt(idx++);
			if (ch != '\'' && ch != '"')
				continue;
			while (idx < html.length() && html.charAt(idx) != ch)
				idx++;
			idx++; // skip past end quote.
		}
	}

	private void skipWhitespace()
	{
		while (idx < html.length() && Character.isWhitespace(html.charAt(idx)))
			idx++;
	}

	private char eat()
	{
		char ch = peek();
		idx++;
		return ch;
	}

	private char peek()
	{
		return (idx < html.length()) ? html.charAt(idx) : ' ';
	}

	private char peek(int delta)
	{
		int x = idx + delta;
		return (x >= 0 && x < html.length()) ? html.charAt(x) : ' ';
	}


	static final Set<String> emptyTags
			= makeSet("area,base,basefont,br,col,frame,hr,img,input,isindex,link,meta,param");

	static Set<String> makeSet(@NotNull String list)
	{
		Set<String> set = new HashSet<>();
		for (String s : StringUtils.splitByChar(list, ','))
			set.add(s);
		return Collections.unmodifiableSet(set);
	}

	static class StackElement
	{
		final String tag;
		final int closePriority;

		StackElement(String tag)
		{
			this.tag = tag;
			Integer p = closePriorityMap.get(tag);
			this.closePriority = (p != null) ? p : defaultClosePriority;
		}
	}

	static final Map<String, Integer> closePriorityMap;
	static final int defaultClosePriority;

	static {
		Map<String, Integer> map = new HashMap<>();
		int p = 0;
		addClosePriority(++p, "form", map);
		addClosePriority(++p, "table", map);
		addClosePriority(++p, "td,th", map);
		addClosePriority(++p, "div,p", map);
		defaultClosePriority = ++p;
		addClosePriority(++p, "span,nobr", map);
		addClosePriority(++p, "b,font,i", map);
		closePriorityMap = Collections.unmodifiableMap(map);
	}

	private static void addClosePriority(int level, @NotNull String names, @NotNull Map<String, Integer> map)
	{
		for (String s : StringUtils.splitByChar(names, ','))
			map.put(s, level);
	}

	public static void main(String[] args)
	{
		try {
			StringBuilder sb = new StringBuilder();
			int x;
			Reader r = new InputStreamReader(System.in);
			while ((x = r.read()) >= 0)
				sb.append((char)x);
			Logger logger = new DatePrefixedLogger();
			logger.ln("Read " + sb.length() + " chars input.");
			SimpleHtmlParser shp = new SimpleHtmlParser(logger, sb.toString(), 0);
			String s = shp.parseSubTree();
			System.err.print("SUBTREE: {{{");
			System.out.print(s);
			System.out.flush();
			System.err.println("}}}");
		}
		catch (Exception ex) {
			System.err.println("Got exception: " + ex.getMessage());
			ex.printStackTrace();
		}
	}
}
