/* Copyright (C) 2008-2012 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <execinfo.h>
#include <sys/stat.h>
#include <CommonCrypto/CommonDigest.h>
#import <ExceptionHandling/NSExceptionHandler.h>
#include "GBDebug.h"
#include <pthread.h>

const char* glimmerAuthString = "org.glimmerblocker.panelprefs";

NSString *PROXY_LAUNCHD_ID = @"org.glimmerblocker.proxy";
NSString *UPDATER_LAUNCHD_ID = @"org.glimmerblocker.installer";

NSString *INSTALLER_BUNDLE_ID = @"org.glimmerblocker.installer";

NSString *LAST_UPDATE_CHECK_MARKER_PATH = @"/Library/GlimmerBlocker/LastUpdateCheck/Marker.txt";

NSString *CHANNEL_DEV_PATH = @"/Library/GlimmerBlocker/UpdaterUseDevChannel.txt";
NSString *CHANNEL_BETA_PATH = @"/Library/GlimmerBlocker/UpdaterUseBetaChannel.txt";

// http://www.cocoadev.com/index.pl?StackTraces

@implementation GBDebug

#pragma mark ------------------- stack frame dump

+ (void)debugDumpExceptionStackframe:(NSException*)ex
{
	NSLog(@"stacktrace dump:\n%@", [ex callStackSymbols]);
}

+ (void)debugDumpStackframe
{
	@try {
		[NSException raise:@"dummy" format:@"my format"];
	}
	@catch (NSException* ex) {
		[GBDebug debugDumpExceptionStackframe:ex];
	}
}

#pragma mark ------------------- codesign

+ (BOOL)verifyCodesignForBundle:(NSString*)identifier
						 atPath:(NSString*)path
{
	NSString *requirements = [NSString stringWithFormat:@"-R=identifier %@ and certificate leaf = %@",
							  identifier, @"H\"c9d85c0c3b7a8d1cb8060aa34df69713cfff637c\""];
	NSArray *args = @[@"--verify", requirements, path];
	NSMutableString *s = [NSMutableString stringWithCapacity:10];
	int status = [GBDebug executeCommandlineTool:@"codesign" withArgs:args stdoutString:s];
	NSLog(@"codesign: %@", s);
	if (status) {
		NSLog(@"codesign verification FAILED (%d): %@", status, requirements);
		NSLog(@"  for '%@' at path: %@", identifier, path);
	}
	return !status;
}

+ (BOOL)verifyCodesign
{
	// be paranoid about "man" leaving files all over the place.
	unlink("/Library/PreferencePanes/GlimmerBlocker.prefPane/Contents/Resources/.lesshst");
	return true
	&& [GBDebug verifyCodesignForBundle:@"org.glimmerblocker.installer"
								 atPath:[GBDebug updaterBundlePath]]
	&& [GBDebug verifyCodesignForBundle:@"org.glimmerblocker.proxyApp"
								 atPath:@"/Library/PreferencePanes/GlimmerBlocker.prefPane/Contents/GlimmerBlockerProxy.app"]
	&& [GBDebug verifyCodesignForBundle:@"org.glimmerblocker.prefsPane"
								 atPath:@"/Library/PreferencePanes/GlimmerBlocker.prefPane"];
}

+ (NSData*)md5:(NSData*)data
{
	unsigned char md[CC_MD5_DIGEST_LENGTH];
	CC_MD5([data bytes], (int)[data length], md);
	return [NSData dataWithBytes:md length:CC_MD5_DIGEST_LENGTH];
}

+ (NSString*)md5Hex:(NSData*)data
{
	unsigned char md[CC_MD5_DIGEST_LENGTH];
	CC_MD5([data bytes], (int)[data length], md);
	return [NSString stringWithFormat:@"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
			md[0], md[1], md[2], md[3], md[4], md[5], md[6], md[7],
			md[8], md[9], md[10], md[11], md[12], md[13], md[14], md[15]
			];
}

#pragma mark ------------------- launchd

+ (NSString*)proxyPlistPath
{
	return [NSString stringWithFormat:@"/Library/LaunchDaemons/%@.plist", PROXY_LAUNCHD_ID];
}

+ (NSString*)updaterPlistPath
{
	return [NSString stringWithFormat:@"/Library/LaunchAgents/%@.plist", UPDATER_LAUNCHD_ID];
}

+ (NSString*)updaterBundlePath
{
	return @"/Library/PreferencePanes/GlimmerBlocker.prefPane/Contents/GlimmerBlockerInstaller.app";
}

+ (NSString*)proxyBundlePath
{
	return @"/Library/PreferencePanes/GlimmerBlocker.prefPane/Contents/GlimmerBlockerProxy.app";
}

+ (NSString*)proxyBinaryPath
{
	return [[self proxyBundlePath] stringByAppendingString:@"/Contents/MacOS/GlimmerBlockerProxy" ];
}

#pragma mark ------------------- URLs

+ (NSString*)percentEscapeUrlParameterValue:(NSString*)s
{
	NSString* e = [s stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding];
	// '&' isn't encoded by stringByAddingPercentEscapesUsingEncoding
	e = [e stringByReplacingOccurrencesOfString:@"&" withString:@"%26"];
	e = [e stringByReplacingOccurrencesOfString:@"+" withString:@"%2b"];
	return e;
}

#pragma mark ------------------- GUI

+ (void)setPop:(NSPopUpButton*)popup
		values:(__strong NSString*[])values
 selectedValue:(NSString*)selectedValue
  defaultValue:(NSString*)defaultValue
{
	NSString* sv = selectedValue ? selectedValue : defaultValue;
	for (int i = 0; values[i]; i++) {
		if ([values[i] isEqual:sv]) {
			[popup selectItemAtIndex:i];
			return;
		}
	}
	NSLog(@"Did not find popup item, selectedValue = '%@', defaultValue = '%@'", selectedValue, defaultValue);
	for (int i = 0; values[i]; i++)
		NSLog(@"  values[%d] = '%@'", i, values[i]);
}

+ (void)setVertMatrix:(NSMatrix*)matrix
			   values:(__strong NSString*[])values
		selectedValue:(NSString*)selectedValue
		 defaultValue:(NSString*)defaultValue
{
	NSString* sv = selectedValue ? selectedValue : defaultValue;
	for (int i = 0; values[i]; i++) {
		if ([values[i] isEqual:sv]) {
			[matrix deselectAllCells];
			NSInteger rows = [matrix numberOfRows];
			[matrix selectCellAtRow:(i % rows) column:(i / rows)];
			return;
		}
	}
	NSLog(@"Did not find popup item, selectedValue = '%@', defaultValue = '%@'", selectedValue, defaultValue);
	for (int i = 0; values[i]; i++)
		NSLog(@"  values[%d] = '%@'", i, values[i]);
}

#pragma mark ------------------- cmd line

static inline BOOL isShellSafe(char ch)
{
	return (ch >= 'A' && ch <= 'Z')
	|| (ch >= 'a' && ch <= 'z')
	|| (ch >= '0' && ch <= '9')
	|| ch == '.' || ch == '-' || ch == '/' || ch == '=' || ch == '_' || ch == '+' || ch == '@' || ch == ':';
}

+ (NSString*)shellQuotedString:(NSString*)src
{
	if (!src)
		return @"''";
	NSUInteger len = [src length];
	BOOL needQuoting = NO;
	for (int i = 0; i < len; i++) {
		char ch = [src characterAtIndex:i];
		if (isShellSafe(ch))
			continue;
		needQuoting = YES;
		break;
	}
	if (!needQuoting)
		return src;
	NSMutableString *sb = [NSMutableString stringWithCapacity:len + 5];
	[sb appendString:@"'"];
	for (int i = 0; i < len; i++) {
		char ch = [src characterAtIndex:i];
		if (ch < 32 || ch == '\'')
			[sb appendString:@"\\"];
		[sb appendFormat:@"%c", ch];
	}
	[sb appendString:@"'"];
	return [NSString stringWithString:sb];
}

+ (NSString*)shellQuoted:(NSArray*)a
{
	NSUInteger num = [a count];
	if (!num)
		return @"";
	NSMutableString *sb = [NSMutableString stringWithCapacity:50];
	for (int i = 0; i < num; i++) {
		if (i > 0)
			[sb appendString:@" "];
		[sb appendString:[self shellQuotedString:a[i]]];
	}
	return [NSString stringWithString:sb];
}

+ (NSString*)findCommandAbsPath:(NSString*)name usingPathList:(NSString*)pathList
{
	NSFileManager *fm = [NSFileManager defaultManager];
	for (NSString *dir in [pathList componentsSeparatedByString:@":"]) {
		if (!dir.length)
			continue;
		NSString *p2 = [NSString stringWithFormat:@"%@/%@", dir, name];
		if ([fm fileExistsAtPath:p2]) {
			//NSLog(@"found %@ as %@", name, p2);
			return p2;
		}
	}
	return NULL;
}

+ (NSString*)findCommandAbsPath:(NSString*)name
{
	if ([name hasPrefix:@"/"])
		return name;
	char* pathList = getenv("PATH");
	NSString *path = pathList ? [GBDebug findCommandAbsPath:name usingPathList:@(pathList)] : NULL;
	if (path)
		return path;
	path = [GBDebug findCommandAbsPath:name usingPathList:@"/bin:/usr/bin:/sbin:/usr/sbin:"];
	if (path)
		return path;
	NSLog(@"Can't find location of command %@ in any std path", name);
	return name;
}

+ (int)executeCommandlineTool:(NSString*)cmd
					 withArgs:(NSArray*)args
{
	return [GBDebug executeCommandlineTool:cmd withArgs:args optStdinData:NULL optStdoutData:NULL];
}

/*
+ (int)executeCommandlineTool:(NSString*)cmd
					 withArgs:(NSArray*)args
				gettingStdOut:(NSString**)stdoutText
{
	if (stdoutText && *stdoutText)
		[NSException raise:@"InvalidStdoutParam" format:@"stdout param must be NULL or point to NULL"];
	NSMutableData* outputData = stdoutText ? [NSMutableData dataWithCapacity:250] : NULL;
	int status = [GBDebug executeCommandlineTool:cmd withArgs:args optStdinData:NULL optStdoutData:outputData];
	if (stdoutText)
		*stdoutText = [[NSString alloc] initWithData:outputData encoding:NSUTF8StringEncoding];
	return status;
}
 */

+ (int)executeCommandlineTool:(NSString*)cmd
					 withArgs:(NSArray*)args
				stdoutString:(NSMutableString*)stdoutString
{
	NSMutableData *stdout = stdoutString ? [NSMutableData dataWithCapacity:100] : NULL;
	int x = [GBDebug executeCommandlineTool:cmd withArgs:args optStdinData:NULL optStdoutData:stdout];
	[stdoutString appendString:[[NSString alloc] initWithData:stdout encoding:NSUTF8StringEncoding]];
	return x;
}


typedef void (^GbFileHandleObserver)(NSNotification*);

+ (int)executeCommandlineTool:(NSString*)cmd
					 withArgs:(NSArray*)args
				 optStdinData:(NSData*)stdinData
				optStdoutData:(NSMutableData*)stdoutData
{
	cmd = [GBDebug findCommandAbsPath:cmd];
	BOOL stdinNeeded = !![stdinData length];
	//
	NSPipe* stdoutPipe = [[NSPipe alloc] init];
	NSFileHandle *stdoutFileHandle = [stdoutPipe fileHandleForReading];
	//
	NSPipe* stderrPipe = [[NSPipe alloc] init];
	[[stderrPipe fileHandleForReading] readToEndOfFileInBackgroundAndNotify];
	//
	NSPipe* pipeStdin = NULL;
	NSFileHandle *fileHandleStdin = NULL;
	//
	NSTask *task = [[NSTask alloc] init];
	task.launchPath = cmd;
	task.standardOutput = stdoutPipe;
	task.standardError = stderrPipe;
	task.arguments = args;
	//NSLog(@"Runs: %@ %@", task.launchPath, [task.arguments componentsJoinedByString:@" "]);
	NSNotificationCenter *nc = [NSNotificationCenter defaultCenter];
	GbFileHandleObserver stderrObserver = ^(NSNotification *notification){
		NSData *myData = [notification.userInfo objectForKey:NSFileHandleNotificationDataItem];
		NSString *s = [[NSString alloc] initWithData:myData encoding:NSUTF8StringEncoding];
		NSLog(@"%@", s);
	};
	[nc addObserverForName:NSFileHandleReadToEndOfFileCompletionNotification
					object:[stderrPipe fileHandleForReading]
					 queue:nil
				usingBlock:stderrObserver];
	if (stdinNeeded) {
		pipeStdin = [[NSPipe alloc] init];
		fileHandleStdin = [pipeStdin fileHandleForWriting];
		task.standardInput = pipeStdin;
	}
	//NSLog(@"Launches task");
	@try {
		[task launch];
	}
	@catch (NSException* ex) {
		[GBDebug debugDumpExceptionStackframe:ex];
		NSLog(@"Got exception running command line tool %@: %@ / %@", cmd, [ex name], [ex reason]);
		return -1;
	}
	if (stdinNeeded) {
		//NSLog(@"Sets stdin (%ld bytes)", [stdinData length]);
		[fileHandleStdin writeData:stdinData];
		[fileHandleStdin closeFile];
	}
	//NSLog(@"Reads stdout");
	NSData *data = [stdoutFileHandle readDataToEndOfFile];
	//NSLog(@"Got %ld bytes stdout", [data length]);
	if (stdoutData)
		[stdoutData appendData:data];
	//
	[task waitUntilExit];
	return [task terminationStatus];
}

#pragma mark ------------------- saving

+ (void)safeSaveFileAtPath:(NSString *)path contents:(NSData *)data;
{
	NSString* tmp = [path stringByAppendingString:@".tmp"];
	NSFileManager* fm = [NSFileManager defaultManager];
	NSDictionary* attrs = @{NSFilePosixPermissions: @0644L};
	[fm createFileAtPath:tmp contents:[NSData data] attributes:attrs];
	NSFileHandle* fh = [NSFileHandle fileHandleForWritingAtPath:tmp]; // can only write to existing files!
	[fh writeData:data];
	[fh synchronizeFile];
	[fh closeFile];
	for (int i = 4; i >= 0; i--) {
		NSString* p1 = i ? [NSString stringWithFormat:@"%@.old-%d", path, i] : path;
		NSString* p2 = [NSString stringWithFormat:@"%@.old-%d", path, i + 1];
		NSError*__autoreleasing error = NULL;
		if ([fm fileExistsAtPath:p2] && ![fm removeItemAtPath:p2 error:&error])
			NSLog(@" FAILED ZAPPING %@", p2);
		if ([fm fileExistsAtPath:p1] && ![fm moveItemAtPath:p1 toPath:p2 error:NULL]) {
			NSLog(@"FAILED RENAMING %@", p1);
			NSLog(@"             TO %@", p2);
		}
	}
	if (![fm moveItemAtPath:tmp toPath:path error:NULL]) {
		NSLog(@"FAILED RENAMING %@", tmp);
		NSLog(@"             TO %@", path);
	}
}


#pragma mark ------------------- inode fingerprint

+ (NSString*)fingerprintOptionForInstalledPanel:(NSBundle*)prefsPanelBundle
{
	NSString* vers = [prefsPanelBundle infoDictionary][@"CFBundleVersion"];
	return [NSString stringWithFormat:@"<string>--gb-version=%@</string>",
			vers ? vers : @"unknown"];
}

+ (BOOL)fingerprintForRunningServerIsValid:(NSBundle*)prefsPanelBundle
{
	NSString* jobLabel = @"org.glimmerblocker.proxy";
	NSString* plistPath = [NSString stringWithFormat:@"/Library/LaunchDaemons/%@.plist", jobLabel];
	NSFileManager* fm = [NSFileManager defaultManager];
	if (![fm fileExistsAtPath:plistPath])
		return YES;
	NSData* data = [fm contentsAtPath:plistPath];
	if (!data)
		return NO;
	NSString* expected = [GBDebug fingerprintOptionForInstalledPanel:prefsPanelBundle];
	NSString* s = [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
	BOOL ok = [s rangeOfString:expected].location != NSNotFound;
	if (!ok)
		DebugNSLog(@"Did not have fingerprint {{%@}} in %@", expected, s);
	NSLog(@"Fingerprint for running server is %@.", ok ? @"ok" : @"missing/bad");
	return ok;
}

@end
