/* Copyright (C) 2008-2012 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "Xml.h"

@implementation Xml

+ (BOOL)isSimpleName:(NSString*)name startIndex:(int)startIndex
{
	NSInteger len = [name length];
	for (int i = startIndex; i < len; i++) {
		unichar ch = [name characterAtIndex:i];
		if (ch == '-' || (ch >= '0' && ch <= '9') || (ch >= 'a' && ch <= 'z'))
			continue;
		return NO;
	}
	return len > 0;
}

+ (BOOL)isSimpleName:(NSString*)name
{
	return [Xml isSimpleName:name startIndex:0];
}

#pragma mark -------------- Create elements

+ (NSXMLElement*)createElementWithName:(NSString*)name
{
	return (NSXMLElement*)[NSXMLNode elementWithName:name];
}

#pragma mark -------------- Singletons

+ (NSXMLElement*)singletonElementForXPath:(NSString*)xpath
								forParent:(NSXMLElement*)parentE
{
	return [self singletonElementForXPath:xpath forParent:parentE asFirstChild:NO];
}

+ (NSXMLElement*)singletonElementForXPath:(NSString*)xpath
								forParent:(NSXMLElement*)parentE
							 asFirstChild:(BOOL)asFirstChild
{
	NSArray* pathItems = [xpath componentsSeparatedByString:@"/"];
	NSUInteger compCount = [pathItems count];
	NSXMLElement* e = parentE;
	for (NSUInteger compIdx = 0; true; compIdx++) {
		NSString* s = pathItems[compIdx];
		if (![Xml isSimpleName:s]) {
			[NSException raise:@"Invalid element name in singleton xpath"
						format:@"xpath = '%@'", xpath];
		}
		BOOL leaf = (compIdx == compCount - 1);
		NSArray* childs = [e elementsForName:s];
		NSInteger numChilds = [childs count];
		if (leaf && numChilds > 1) {
			for (int x = 1; x < numChilds; x++)
				[childs[compIdx] detach];
		}
		if (numChilds) {
			e = childs[0];
		} else {
			NSXMLElement* newE = [Xml createElementWithName:s];
			if (leaf && asFirstChild)
				[e insertChild:newE atIndex:0];
			else
				[e addChild:newE];
			e = newE;
		}
		if (leaf)
			return e;
	}
	assert(false);
	//[NSException raise:@"shouldn't fell through" format:@"xpath = '%@'", xpath];
}

+ (NSXMLElement*)singletonDateElement:(NSDate*)date
							 forXPath:(NSString*)xpath
							fromParent:(NSXMLElement*)parentE
					   updateExisting:(BOOL)updateExisting
{
	NSXMLElement* e = [Xml singletonElementForXPath:xpath forParent:parentE asFirstChild:YES];
	if (!updateExisting && [Xml getAttribute:e withName:@"msecs"])
		return e;
	NSString* msecs = [NSString stringWithFormat:@"%.0f", [date timeIntervalSince1970] * 1000];
	[Xml setAttribute:msecs withName:@"msecs" inElement:e];
	[Xml setAttribute:[date description] withName:@"iso" inElement:e];
	return e;
}

#pragma mark -------------- Attributes

+ (NSString*)getAttribute:(NSXMLElement*)e withName:(NSString*)name
{
	return [[e attributeForName:name] stringValue];
}

+ (int)getBoolAttribute:(NSXMLElement*)e withName:(NSString*)name
{
	return [Xml getIntAttribute:e withName:name] ? YES : NO;
}

+ (int)getIntAttribute:(NSXMLElement*)e withName:(NSString*)name
{
	return [[Xml getAttribute:e withName:name] intValue];
}

+ (void)setAttribute:(NSString*)value withName:(NSString*)name inElement:(NSXMLElement*)e
{
	NSXMLNode* attr = [e attributeForName:name];
	if ([value length]) {
		if (![[attr stringValue] isEqual:value]) {
			[attr detach];
			[e addAttribute:[NSXMLNode attributeWithName:name stringValue:value]];
		}
	} else {
		[attr detach];
	}
}

+ (void)setBool:(BOOL)value withName:(NSString*)name inElement:(NSXMLElement*)e
{
	[Xml setAttribute:(value ? @"1" : NULL) withName:name inElement:e];
}

+ (void)setInt:(int)value withName:(NSString*)name inElement:(NSXMLElement*)e
{
	[Xml setAttribute:[NSString stringWithFormat:@"%d", value] withName:name inElement:e];
}

+ (void)reorderChilds:(NSArray*)childs
{
	if ([childs count] <= 1)
		return;
	for (NSUInteger i = 0; i < [childs count]; i++) {
		NSXMLElement *childE = childs[i];
		if (!childE) {
			NSLog(@"NULL element in array in reorder childs");
			continue;
		}
		NSXMLElement *parentE = (NSXMLElement*)[childE parent];
		if (!parentE) {
			NSLog(@"Element without parent in reorder childs.");
			continue;
		}
		[childE detach];
		[parentE addChild:childE];
	}
}

#pragma mark -------------------------------- XPath

+ (NSArray *)nodesForXPath:(NSString *)xpath withRoot:(NSXMLElement*)rootE
{
	NSAssert(rootE, @"rootE");
	NSError *__autoreleasing error = NULL;
	NSArray* a = [rootE nodesForXPath:xpath error:&error];
	if (error)
		NSLog(@"Xml.nodesForXPath failed with '%@' for xpath '%@' at location '%@'",
			  error, xpath, [rootE XPath]);
	if (!a)
		a = @[];
	return a;
}

+ (NSXMLNode *)nodeForXPath:(NSString *)xpath withRoot:(NSXMLElement*)rootE
{
	NSArray* a = [self nodesForXPath:xpath withRoot:rootE];
	return [a count] > 0 ? a[0] : NULL;
}

+ (void)remove:(NSString*)xpath withRoot:(NSXMLElement*)rootE
{
	NSArray* a = [self nodesForXPath:xpath withRoot:rootE];
	for (NSUInteger i = 0; i < [a count]; i++) {
		NSXMLNode* e = a[i];
		[e detach];
	}
}

+ (void)zapChildsAndAttributes:(NSXMLElement*)e
{
	for (NSXMLNode* a in [e attributes]) {
		if (![[a name] isEqual:@"rule-id"]) {
			//NSLog(@"Zaps attr: %@", a);
			[a detach];
		}
	}
	[e setChildren:@[]];
}

#pragma mark -------------------------------- simple get/set

+ (NSString*)stringForXPath:(NSString*)xpath withRoot:(NSXMLElement*)rootE
{
	NSArray* a = [Xml nodesForXPath:xpath withRoot:rootE];
	return [a count] ? [a[0] stringValue] : NULL;
}

+ (void)setText:(NSString*)value
	   forXPath:(NSString*)xpath
	   withRoot:(NSXMLElement*)rootE
	 usingCDATA:(BOOL)usingCDATA
{
	NSArray* pathItems = [xpath componentsSeparatedByString:@"/"];
	NSUInteger num = [pathItems count];
	NSXMLElement* e = rootE;
	for (NSUInteger i = 0; i < num; i++) {
		NSString* s = pathItems[i];
		if (![s length]) {
			[NSException raise:@"Invalid setString xpath"
						format:@"xpath = '%@'", xpath];
		}
		if ([s isEqual:@"."])
			continue;
		if ([s hasPrefix:@"@"]) {
			if (i + 1 < num) {
				[NSException raise:@"Invalid setString xpath"
							format:@"xpath = '%@'", xpath];
			}
			if (![Xml isSimpleName:s startIndex:1]) {
				[NSException raise:@"Invalid attribute name in setString xpath"
							format:@"xpath = '%@'", xpath];
			}
			NSString* attrName = [s substringFromIndex:1];
			[Xml setAttribute:value withName:attrName inElement:e];
			return;
		}
		if (![Xml isSimpleName:s]) {
			[NSException raise:@"Invalid element name in setString xpath"
						format:@"xpath = '%@'", xpath];
		}
		NSArray* childs = [e elementsForName:s];
		if ([childs count]) {
			e = childs[0];
		} else {
			NSXMLElement* newE = [Xml createElementWithName:s];
			[e addChild:newE];
			e = newE;
		}
	}
	for (NSXMLElement* zapE in [e children])
		[zapE detach];
	if (![value length])
		return;
	if (usingCDATA) {
		//NSLog(@"SetCDATA in %@: {{{%@}}}", [e name], value);
		NSXMLNode* cdata = [[NSXMLNode alloc] initWithKind:NSXMLTextKind options:NSXMLNodeIsCDATA];
		[cdata setStringValue:value];
		[e addChild:cdata];
	} else {
		[e setStringValue:value];
	}
}

+ (void)setString:(NSString*)value forXPath:(NSString*)xpath withRoot:(NSXMLElement*)rootE
{
	[self setText:value forXPath:xpath withRoot:rootE usingCDATA:NO];
}

+ (void)setCData:(NSString*)value forXPath:(NSString*)xpath withRoot:(NSXMLElement*)rootE
{
	[self setText:value forXPath:xpath withRoot:rootE usingCDATA:YES];
}

+ (double)doubleForXPath:(NSString*)xpath withRoot:(NSXMLElement*)rootE
{
	return [[Xml stringForXPath:xpath withRoot:rootE] doubleValue];
}

+ (void)setDouble:(double)value forXPath:(NSString*)xpath withRoot:(NSXMLElement*)rootE
{
	[Xml setString:[NSString stringWithFormat:@"%f", value] forXPath:xpath withRoot:rootE];
}

+ (int)intForXPath:(NSString*)xpath withRoot:(NSXMLElement*)rootE
{
	return [[Xml stringForXPath:xpath withRoot:rootE] intValue];
}

+ (void)setInt:(int)value forXPath:(NSString*)xpath withRoot:(NSXMLElement*)rootE
{
	[Xml setString:[NSString stringWithFormat:@"%d", value] forXPath:xpath withRoot:rootE];
}

+ (BOOL)boolForXPath:(NSString*)xpath withRoot:(NSXMLElement*)rootE
{
	NSString* s = [Xml stringForXPath:xpath withRoot:rootE];
	return s && [s length] && ![s isEqual:@"false"] && ![s isEqual:@"0"];
}

+ (void)setBool:(BOOL)value forXPath:(NSString*)xpath withRoot:(NSXMLElement*)rootE
{
	[Xml setString:(value ? @"1" : NULL) forXPath:xpath withRoot:rootE];
}

+ (NSString*)dateFromMillisecsXPath:(NSString*)xpath withRoot:(NSXMLElement*)rootE
{
	double msecs = [Xml doubleForXPath:xpath withRoot:rootE];
	if (!msecs)
		return NULL;
	NSDate* date = [NSDate dateWithTimeIntervalSince1970:msecs / 1000.0];
	//
	NSDateFormatter* fmt = [[NSDateFormatter alloc] init];
	[fmt setDateStyle:NSDateFormatterMediumStyle];
	[fmt setTimeStyle:NSDateFormatterShortStyle];
	return [fmt stringFromDate:date];
}

#pragma mark -------------------------------- regexp

+ (BOOL)regex:(NSString*)regex matchesString:(NSString*)text
{
	NSString *fmt = [NSString stringWithFormat:@"SELF MATCHES '%@'", regex];
	NSPredicate *predicate = [NSPredicate predicateWithFormat:fmt];
	return [predicate evaluateWithObject:text];
}

+ (NSString*)getHttpUrlSyntaxError:(NSString*)url
{
	if (![url length])
		return @"No URL was specified";
	if (![url hasPrefix:@"http://"] && ![url hasPrefix:@"https://"])
		return @"URL must start with 'http://' or 'https://'";
	if (![NSURL URLWithString:url])
		return @"Invalid URL (NSURL)";
	if (![Xml regex:@"^https?://[^:/]+(:[0-9]+)?(/[^[:space:]]*)?[:space:]*$" matchesString:url])
		return @"Invalid URL syntax";
	return NULL;
}

#pragma mark -------------------------------- unqiue id

// Faster than using XPath. Was bottleneck in loading new filters.
+ (BOOL)hasAttribute:(NSString*)attrName inSubtree:(NSXMLElement*)parentE withValue:(NSString*)value
{
	NSString* s = [[parentE attributeForName:attrName] stringValue];
	if (s && [s isEqualToString:value])
		return YES;
	for (NSXMLNode* childN in [parentE children]) {
		if ([childN kind] == NSXMLElementKind
			&& [Xml hasAttribute:attrName inSubtree:(NSXMLElement*)childN withValue:value]) {
			return YES;
		}
	}
	return NO;
}

+ (int)makeUniqueIdWithRoot:(NSXMLElement*)rootE withAttributeName:(NSString*)attrName
{
	while (true) {
		int r = (int)arc4random() & 0x3FFFFFFF;
		if (r < 1234567)
			continue;
		NSString* value = [NSString stringWithFormat:@"%d", r];
		if (![Xml hasAttribute:attrName inSubtree:rootE withValue:value])
			return r;
	}
}

+ (int)makeUniqueIdWithRoot:(NSXMLElement*)rootE
{
	return [Xml makeUniqueIdWithRoot:rootE withAttributeName:@"id"];
}

#pragma mark -------------------------------- print-to-string

+ (NSData*)documentToData:(NSXMLDocument*)doc
{
	int options = NSXMLNodePrettyPrint|NSXMLNodePreserveCDATA;
	NSMutableData* data = [NSMutableData dataWithData:[doc XMLDataWithOptions:options]];
	[data appendBytes:"\n" length:1]; // XMLDataWithOptions doesn't end with newline.
	return data;
}

+ (NSString*)documentToString:(NSXMLDocument*)doc;
{
	NSData* data = [Xml documentToData:doc];
	return [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
}

// Cocoa doesn't support this for anything but full documents,
// and using [NSString stringWithFormat:@"%@", rootE] is dubious long-term.
+ (NSData*)subtreeToData:(NSXMLElement*)parentE;
{
	parentE = [parentE copy];
	NSXMLDocument* doc = [[NSXMLDocument alloc] initWithRootElement:parentE];
	NSData* data = [Xml documentToData:doc];
	return data;
}

+ (NSString*)subtreeToString:(NSXMLElement*)parentE
{
	NSData* data = [Xml subtreeToData:parentE];
	return [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
}

@end
