/* Copyright (C) 2008-2012 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "XmlDoc.h"
#import "Xml.h"
#import "GBDebug.h"

@implementation XmlDoc

#ifdef DEBUG
//#define DEBUG_LOG_XML 1
#endif

- (instancetype)initWithRootElement:(NSXMLElement*)rootE
{
	if (!(self = [super init]))
		return NULL;
	[self reloadWithRootElement:rootE];
	return self;
}

- (instancetype)initWithRootElementTagName:(NSString*)tagName
{
	if (!(self = [super init]))
		return NULL;
	[self reloadWithRootElement:[Xml createElementWithName:tagName]];
	return self;
}

- (instancetype)initWithURL:(NSURL*)url withData:(NSData*)data
{
	if (!(self = [super init]))
		return self;
	_url = url;
	NSError* __autoreleasing tmp = NULL;
	_doc = [[NSXMLDocument alloc] initWithData:data
									   options:NSXMLNodePreserveCDATA
										 error:&tmp];
	_error = tmp;
#ifdef DEBUG_LOG_XML
	NSLog(@"Got xml: %@", _doc);
#endif
	if (_error) {
		NSLog(@"GlimmerBlocker got error parsing xml from: %@", _url);
		NSLog(@"  error: %@", _error);
	}
	if (_doc && !_error) {
		_rootE = [_doc rootElement];
		if (!_rootE) {
			NSLog(@"Invalid xml document (no root element) loaded from %@", _url);
			_doc = NULL;
		}
	}
	return self;
}

- (void)releaseDoc
{
	_doc = NULL;
	_rootE = NULL;
	_error = NULL;
}

- (void)dealloc
{
	[self releaseDoc];
}

- (void)reloadWithRootElement:(NSXMLElement*)rootE
{
	NSAssert(rootE, @"rootE");
	[self releaseDoc];
	_rootE = rootE;
	NSXMLNode* parent = [rootE parent];
	if (parent) {
		NSAssert([parent isKindOfClass:[NSXMLDocument class]], @"root-parent is doc");
		_doc = (NSXMLDocument*)parent;
	} else {
		_doc = [[NSXMLDocument alloc] initWithRootElement:_rootE];
	}
	[_doc setCharacterEncoding:@"UTF-8"]; // used when saving document.
	[_doc setDocumentContentKind:NSXMLDocumentXMLKind];
}

- (NSURL*) url
{
	return _url;
}

- (NSError*) error
{
	return _error;
}

- (NSXMLDocument*) doc
{
	return _doc;
}

- (NSXMLElement*) rootE
{
	return _rootE;
}

- (NSXMLElement*) byXPath:(NSString*)xpath
{
	NSError*__autoreleasing error = NULL;
	NSArray* a = [_rootE nodesForXPath:xpath error:&error];
	if (error) {
		NSLog(@"XmlDoc.byXPath failed for %@", xpath);
		NSLog(@"    %@", error);
		return NULL;
	}
	if ([a count] > 1)
		NSLog(@"Warning: matched %ld elements for: %@", (long)[a count], xpath);
	return [a count] ? a[0] : NULL;
}

- (NSXMLElement*)createRootChildWithName:(NSString*)name
{
	NSXMLElement* e = [Xml createElementWithName:name];
	[_rootE addChild:e];
	return e;
}

- (NSData*)printToData;
{
	return [Xml documentToData:_doc];
}

- (NSString*)printToString
{
	return [Xml documentToString:_doc];
}

#pragma mark -------------------------------- XPath scalars

- (NSString*)stringForXPath:(NSString*)xpath
{
	return [Xml stringForXPath:xpath withRoot:_rootE];
}

- (void)setString:(NSString*)value forXPath:(NSString*)xpath
{
	[Xml setString:value forXPath:xpath withRoot:_rootE];
}

- (double)doubleForXPath:(NSString*)xpath
{
	return [Xml doubleForXPath:xpath withRoot:_rootE];
}

- (void)setDouble:(double)value forXPath:(NSString*)xpath
{
	[Xml setDouble:value forXPath:xpath withRoot:_rootE];
}

- (int)intForXPath:(NSString*)xpath
{
	return [Xml intForXPath:xpath withRoot:_rootE];
}

- (void)setInt:(int)value forXPath:(NSString*)xpath
{
	[Xml setInt:value forXPath:xpath withRoot:_rootE];
}

- (BOOL)boolForXPath:(NSString*)xpath
{
	return [Xml boolForXPath:xpath withRoot:_rootE];
}

- (void)setBool:(BOOL)value forXPath:(NSString*)xpath
{
	[Xml setBool:value forXPath:xpath withRoot:_rootE];
}

- (NSString*)dateFromMillisecsXPath:(NSString*)xpath
{
	return [Xml dateFromMillisecsXPath:xpath withRoot:_rootE];
}

#pragma mark -------------------------------- XPath

- (NSArray *)nodesForXPath:(NSString *)xpath
{
	return [Xml nodesForXPath:xpath withRoot:_rootE];
}

- (NSXMLNode *)nodeForXPath:(NSString *)xpath
{
	return [Xml nodeForXPath:xpath withRoot:_rootE];
}

- (void)removeElementsForXPath:(NSString*)xpath
{
	[Xml remove:xpath withRoot:_rootE];
}

- (NSXMLElement*)singletonElementForXPath:(NSString*)xpath
{
	return [Xml singletonElementForXPath:xpath forParent:_rootE];
}

- (NSXMLElement*)singletonElementForXPath:(NSString*)xpath
							 asFirstChild:(BOOL)asFirstChild
{
	return [Xml singletonElementForXPath:xpath forParent:_rootE asFirstChild:asFirstChild];
}

#pragma mark -------------------------------- unqiue id

- (int)makeUniqueId
{
	return [Xml makeUniqueIdWithRoot:_rootE];
}

@end
